// Fill out your copyright notice in the Description page of Project Settings.

#include "ActorComponents/DDCompositeSkeletalMeshComponent.h"

#include "DataAssets/DDDataAssets.h"

#include "Engine/SkinnedAssetCommon.h"
#include "Engine/Texture2D.h"
#include "Engine/SkeletalMesh.h"
#include "Rendering/SkeletalMeshRenderData.h"
#include "Materials/MaterialInstanceDynamic.h"
#include "Animation/Skeleton.h"
#include "Animation/AnimSequence.h"
#include "Animation/AnimInstanceProxy.h"

#include "Controllers/DDTextureEditingController.h"
#include "Controllers/ImageProcessor/DDBlendTexture.h"
#include "Controllers/ImageProcessor/DDColorize.h"

#include "Components/SkeletalMeshComponent.h"
#include <SkeletalMeshAttributes.h>


UDDCompositeSkeletalMeshComponent::UDDCompositeSkeletalMeshComponent()
{
}

void UDDCompositeSkeletalMeshComponent::CompositeCharacter()
{
	SetMorphTargets();
	SetTextures();
	ApplyEdits();
}

void UDDCompositeSkeletalMeshComponent::OnRegister()
{
	Super::OnRegister();
	Bake();
	CompositeCharacter();
}

void UDDCompositeSkeletalMeshComponent::BeginPlay()
{
	Super::BeginPlay();
	Bake();
	CompositeCharacter();
}

void UDDCompositeSkeletalMeshComponent::CompositeSkin(USkeletalMesh* SKM)
{
	check(CharacterDescriptorData);
	FDDCharacter& Character = CharacterDescriptorData->groups[GroupIndex].characters[CharacterIndex];

	TArray<FSkeletalMaterial>& Materials_SM = SKM->GetMaterials();

	// For each skinned render object in the template asset
	for (auto& TemplateSkinnedRenderObject : AssetCatalogData->templateAssets[Character.blends[0].templateAssetIndex].skinnedRenderObjects)
	{
		TArray<UTexture2D*> Textures;
		TArray<float> Weights;
		float totalWeight = 0.0f;

		// For each texture weight in the character
		for (auto& TextureWeight : Character.blends[0].textureWeights)
		{
			// looking for the texture in the template asset
			for (auto& ShapeSkinnedRenderObject : AssetCatalogData->shapeAssets[TextureWeight.index].skinnedRenderObjects)
			{
				if (ShapeSkinnedRenderObject.meshName.Equals(TemplateSkinnedRenderObject.meshName))
				{
					auto& ACMaterial = AssetCatalogData->materials[ShapeSkinnedRenderObject.materialExportIndex[0]];

					if (ACMaterial.textureExports.Num() > 0) {
						Textures.Add(AssetCatalogData->textures[ACMaterial.textureExports[0].textureVariationIndices[0]].texture);
						Weights.Add(TextureWeight.weight);
						totalWeight += TextureWeight.weight;
					}
				}
			}
		}

		if (Textures.Num() > 0)
		{

			// Searching the material we need
			FSkeletalMaterial* Material_SM = nullptr;
			int32 MaterialIndex = 0;
			for (size_t i = 0; i < Materials_SM.Num(); i++)
			{
				if (Materials_SM[i].MaterialSlotName.IsEqual(*TemplateSkinnedRenderObject.materialNames[0], ENameCase::IgnoreCase))
				{
					Material_SM = &Materials_SM[i];
					MaterialIndex = i;
					break;
				}
			}

			if (Material_SM == nullptr)
			{
				UE_LOG(LogTemp, Warning, TEXT("Material %s does not exist"), *TemplateSkinnedRenderObject.materialNames[0]);
				continue;
			}

			// Adding the texture already existing in the material (the template's default)
			for (auto& TextureExport : AssetCatalogData->materials[TemplateSkinnedRenderObject.materialExportIndex[0]].textureExports)
			{
				if (TextureExport.slot.Compare("ALBEDO") == 0) // Assuming only albedo for now
				{
					Textures.Add(AssetCatalogData->textures[TextureExport.textureVariationIndices[0]].texture);
					Weights.Add(1.0f - totalWeight);
					break;
				}
			}

			// Overlays
			TArray<FDDVariantRegion> characterVariantRegions;
			for (auto& overlayAssetIndex : Character.overlayAssetsIndices)
			{
				auto& overlayAsset = AssetCatalogData->overlayAssets[overlayAssetIndex.overlayAssetIndex];
				for (auto& overlayMaterialConnection : overlayAsset.overlayMaterialConnections)
				{
					auto& assetCatalogMaterial = AssetCatalogData->materials[overlayMaterialConnection.materialIndex];
					if (TemplateSkinnedRenderObject.materialNames[0].ToLower().Equals(assetCatalogMaterial.name.ToLower()))
					{
						for (auto& atlasInformation : overlayMaterialConnection.overlayAtlasInformations)
						{
							characterVariantRegions.Add(atlasInformation.variantRegions[overlayAssetIndex.variationIndex]);
						}
					}
				}
			}

			// Build the texture composition graph
			BuildTextureCompositionGraph(TemplateSkinnedRenderObject.meshName, Textures, Weights);

			// Process the graph
			UDDTextureEditingController::ProcessImageGameThread(MeshNameToImageProcessorArray[TemplateSkinnedRenderObject.meshName].ImageProcessors.Last());

			// Make the material dynamic and put the correct processed texture on it
			UMaterialInterface* OldMaterial = GetMaterial(MaterialIndex);
			UMaterialInstanceDynamic* DynamicMaterial = CreateAndSetMaterialInstanceDynamic(MaterialIndex);
			if (DynamicMaterial)
			{
				FName TextureParameterName = "BC_Texture";
				UTextureRenderTarget2D* RT = MeshNameToImageProcessorArray[TemplateSkinnedRenderObject.meshName].ImageProcessors.Last()->GetResultRenderTarget();
				DynamicMaterial->SetTextureParameterValue(TextureParameterName, RT);
			}
		}
	}
}

void UDDCompositeSkeletalMeshComponent::BuildTextureCompositionGraph(FString MeshName, TArray<UTexture2D*> Textures,
	TArray<float> Weights)
{
	TArray<UDDImageProcessor*> ImageProcessors;

	// Adding the image processors
	for (auto& Texture : Textures)
	{
		UDDImageTexture* ImageTexture = NewObject<UDDImageTexture>();
		ImageTexture->Initialize(Texture->GetSizeX(), Texture->GetSizeY());
		ImageTexture->SetInputTexture(Texture);
		ImageProcessors.Add(ImageTexture);
	}

	// Using the colorize to extract the weight of the texture
	for (int32 i = 0; i < Textures.Num(); i++)
	{
		float Weight = Weights[i];
		UDDColorize* Colorize = NewObject<UDDColorize>();
		Colorize->Initialize(Textures[i]->GetSizeX(), Textures[i]->GetSizeY());
		Colorize->SetDependency(ImageProcessors[i]);
		Colorize->SetColor(FColor(255 * Weight, 255 * Weight, 255 * Weight, 255));
		ImageProcessors.Add(Colorize);
	}

	// Doing the blending
	int32 Operator1 = Textures.Num(); // This is the index of the first colorizer
	for (int32 i = 1; i < Textures.Num(); i++)
	{
		int32 Operator2 = Textures.Num() + i;
		int32 sizeX = std::max(ImageProcessors[Operator1]->GetResultRenderTarget()->SizeX, ImageProcessors[Operator2]->GetResultRenderTarget()->SizeX);
		int32 sizeY = std::max(ImageProcessors[Operator1]->GetResultRenderTarget()->SizeY, ImageProcessors[Operator2]->GetResultRenderTarget()->SizeY);
		UDDBlendTexture* BlendTexture = NewObject<UDDBlendTexture>();
		BlendTexture->Initialize(sizeX, sizeY);
		BlendTexture->SetOperation(EDDBlendTextureOperation::BTO_Add);
		BlendTexture->SetDependency(ImageProcessors[Operator1], 0);
		BlendTexture->SetDependency(ImageProcessors[Operator2], 1);
		ImageProcessors.Add(BlendTexture);

		Operator1 = ImageProcessors.Num() - 1; // Setting the next operation to be this blend
	}

	FImageProcessorArray ImageProcessorArray;
	ImageProcessorArray.ImageProcessors = ImageProcessors;

	MeshNameToImageProcessorArray.Add(MeshName, ImageProcessorArray);
}

void UDDCompositeSkeletalMeshComponent::TickComponent(float DeltaTime, ELevelTick TickType,
	FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	//BoneVisualizer(this);
	//BoneDebugger(this, BoneTransformFrame1);
	//BoneDebugger(this);
}

void UDDCompositeSkeletalMeshComponent::Bake()
{
	//USkinnedAsset* NewMesh = DuplicateObject<USkinnedAsset>(GetSkinnedAsset(), nullptr);
	//if (NewMesh && NewMesh->GetResourceForRendering())
	//{
	//	SetSkinnedAsset(NewMesh);
	//}

	TArray<USceneComponent*> Children;
	this->GetChildrenComponents(true, Children);
	//for (auto Child : Children)
	//{
	//	if (USkeletalMeshComponent* Garment = Cast<USkeletalMeshComponent>(Child))
	//	{
	//		USkinnedAsset* NewGarment = DuplicateObject<USkinnedAsset>(Garment->GetSkinnedAsset(), nullptr);
	//		if (NewGarment && NewGarment->GetResourceForRendering())
	//		{
	//			Garment->SetSkinnedAsset(NewGarment);
	//			//SetRig(Garment, BoneTransformsPerFrame);
	//		}
	//	}
	//}

	// Setting the rigs
	auto BoneTransformsPerFrame = GetRig(GetSkinnedAsset(), AnimationSequence);
	SetRig(this, BoneTransformsPerFrame);
	for (auto Child : Children)
	{
		if (USkeletalMeshComponent* Garment = Cast<USkeletalMeshComponent>(Child))
		{
			SetRig(Garment, BoneTransformsPerFrame);
		}
	}
}

TArray<TArray<FTransform>> UDDCompositeSkeletalMeshComponent::GetRig(const USkinnedAsset* SKAsset,
	const UAnimSequence* AnimSequence)
{
	if (!AnimSequence || !SKAsset)
	{
		UE_LOG(LogTemp, Warning, TEXT("Invalid AnimationSequence or SkeletalMesh"));
		return TArray<TArray<FTransform>>();
	}

	int32 TotalFrames = AnimSequence->GetNumberOfSampledKeys();
	float AnimationLength = AnimSequence->GetPlayLength();

	// Ensuring we have a valid skeleton
	const USkeleton* Skeleton = SKAsset->GetSkeleton();

	if (!Skeleton)
	{
		UE_LOG(LogTemp, Warning, TEXT("Skeleton is invalid"));
		return TArray<TArray<FTransform>>();
	}

	// Create arrays to store bone transforms for each frame
	TArray<TArray<FTransform>> BoneTransformsPerFrame;
	BoneTransformsPerFrame.SetNum(TotalFrames);

	// Get reference skeleton to access bone hierarchy
	const FReferenceSkeleton& RefSkeleton = Skeleton->GetReferenceSkeleton();
	const int32 NumBones = RefSkeleton.GetNum();

	// Initialize storage for each frame
	for (auto& FrameTransforms : BoneTransformsPerFrame)
	{
		FrameTransforms.SetNum(NumBones);
	}

	// Initialize FBoneContainer with the required bones and the skeleton
	TArray<FBoneIndexType> ReqBones;
	for (int32 BoneIndex = 0; BoneIndex < RefSkeleton.GetNum(); ++BoneIndex)
	{
		ReqBones.Add(static_cast<FBoneIndexType>(BoneIndex));
	}
	TArrayView<FBoneIndexType> ReqBonesArrV(ReqBones);
	FBoneContainer BoneContainer(ReqBonesArrV, UE::Anim::FCurveFilterSettings(),
		*static_cast<UObject*>(const_cast<USkeleton*>(Skeleton)));

	// Iterate through each frame
	for (int32 Frame = 0; Frame < TotalFrames; ++Frame)
	{
		// Calculate time for this frame
		float Time = (AnimationLength / (TotalFrames - 1)) * Frame;

		UE_LOG(LogTemp, Warning, TEXT("Frame %d: Time %f"), Frame, Time);

		// Initialize pose data
		FCompactPose Pose;
		FBlendedCurve Curve;
		UE::Anim::FStackAttributeContainer Attributes;
		FMemMark Mark(FMemStack::Get());
		FAnimationPoseData AnimationPoseData(Pose, Curve, Attributes);
		Pose.SetBoneContainer(&BoneContainer);

		// Extract bone pose for the current frame
		FAnimExtractContext ExtractionContext(Time, false);
		AnimSequence->GetBonePose(AnimationPoseData, ExtractionContext);

		// Get bone transforms for this frame
		for (int32 BoneIndex = 0; BoneIndex < NumBones; ++BoneIndex)
		{
			FTransform BoneTransform = Pose.GetBones()[BoneIndex];
			BoneTransformsPerFrame[Frame][BoneIndex] = BoneTransform;

			UE_LOG(LogTemp, Warning, TEXT("Bone %d: Transform %s"), BoneIndex, *BoneTransform.ToString());
		}
	}
	return BoneTransformsPerFrame;
}

void UDDCompositeSkeletalMeshComponent::BoneDebugger(USkeletalMeshComponent* SKMC)
{
	const FReferenceSkeleton& RefSkeleton = SKMC->GetSkinnedAsset()->GetRefSkeleton();
	for (int32 BoneIndex = 0; BoneIndex < RefSkeleton.GetNum(); ++BoneIndex)
	{
		FVector BoneLocationWorldSpace;
		{
			FTransform BoneLocalToComponentTransform = RefSkeleton.GetRefBonePose()[BoneIndex];//SKMC->GetBoneTransform(BoneIndex).GetLocation();
			FMatrix44f BoneLocalToComponentMatrix = (FMatrix44f)BoneLocalToComponentTransform.ToMatrixWithScale();
			FMatrix44f ComponentToWorldMatrix = (FMatrix44f)SKMC->GetComponentTransform().ToMatrixWithScale();
			FMatrix44f BoneLocalToWorld = BoneLocalToComponentMatrix * ComponentToWorldMatrix;

			BoneLocationWorldSpace = (FVector)BoneLocalToWorld.TransformPosition(FVector3f(0, 0, 0));
			DrawDebugSphere(SKMC->GetWorld(), BoneLocationWorldSpace, 5.0f, 12, FColor::Red, false, -1.0f, 0);
		}


		// Draw bone locations
		int32 ParentIndex = RefSkeleton.GetParentIndex(BoneIndex);

		// Draw connections to parent bones
		if (ParentIndex != INDEX_NONE)
		{
			FTransform ParentLocation = RefSkeleton.GetRefBonePose()[ParentIndex];
			FMatrix44f BoneLocalToComponentMatrix = (FMatrix44f)ParentLocation.ToMatrixWithScale();
			FMatrix44f ComponentToWorldMatrix = (FMatrix44f)SKMC->GetComponentTransform().ToMatrixWithScale();
			FMatrix44f BoneLocalToWorld = BoneLocalToComponentMatrix * ComponentToWorldMatrix;

			FVector ParentBoneLocationWorldSpace = (FVector)BoneLocalToWorld.TransformPosition(FVector3f(0, 0, 0));
			DrawDebugLine(SKMC->GetWorld(), BoneLocationWorldSpace, ParentBoneLocationWorldSpace, FColor::Green, false, -1.0f, 0, 1.0f);
		}
	}
}

void UDDCompositeSkeletalMeshComponent::ApplyEdits()
{
	// Geometry editing
	ENQUEUE_RENDER_COMMAND(LockAndModifyVertexBuffer)(
		[this](FRHICommandListImmediate& RHICmdList)
		{
			auto RenderResource = GetSkinnedAsset()->GetResourceForRendering();
			auto LOD = &RenderResource->LODRenderData[0];
			auto StaticVertexBuffers = &LOD->StaticVertexBuffers;
			auto PositionVB = &StaticVertexBuffers->PositionVertexBuffer;
			check(PositionVB->GetAllowCPUAccess());
			auto ColorVB = &StaticVertexBuffers->ColorVertexBuffer;
			check(ColorVB->GetAllowCPUAccess());
			auto StaticVertexBuffer = &StaticVertexBuffers->StaticMeshVertexBuffer;
			check(StaticVertexBuffer->GetAllowCPUAccess());
			auto TangentsVB = &StaticVertexBuffer->TangentsVertexBuffer;


			// Update positions and tangents (tangents include normals)
			void* PositionVertexBufferData = RHICmdList.LockBuffer(PositionVB->VertexBufferRHI, 0, PositionVB->GetAllocatedSize(), RLM_WriteOnly);
			void* ColorVertexBufferData = RHICmdList.LockBuffer(ColorVB->VertexBufferRHI, 0, ColorVB->GetAllocatedSize(), RLM_WriteOnly);
			for (int32 SectionIndex = 0; SectionIndex < LOD->RenderSections.Num(); SectionIndex++)
			{
				auto Section_SM = &LOD->RenderSections[SectionIndex];
				const int32 NumSoftVertices = Section_SM->GetNumVertices();
				for (int32 VertexIndex = 0; VertexIndex < NumSoftVertices; VertexIndex++)
				{
					const int32 VBIndex = Section_SM->GetVertexBufferIndex() + VertexIndex;


					float VertexOpacity = PositionVB->VertexPosition(VBIndex).Z / 10.0;
					VertexOpacity = sinf(VertexOpacity);

					VertexOpacity = 1.0;
					for (int32 i = 0; i < AttachmentCullData.Num(); i++)
					{
						if (AttachmentCullData[i] != nullptr && !AttachmentCullData[i]->VertexOpacity.IsEmpty())
							VertexOpacity *= AttachmentCullData[i]->VertexOpacity[VBIndex];
					}
					
					ColorVB->VertexColor(VBIndex) = FColor(255u * VertexOpacity, 0, 0, 255u);
				}
			}
			FMemory::Memcpy(PositionVertexBufferData, PositionVB->GetVertexData(), PositionVB->GetAllocatedSize());
			FMemory::Memcpy(ColorVertexBufferData, ColorVB->GetVertexData(), ColorVB->GetAllocatedSize());
			RHICmdList.UnlockBuffer(PositionVB->VertexBufferRHI);
			RHICmdList.UnlockBuffer(ColorVB->VertexBufferRHI);
		}
		);

	// Finish
	MarkRenderStateDirty();
}

void UDDCompositeSkeletalMeshComponent::BoneDebugger(USkeletalMeshComponent* SKMC, TArray<FTransform> Transforms)
{
	const FReferenceSkeleton& RefSkeleton = SKMC->GetSkinnedAsset()->GetRefSkeleton();
	TArray<FMeshBoneInfo> SkelRefBonesInfo = RefSkeleton.GetRefBoneInfo();

	check(RefSkeleton.GetNum() == Transforms.Num());

	for (int32 BoneIndex = 0; BoneIndex < RefSkeleton.GetNum(); ++BoneIndex)
	{
		// Draw bone locations
		{
			FTransform BoneLocalToComponentTransform = Transforms[BoneIndex];
			FMatrix44f BoneLocalToComponentMatrix = (FMatrix44f)BoneLocalToComponentTransform.ToMatrixWithScale();
			FMatrix44f ComponentToWorldMatrix = (FMatrix44f)SKMC->GetComponentTransform().ToMatrixWithScale();
			FMatrix44f BoneLocalToWorld = BoneLocalToComponentMatrix * ComponentToWorldMatrix;

			FVector BoneLocationWorldSpace = (FVector)BoneLocalToWorld.TransformPosition(FVector3f(0, 0, 0));

			DrawDebugSphere(SKMC->GetWorld(), BoneLocationWorldSpace, 2.0f, 12, FColor::Yellow, false, -1.0f, 0);
		}
	}
}

void UDDCompositeSkeletalMeshComponent::BoneVisualizer(USkeletalMeshComponent* SKMC)
{
	const FReferenceSkeleton& RefSkeleton = SKMC->GetSkinnedAsset()->GetRefSkeleton();
	TArray<FMeshBoneInfo> SkelRefBonesInfo = RefSkeleton.GetRefBoneInfo();

	for (int32 BoneIndex = 0; BoneIndex < RefSkeleton.GetNum(); ++BoneIndex)
	{
		FVector ChildBoneLocationWorldSpace;
		// Draw bone locations
		{
			FTransform BoneLocalToWorldTransform = SKMC->GetBoneTransform(SkelRefBonesInfo[BoneIndex].Name, ERelativeTransformSpace::RTS_World);
			FMatrix44f BoneLocalToWorldMatrix = (FMatrix44f)BoneLocalToWorldTransform.ToMatrixWithScale();
			FMatrix44f WorldToComponentMatrix = (FMatrix44f)SKMC->GetComponentTransform().ToMatrixWithScale().Inverse();
			FMatrix44f WorldToLocal_overrride = WorldToComponentMatrix * SKMC->GetRefPoseOverride()->RefBasesInvMatrix[BoneIndex];

			FMatrix44f FinalBoneMatrix = BoneLocalToWorldMatrix * WorldToLocal_overrride * BoneLocalToWorldMatrix;
			FVector BoneLocationWorldSpace = (FVector)FinalBoneMatrix.TransformPosition(FVector3f(0, 0, 0));

			DrawDebugSphere(SKMC->GetWorld(), BoneLocationWorldSpace, 2.0f, 12, FColor::Red, false, -1.0f, 0);
			ChildBoneLocationWorldSpace = BoneLocationWorldSpace;
		}

		// Draw connections to parent bones
		int32 ParentIndex = RefSkeleton.GetParentIndex(BoneIndex);
		if (ParentIndex != INDEX_NONE)
		{
			FTransform BoneLocalToWorldTransform = SKMC->GetBoneTransform(SkelRefBonesInfo[ParentIndex].Name, ERelativeTransformSpace::RTS_World);
			FMatrix44f BoneLocalToWorldMatrix = (FMatrix44f)BoneLocalToWorldTransform.ToMatrixWithScale();
			FMatrix44f WorldToComponentMatrix = (FMatrix44f)SKMC->GetComponentTransform().ToMatrixWithScale().Inverse();
			FMatrix44f WorldToLocal_overrride = WorldToComponentMatrix * SKMC->GetRefPoseOverride()->RefBasesInvMatrix[ParentIndex];

			FMatrix44f FinalBoneMatrix = BoneLocalToWorldMatrix * WorldToLocal_overrride * BoneLocalToWorldMatrix;
			FVector BoneLocationWorldSpace = (FVector)FinalBoneMatrix.TransformPosition(FVector3f(0, 0, 0));

			DrawDebugLine(SKMC->GetWorld(), ChildBoneLocationWorldSpace, BoneLocationWorldSpace, FColor::Green, false, -1.0f, 0, 1.0f);
		}
	}
}

void UDDCompositeSkeletalMeshComponent::SetRig(USkeletalMeshComponent* SKMC,
	TArray<TArray<FTransform>> BoneTransformsPerFrame)
{
	if (!SKMC || !SKMC->GetSkinnedAsset()->GetSkeleton())
	{
		UE_LOG(LogTemp, Warning, TEXT("Invalid USkinnedAsset"));
		return;
	}

	if (BoneTransformsPerFrame.Num() <= 0)
	{
		UE_LOG(LogTemp, Warning, TEXT("BoneTransformsPerFrame is invalid."));
		return;
	}
	USkeleton* OriginalSkeleton = SKMC->GetSkinnedAsset()->GetSkeleton();


	// Prepare bone hierarchy data
	TArray<FTransform> BoneTransforms;
	TArray<FMeshBoneInfo> BoneInfos;

	// Getting the bone transforms from the morphs and the animation
	TArray<FTransform> SkeletonBoneTransforms =
		GetBoneTransformsForMorp(OriginalSkeleton->GetReferenceSkeleton(), BoneTransformsPerFrame);

	for (int32 BoneIndex = 0; BoneIndex < OriginalSkeleton->GetReferenceSkeleton().GetNum(); ++BoneIndex)
	{
		FTransform BoneTransform = OriginalSkeleton->GetReferenceSkeleton().GetRefBonePose()[BoneIndex];
		FMeshBoneInfo BoneInfo = OriginalSkeleton->GetReferenceSkeleton().GetRefBoneInfo()[BoneIndex];
		BoneInfos.Add(BoneInfo);
		BoneTransforms.Add(SkeletonBoneTransforms[BoneIndex]);
	}

	//USkeletalMesh* Mesh = SKMC->SkeletalMesh;

	//// Modify reference skeleton
	//FReferenceSkeletonModifier SkeletonModifier(Mesh->RefSkeleton, Mesh->Skeleton);
	//Mesh->RefSkeleton.Empty();
	//for (int32 i = 0; i < BoneInfos.Num(); i++)
	//{
	//	SkeletonModifier.Add(BoneInfos[i], BoneTransforms[i]);
	//}

	//Mesh->RefSkeleton.RebuildRefSkeleton(Mesh->Skeleton, true);
	//Mesh->CalculateInvRefMatrices();

	SKMC->SetRefPoseOverride(SkeletonBoneTransforms);

	SKMC->UpdateBoneBodyMapping();
	SKMC->UpdateBounds();
	SKMC->MarkRenderStateDirty();
	SKMC->UpdateComponentToWorld();
}

TArray<FTransform> UDDCompositeSkeletalMeshComponent::GetBoneTransformsForMorp(FReferenceSkeleton RefSkeleton,
	TArray<TArray<FTransform>> BoneTransformsPerFrame)
{
	const int32 NumBones = BoneTransformsPerFrame[0].Num();
	TArray<FTransform> ReferencePoseTransforms = RefSkeleton.GetRefBonePose();
	TArray<FTransform> FinalBonesTransforms;
	for (int32 BoneIndex = 0; BoneIndex < NumBones; ++BoneIndex)
	{
		FTransform OriginalTransform = ReferencePoseTransforms[BoneIndex];
		FVector DeltaPosition = FVector::Zero();
		FQuat DeltaRotation = FQuat::Identity;
		for (auto& ShapeWeight : CharacterDescriptorData->groups[GroupIndex].characters[CharacterIndex].blends[0].shapeWeights)
		{
			int shapeIndex = AssetCatalogData->shapeAssets[ShapeWeight.index].shapeIndex;
			const FTransform& ShapeBoneTransform = BoneTransformsPerFrame[shapeIndex + 1][BoneIndex];

			// Position calculation 
			FVector PosChange = FMath::Lerp(OriginalTransform.GetLocation(), ShapeBoneTransform.GetLocation(), ShapeWeight.weight)
				- OriginalTransform.GetLocation();
			DeltaPosition += PosChange;

			// Rotation calculation 
			FQuat RotChange = OriginalTransform.GetRotation().Inverse() * 
				FQuat::Slerp(OriginalTransform.GetRotation(), ShapeBoneTransform.GetRotation(), ShapeWeight.weight);
			DeltaRotation = DeltaRotation * RotChange;
		}

		FTransform FinalBoneTransform = OriginalTransform;
		FinalBoneTransform.SetLocation(OriginalTransform.GetLocation() + DeltaPosition);
		FinalBoneTransform.SetRotation(OriginalTransform.GetRotation() * DeltaRotation);

		FinalBonesTransforms.Add(FinalBoneTransform);
	}

	return FinalBonesTransforms;
}

void UDDCompositeSkeletalMeshComponent::SetMorphTargets()
{
	check(CharacterDescriptorData);
	FDDCharacter& Character = CharacterDescriptorData->groups[GroupIndex].characters[CharacterIndex];

	TArray<USceneComponent*> Children;
	this->GetChildrenComponents(true, Children);
	for (const auto& ShapeWeight : Character.blends[0].shapeWeights)
	{
		SetMorphTarget(FName(*ShapeWeight.name), ShapeWeight.weight);

		// Setting morph targets for the Deformables (garments)
		for (auto Child : Children)
			if (USkeletalMeshComponent* Garment = Cast<USkeletalMeshComponent>(Child))
				if (Garment->FindMorphTarget(FName(*ShapeWeight.name)) != nullptr)
					Garment->SetMorphTarget(FName(*ShapeWeight.name), ShapeWeight.weight);
	}
}

void UDDCompositeSkeletalMeshComponent::SetTextureToMaterial(UMaterialInstanceDynamic* DynamicMaterial,
	const FDDTextureExport& TextureExport,
	const int32 VariationIndex = 0) const
{
	if (TextureExport.slot.Compare("ALBEDO") == 0)
	{
		int32 TextureIndex = TextureExport.textureVariationIndices[VariationIndex];
		UTexture2D* Texture = AssetCatalogData->textures[TextureIndex].texture;
		DynamicMaterial->SetTextureParameterValue("BC_Texture", Texture);
	}
	else if (TextureExport.slot.Compare("NORMAL") == 0)
	{
		int32 TextureIndex = TextureExport.textureVariationIndices[VariationIndex];
		UTexture2D* Texture = AssetCatalogData->textures[TextureIndex].texture;
		DynamicMaterial->SetTextureParameterValue("N_Texture", Texture);
	}
	else if (TextureExport.slot.Compare("MASK") == 0)
	{
		int32 TextureIndex = TextureExport.textureVariationIndices[VariationIndex];
		UTexture2D* Texture = AssetCatalogData->textures[TextureIndex].texture;
		DynamicMaterial->SetTextureParameterValue("ARM_Texture", Texture);
	}
}

void UDDCompositeSkeletalMeshComponent::SetTextures()
{
	check(CharacterDescriptorData);

	// Compositing the skin (with blendings)
	CompositeSkin(this->GetSkeletalMeshAsset());


	// Setting textures for the Deformables (garments)
	FDDCharacter& Character = CharacterDescriptorData->groups[GroupIndex].characters[CharacterIndex];
	TArray<USceneComponent*> Children;
	this->GetChildrenComponents(true, Children);
	for (auto& Child : Children)
	{
		if (USkeletalMeshComponent* Garment = Cast<USkeletalMeshComponent>(Child))
		{
			TArray<FSkeletalMaterial>& DeformableMaterials_SM = Garment->GetSkinnedAsset()->GetMaterials();
			for (auto& DeformableAssetIndex : Character.blends[0].deformableAssetsIndices)
			{
				if (DeformableAssetIndex.name.Equals(Garment->GetName()))
				{
					FDDDeformableAsset DeformableAsset = AssetCatalogData->deformableAssets[DeformableAssetIndex.
						deformableAssetIndex];
					for (auto& SkinnedRenderObject : DeformableAsset.skinnedRenderObjects)
					{
						for (auto& MaterialIndex : SkinnedRenderObject.materialExportIndex)
						{
							auto& AssetCatalogMaterial = AssetCatalogData->materials[MaterialIndex];
							for (size_t i = 0; i < DeformableMaterials_SM.Num(); i++)
							{
								auto Material_SM = &DeformableMaterials_SM[i];
								FString MaterialName_SM = Material_SM->MaterialSlotName.GetPlainNameString().ToLower();
								FString MaterialName_Template = AssetCatalogMaterial.name.ToLower();

								if (MaterialName_SM.StartsWith(MaterialName_Template))
									// Using StartsWith because Unreal add some things
								{
									UMaterialInterface* OldMaterial = Garment->GetMaterial(i);
									UMaterialInstanceDynamic* DynamicMaterial = Garment->
										CreateAndSetMaterialInstanceDynamic(i);
									if (DynamicMaterial)
									{
										for (auto& TextureExport : AssetCatalogMaterial.textureExports)
										{
											int32 VariationIndex = TextureExport.textureVariationIndices.Num() > 1
												? DeformableAssetIndex.variationIndex
												: 0;
											SetTextureToMaterial(DynamicMaterial, TextureExport, VariationIndex);
										}
									}
								}
							}
						}
					}
					break;
				}
			}
		}
	}
}