// Fill out your copyright notice in the Description page of Project Settings.

#include "ActorComponents/DDEditableSkeletalMeshComponent.h"
#include "RenderGraphUtils.h"
#include "RHI.h"
#include "RHIResources.h"
#include "RenderResource.h"
#include "GlobalShader.h"
#include "ShaderParameterUtils.h"
#include "Rendering/SkeletalMeshRenderData.h"
#include "Rendering/PositionVertexBuffer.h"
#include "Rendering/StaticMeshVertexBuffer.h"
#include "Rendering/SkeletalMeshLODModel.h"
#include "Rendering/SkeletalMeshModel.h"
#include "Engine/TextureRenderTarget2D.h"
#include "Engine/Texture2D.h"
#include "Engine/SkinnedAssetCommon.h"
#include "Materials/MaterialInstance.h"
#include "Materials/MaterialInstanceDynamic.h"


UDDEditableSkeletalMeshComponent::UDDEditableSkeletalMeshComponent()
{

}

void UDDEditableSkeletalMeshComponent::BeginPlay()
{
	Super::BeginPlay();

	//BeginEditing();
	//FinishEditing();

	//// Make sure the edits are applied before the next character is edited
	//FlushRenderingCommands();
}

void UDDEditableSkeletalMeshComponent::OnRegister() 
{
	Super::OnRegister();
	
}

void UDDEditableSkeletalMeshComponent::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);
	if (State == EDDEditableSkeletalMeshComponentState::ESMCS_Editing)
		ApplyEdits();
}

void UDDEditableSkeletalMeshComponent::BeginEditing()
{
	check(State == EDDEditableSkeletalMeshComponentState::ESMCS_Initial);
	State = EDDEditableSkeletalMeshComponentState::ESMCS_Editing;

	// Construct the UV lookup channel
	FSkeletalMeshRenderData* RenderData = SkeletalMesh->GetResourceForRendering();
	check(RenderData);
	check(RenderData->LODRenderData.Num() > 0);
	FSkeletalMeshLODRenderData& LODData = RenderData->LODRenderData[0];
	uint32 VertCount = LODData.StaticVertexBuffers.StaticMeshVertexBuffer.GetNumVertices();
	int TexDim = FMath::CeilToInt32(FMath::Sqrt((double)VertCount));
	uint32 UV_ChannelCount = LODData.StaticVertexBuffers.StaticMeshVertexBuffer.GetNumTexCoords();
	check(UV_ChannelCount == 2);
	ENQUEUE_RENDER_COMMAND(LockAndModifyVertexBuffer)(
		[&](FRHICommandListImmediate& RHICmdList)
		{
			auto RenderResource = SkeletalMesh->GetResourceForRendering();
			auto LOD = &RenderResource->LODRenderData[0];
			auto StaticVertexBuffers = &LOD->StaticVertexBuffers;
			auto StaticVertexBuffer = &StaticVertexBuffers->StaticMeshVertexBuffer;
			check(StaticVertexBuffer->GetAllowCPUAccess());
			auto TexCoordsVB = &StaticVertexBuffer->TexCoordVertexBuffer;
			void* TexCoordsBufferData = RHICmdList.LockBuffer(TexCoordsVB->VertexBufferRHI, 0, TexCoordsVB->VertexBufferRHI->GetSize(), RLM_WriteOnly);
			for (int32 SectionIndex = 0; SectionIndex < LOD->RenderSections.Num(); SectionIndex++)
			{
				auto Section_SM = &LOD->RenderSections[SectionIndex];
				const int32 NumSoftVertices = Section_SM->GetNumVertices();
				for (int32 VertexIndex = 0; VertexIndex < NumSoftVertices; VertexIndex++)
				{
					const int32 VBIndex = Section_SM->GetVertexBufferIndex() + VertexIndex;
					check((uint32)VBIndex < VertCount);
					uint32 uvChannel = 1;
					//FVector2f TexCoordValue = StaticVertexBuffer->GetVertexUV(VBIndex, uvChannel);
					int32 texCoordX = VBIndex % TexDim;
					int32 texCoordY = VBIndex / TexDim;
					float halfTextelW = 0.5f / TexDim;
					float halfTextelH = 0.5f / TexDim;
					FVector2f TexCoordValue = FVector2f(
						(float)texCoordX / TexDim + halfTextelW,
						(float)texCoordY / TexDim + halfTextelH);
					StaticVertexBuffer->SetVertexUV(VBIndex, uvChannel, TexCoordValue);

				}
			}
			FMemory::Memcpy(TexCoordsBufferData, StaticVertexBuffer->GetTexCoordData(), StaticVertexBuffer->GetTexCoordSize());
			RHICmdList.UnlockBuffer(TexCoordsVB->VertexBufferRHI);
		}
		);

	// Set the default cull info texture
	UserDataTexture_0 = UTexture2D::CreateTransient(TexDim, TexDim, PF_B8G8R8A8);
	check(UserDataTexture_0);
	//UserDataTexture_0->MipGenSettings = TMGS_NoMipmaps;
	UserDataTexture_0->SRGB = false;
	UserDataTexture_0->NeverStream = true; // Prevent streaming issues
	UserDataTexture_0->CompressionSettings = TC_Default;
	UserDataTexture_0->LODGroup = TEXTUREGROUP_Pixels2D;
	UserDataTexture_0->UpdateResource();
	TArray<FColor> PixelData;
	PixelData.SetNumUninitialized(TexDim * TexDim);
	for (int32 i = 0; i < TexDim * TexDim; ++i)
		PixelData[i] = FColor(255, 255, 255, 255);  // White color
	FTexture2DResource* TexResource = (FTexture2DResource*)UserDataTexture_0->GetResource();
	check(TexResource);
	UserDataTextureUpdateRegion = FUpdateTextureRegion2D(0, 0, 0, 0, TexDim, TexDim);
	UserDataTexture_0->UpdateTextureRegions(0, 1,
		&UserDataTextureUpdateRegion, TexDim * sizeof(FColor),
		sizeof(FColor),
		(uint8*)PixelData.GetData());

	// Morph target editing
	checkf(MorphShapingModel == nullptr, TEXT("MorphShapingModel should be null before initializing it."));
	MorphShapingModel = NewObject<UDDMorphShapingModel>(this);
	MorphShapingModel->Initialize(CEConfigData, SkeletalMesh);
	checkf(MorphShapingController == nullptr, TEXT("MorphShapingController should be null before initializing it."));
	MorphShapingController = NewObject<UDDMorphShapingController>(this);
	MorphShapingController->Initialize(CEConfigData);
	MorphShapingController->SetModel(MorphShapingModel);

	// Texture editing
	checkf(TextureEditingModel == nullptr, TEXT("TextureEditingModel should be null before initializing it."));
	TextureEditingModel = NewObject<UDDTextureEditingModel>(this);
	TextureEditingModel->Initialize(SkinCompositionEditorData, SkeletalMesh);
	checkf(TextureEditingController == nullptr, TEXT("TextureEditingController should be null before initializing it."));
	TextureEditingController = NewObject<UDDTextureEditingController>(this);
	TextureEditingController->Initialize(CEConfigData, SkinCompositionEditorData);
	TextureEditingController->SetModel(TextureEditingModel);

	// Iterate over all materials in the skeletal mesh and replace them with dynamic materials,
	// so that we can change its textures to render targets
	TArray<FSkeletalMaterial>& Materials_SM = SkeletalMesh->GetMaterials();
	for (size_t i = 0; i < Materials_SM.Num(); i++)
	{
		auto Material_SM = &Materials_SM[i];
		FString MaterialName_SM = Material_SM->MaterialSlotName.GetPlainNameString().ToLower();
		if (MaterialName_SM.StartsWith("m_"))
			MaterialName_SM = MaterialName_SM.RightChop(2);

		auto TextureBindingKey = FDDTextureEditingBindingKey(MaterialName_SM, EDDTextureEditingMaterialKey::MK_Albedo);
		if (TextureEditingModel->SupportsMesh(TextureBindingKey))
		{
			UMaterialInterface* OldMaterial = GetMaterial(i);
			UMaterialInstanceDynamic* DynamicMaterial = CreateAndSetMaterialInstanceDynamic(i);

			if (DynamicMaterial)
			{
				UTextureRenderTarget2D* RT = TextureEditingModel->GetMeshRenderTarget(TextureBindingKey);
				DynamicMaterial->SetTextureParameterValue("BC_Texture", RT);

				//if (AttachmentCullMaskNameToTexture.Contains(MaterialName_SM))
				{
					//UTexture2D* CullMask = AttachmentCullMaskNameToTexture[MaterialName_SM];
					DynamicMaterial->SetTextureParameterValue("CM_Texture", UserDataTexture_0);
				}
			}
		}
	}

	// Now that the textures have been replaced with render targets, we need
	// to draw to those render targets for the first time, otherwise they'll be black
	TextureEditingController->ResetToConfigValues();
}

void UDDEditableSkeletalMeshComponent::FinishEditing()
{
	if (State == EDDEditableSkeletalMeshComponentState::ESMCS_Initial)
	{
		// do nothing
	}
	else if (State == EDDEditableSkeletalMeshComponentState::ESMCS_Editing)
	{
		ApplyEdits();
	}
	State = EDDEditableSkeletalMeshComponentState::ESMCS_Finished;
}

void UDDEditableSkeletalMeshComponent::ApplyCullData(UDDAttachmentVertexCacheData* AtthVertexCacheData)
{
	AttachmentVertexCacheData = AtthVertexCacheData;
	bCullDataMarkedDirty = true;

	check(UserDataTexture_0);
	TArray<FColor> PixelData;
	int32 TexWidth = UserDataTexture_0->GetSizeX();
	int32 TexHeight = UserDataTexture_0->GetSizeY();
	check(TexWidth == TexHeight);
	int32 TexDim = TexWidth;
	PixelData.SetNumUninitialized(TexWidth * TexHeight);
	for (int32 i = 0; i < TexWidth * TexHeight; ++i)
		PixelData[i] = FColor(255, 255, 255, 255);  // White color



	auto RenderResource = SkeletalMesh->GetResourceForRendering();
	auto LOD = &RenderResource->LODRenderData[0];
	auto StaticVertexBuffers = &LOD->StaticVertexBuffers;
	auto StaticVertexBuffer = &StaticVertexBuffers->StaticMeshVertexBuffer;
	uint32 VertCount = StaticVertexBuffer->GetNumVertices();
	check(StaticVertexBuffer->GetAllowCPUAccess());
	for (int32 SectionIndex = 0; SectionIndex < LOD->RenderSections.Num(); SectionIndex++)
	{
		auto Section_SM = &LOD->RenderSections[SectionIndex];
		const int32 NumSoftVertices = Section_SM->GetNumVertices();
		for (int32 VertexIndex = 0; VertexIndex < NumSoftVertices; VertexIndex++)
		{
			const int32 VBIndex = Section_SM->GetVertexBufferIndex() + VertexIndex;
			check((uint32)VBIndex < VertCount);
			//uint32 uvChannel = 1;
			//FVector2f TexCoordValue = StaticVertexBuffer->GetVertexUV(VBIndex, uvChannel);
			int32 texCoordX = VBIndex % TexDim;
			int32 texCoordY = VBIndex / TexDim;
			//TexCoordValue = TexCoordValue + FVector2f(texCoordX, texCoordY);
			//StaticVertexBuffer->SetVertexUV(VBIndex, uvChannel, TexCoordValue);


			float vertOpacity = AttachmentVertexCacheData ? AttachmentVertexCacheData->VertexOpacity[VBIndex] : 1.0f;
			PixelData[VBIndex] = FColor(255 * vertOpacity, 255 * vertOpacity, 255 * vertOpacity, 255);
		}
	}



	//// Populate PixelData with random noise
	//for (int32 i = 0; i < TexWidth * TexHeight; ++i)
	//{
	//	// Generate random color components (Noise: RGB values between 0 and 255)
	//	uint8 R = FMath::RandRange(0, 255);
	//	uint8 G = FMath::RandRange(0, 255);
	//	uint8 B = FMath::RandRange(0, 255);
	//	uint8 A = 255; // Full alpha (opaque)

	//	// Create a color and assign it to PixelData
	//	PixelData[i] = FColor(R, G, B, A);
	//}



	// Ensure the resource is ready
	FTexture2DResource* TexResource = (FTexture2DResource*)UserDataTexture_0->GetResource();
	check(TexResource);

	// Update region
	UserDataTexture_0->UpdateTextureRegions(0, 1,
		&UserDataTextureUpdateRegion,
		TexWidth * sizeof(FColor),
		sizeof(FColor),
		(uint8*)PixelData.GetData());
}

void UDDEditableSkeletalMeshComponent::ApplyEdits()
{
	check(State == EDDEditableSkeletalMeshComponentState::ESMCS_Editing);
	if (TextureEditingModel && TextureEditingController && TextureEditingModel->IsMarkedDirty())
		TextureEditingController->ApplyChanges();

	if (MorphShapingModel && MorphShapingController && MorphShapingModel->IsMarkedDirty())
		MorphShapingController->ApplyChanges(this);

	// Unset the flag
	bCullDataMarkedDirty = false;

	// Finish
	MarkRenderStateDirty();
}
