#include "Controllers/DDTextureEditingController.h"
#include "RHICommandList.h"
#include "RenderGraph.h"
#include "RenderGraphUtils.h"
#include "Rendering/Texture2DResource.h"
#include "RenderUtils.h"
#include "GlobalShader.h"
#include "ShaderParameterStruct.h"
#include "ShaderParameterUtils.h"
#include "Engine/TextureRenderTarget2D.h"
#include "DataAssets/DDDataAssets.h"
#include "DataAssets/DDCharacterEditorSkinCompositionData.h"
#include "Models/DDTextureEditingModel.h"

#include "Controllers/ImageProcessor/DDImageTexture.h"
#include "Controllers/ImageProcessor/DDColorize.h"
#include "Controllers/ImageProcessor/DDBrightnessContrast.h"
#include "Controllers/ImageProcessor/DDDilation.h"
#include "Controllers/ImageProcessor/DDSwizzle.h"
#include "Controllers/ImageProcessor/DDBlendTexture.h"
#include "Controllers/ImageProcessor/DDMeshUVColor.h"
#include "Controllers/ImageProcessor/DDMeshUVTexture.h"


void CopyTexture2DToRenderTarget(UTexture2D* SourceTexture, UTextureRenderTarget2D* RenderTarget)
{
    if (!SourceTexture || !RenderTarget)
    {
        UE_LOG(LogTemp, Warning, TEXT("Invalid texture or render target."));
        return;
    }

    ENQUEUE_RENDER_COMMAND(CopyTextureCommand)(
        [SourceTexture, RenderTarget](FRHICommandListImmediate& RHICmdList)
        {
            FRHITexture2D* SourceRHI = SourceTexture->GetResource()->TextureRHI->GetTexture2D();
            FRHITexture2D* TargetRHI = RenderTarget->GetRenderTargetResource()->GetRenderTargetTexture()->GetTexture2D();

            if (SourceRHI && TargetRHI)
            {
                FRHICopyTextureInfo CopyInfo;
                RHICmdList.CopyTexture(SourceRHI, TargetRHI, CopyInfo);
            }
        });
}

bool FDDTextureEditingColorPickerData::SetTextureColors()
{
    if (!texture2D)
    {
        UE_LOG(LogTemp, Warning, TEXT("Invalid texture."));
        return false;
    }

    // Ensure the texture is fully loaded
    texture2D->UpdateResource();
    width = texture2D->GetSizeX();
    height = texture2D->GetSizeY();

    // Create a new render target
    UTextureRenderTarget2D* RenderTarget = NewObject<UTextureRenderTarget2D>();
    RenderTarget->RenderTargetFormat = RTF_RGBA8; // Match format (update as needed based on SourceTexture)
    RenderTarget->InitAutoFormat(width, height);
    RenderTarget->UpdateResourceImmediate(true);

    // Perform the copy
    CopyTexture2DToRenderTarget(texture2D, RenderTarget);

    // Wait for the render thread to finish the operation
    FRenderCommandFence Fence;
    Fence.BeginFence();
    Fence.Wait();

    // Read pixels into the array
    FTextureRenderTargetResource* RenderTargetResource = RenderTarget->GameThread_GetRenderTargetResource();
    if (RenderTargetResource->ReadPixels(textureColors))
    {
        return true;
    }
    else
    {
        UE_LOG(LogTemp, Error, TEXT("Failed to read pixels from the RenderTarget!"));
        return false;
    }
}

FColor FDDTextureEditingColorPickerData::SampleColor(FVector2f TexCoord) const
{
    int32 x = (TexCoord.X * 0.5 + 0.5) * (width - 1);
    int32 y = (TexCoord.Y * 0.5 + 0.5) * (height - 1);
    return textureColors[x + y * width];
}


void UDDTextureEditingController::Initialize(UDDCharacterEditorConfigData* CEConfig, UDDCharacterEditorSkinCompositionData* Config)
{
    CEConfigData = Config;
    HandlerToColorPickerIndex.Reset();
    ColorPickers.Reset();

    TMap<FString, FString> MeshNameToSegmentation;
    TMap<FString, TMap<FString, int32>> SegNameToColorPickerNameToColorPickerDataIndex;
    for (auto& Segmentation : CEConfig->segmentations)
    {
        for (auto& MeshName : Segmentation.targetSkinCompositionMeshes)
        {
            MeshNameToSegmentation.Add(MeshName, Segmentation.name);
        }
        SegNameToColorPickerNameToColorPickerDataIndex.Add(Segmentation.name, TMap<FString, int32>());
    }

    for (int32 ComponentIndex = 0; ComponentIndex < Config->components.Num(); ++ComponentIndex)
    {
        auto& Component = Config->components[ComponentIndex];
        FString meshName = Component.meshes[0];
        if (!MeshNameToSegmentation.Contains(meshName))
            continue;
        auto& SegName = MeshNameToSegmentation[meshName];
        auto& ColorPickerNameToColorPickerDataIndex = SegNameToColorPickerNameToColorPickerDataIndex[SegName];
        for (int32 i = 0; i < Component.handlers.Num(); i++)
        {
            auto Handler = &Component.handlers[i];
            if (Handler->type.Compare("PixelPickerElement") == 0)
            {
                int32 ColorPickerIndex = -1;
                if (!ColorPickerNameToColorPickerDataIndex.Contains(Handler->displayName))
                {
                    check(Handler->textureParams.Num() == 1);
                    ColorPickers.Add(FDDTextureEditingColorPickerData());
                    auto ColorPicker = &ColorPickers.Last();
                    ColorPicker->texture2D = Handler->textureParams[0];
                    ColorPicker->SetTextureColors();

                    ColorPicker->components = { ComponentIndex };
                    ColorPicker->stacks = { Handler->stack };
                    ColorPicker->processes = { Handler->process };

                    ColorPickerIndex = ColorPickers.Num() - 1;
                    ColorPickerNameToColorPickerDataIndex.Add(Handler->displayName, ColorPickerIndex);
                }
                else // already in
                {
                    ColorPickerIndex = ColorPickerNameToColorPickerDataIndex[Handler->displayName];
                    auto ColorPicker = &ColorPickers[ColorPickerIndex];
                    ColorPicker->components.Add(ComponentIndex);
                    ColorPicker->stacks.Add(Handler->stack);
                    ColorPicker->processes.Add(Handler->process);
                }
                check(ColorPickerIndex != -1);
                HandlerToColorPickerIndex.Add(Handler, ColorPickerIndex);
            }
            else
            {
                // Not supporting other types yet
                throw;
            }
        }
    }
}

void UDDTextureEditingController::SetModel(UDDTextureEditingModel* Model)
{
    ModelToControl = Model;
}

void UDDTextureEditingController::ResetToConfigValues()
{
    ModelToControl->MarkedDirty = true;
    for (auto& Component : CEConfigData->components)
    {
        for (int32 HndlerIndex = 0; HndlerIndex < Component.handlers.Num(); HndlerIndex++)
        {
            auto Handler = &Component.handlers[HndlerIndex];
            if (Handler->type.Compare("PixelPickerElement") == 0)
            {
                int32 ColorPickerIndex = HandlerToColorPickerIndex[Handler];
                auto ColorPicker = &ColorPickers[ColorPickerIndex];
                FColor sampledColor = ColorPicker->SampleColor(FVector2f(0.0f, 0.0f));
                if (Handler->floatParams.Num() == 2)
                    sampledColor = ColorPicker->SampleColor(FVector2f(Handler->floatParams[0], Handler->floatParams[1]));

                for (size_t i = 0; i < ColorPicker->components.Num(); i++)
                {
                    auto ImageProcessorComponent = &ModelToControl->ImageProcessorComponents[ColorPicker->components[i]];
                    auto Stack = &ImageProcessorComponent->ImageProcessorStacks[ColorPicker->stacks[i]];
                    auto Process = Stack->InterfaceProcessors[ColorPicker->processes[i]];

                    auto imgProcessor = Process;
                    auto node = Cast<UDDColorize>(imgProcessor);
                    node->SetColor(sampledColor);
                }
            }
        }
    }
}

void UDDTextureEditingController::UpdateModel(
    const FDDCharacterEditorSegmentation* Segmentation,
    const FDDCharacterEditorSkinCompositionHandler* SkinCompositorHandler,
    FVector2f value)
{
    ModelToControl->MarkedDirty = true;
    if (SkinCompositorHandler->type.Compare("PixelPickerElement") == 0)
    {
        int32 ColorPickerIndex = HandlerToColorPickerIndex[SkinCompositorHandler];
        auto ColorPicker = &ColorPickers[ColorPickerIndex];
        FColor sampledColor = ColorPicker->SampleColor(value);
        check(ColorPicker->components.Num() == ColorPicker->stacks.Num());
        check(ColorPicker->components.Num() == ColorPicker->processes.Num());
        for (size_t i = 0; i < ColorPicker->components.Num(); i++)
        {
            auto ImageProcessorComponent = &ModelToControl->ImageProcessorComponents[ColorPicker->components[i]];
            auto Stack = &ImageProcessorComponent->ImageProcessorStacks[ColorPicker->stacks[i]];
            auto Process = Stack->InterfaceProcessors[ColorPicker->processes[i]];

            auto imgProcessor = Process;
            auto node = Cast<UDDColorize>(imgProcessor);
            node->SetColor(sampledColor);
        }
    }
}

void UDDTextureEditingController::ApplyChanges()
{
    ModelToControl->MarkedDirty = false;
    ProcessImageGameThread(ModelToControl->ResultNodes);
}

void UDDTextureEditingController::ProcessImageGameThread(UDDImageProcessor* ResultImageProcessor)
{
    TSet<UDDImageProcessor*> ResultSet;
    ResultSet.Add(ResultImageProcessor);
    ProcessImageGameThread(ResultSet);
}

void UDDTextureEditingController::ProcessImageGameThread(TSet<UDDImageProcessor*> ResultImageProcessors)
{
    // Prepare first
    DDImageProcessorExecuteContext_GameThread contextGT;
    contextGT.Depth = 0;
    for (UDDImageProcessor* Processor : ResultImageProcessors)
        Processor->Prepare();

    // Now exectute the CPU part
    for (UDDImageProcessor* Processor : ResultImageProcessors)
        Processor->Execute(contextGT);

    // Then add to graph and execute in the render thread
    ENQUEUE_RENDER_COMMAND(SceneDrawCompletion)(
        [ResultImageProcessors](FRHICommandListImmediate& RHICmdList)
        {
            FRDGBuilder GraphBuilder(RHICmdList);
            DDImageProcessorExecuteContext_RenderThread contextRT;
            contextRT.GraphBuilder = &GraphBuilder;
            contextRT.Depth = 0;
            for (UDDImageProcessor* Processor : ResultImageProcessors)
                Processor->Execute(contextRT);
            GraphBuilder.Execute();
        });
}
