#include "Controllers/ImageProcessor/DDBlendTexture.h"
#include "RHICommandList.h"
#include "RenderGraph.h"
#include "RenderGraphUtils.h"
#include "RenderUtils.h"
#include "GlobalShader.h"
#include "Engine/TextureRenderTarget2D.h"
#include "RHIDefinitions.h"
#include "RenderResource.h"
#include "RHI.h"
#include "RHIStaticStates.h"
#include "PipelineStateCache.h"
#include "TextureResource.h"


DECLARE_STATS_GROUP(TEXT("Popul8BlendTextureCS"), STATGROUP_Popul8BlendTextureCS, STATCAT_Advanced);
DECLARE_CYCLE_STAT(TEXT("Popul8BlendTextureCS Execute"), STAT_Popul8BlendTextureCS_Execute, STATGROUP_Popul8BlendTextureCS);

#define NUM_THREADS_X 8
#define NUM_THREADS_Y 8
#define NUM_THREADS_Z 1

class FDDBlendTextureCS : public FGlobalShader
{
    DECLARE_GLOBAL_SHADER(FDDBlendTextureCS)
    SHADER_USE_PARAMETER_STRUCT(FDDBlendTextureCS, FGlobalShader)

public:
    BEGIN_SHADER_PARAMETER_STRUCT(FParameters, )
        SHADER_PARAMETER_RDG_TEXTURE_SRV(Texture2D, InputTexture_A)
        SHADER_PARAMETER_RDG_TEXTURE_SRV(Texture2D, InputTexture_B)
        SHADER_PARAMETER_RDG_TEXTURE_SRV(Texture2D, InputTexture_Mask)
        SHADER_PARAMETER_RDG_TEXTURE_UAV(RWTexture2D<float4>, OutputTexture)
        SHADER_PARAMETER_SAMPLER(SamplerState, InputTextureSampler)
        SHADER_PARAMETER(float, Weight)
        SHADER_PARAMETER(int32, Operation)
    END_SHADER_PARAMETER_STRUCT()

public:
    static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
    {
        const FPermutationDomain PermutationVector(Parameters.PermutationId);
        return true;
    }

    static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
    {
        FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
        const FPermutationDomain PermutationVector(Parameters.PermutationId);

        OutEnvironment.SetDefine(TEXT("THREADS_X"), NUM_THREADS_X);
        OutEnvironment.SetDefine(TEXT("THREADS_Y"), NUM_THREADS_Y);
        OutEnvironment.SetDefine(TEXT("THREADS_Z"), NUM_THREADS_Z);
    }
};

IMPLEMENT_GLOBAL_SHADER(FDDBlendTextureCS, "/Plugin/Popul8/DDBlendTexture.usf", "MainCS", SF_Compute);


void UDDBlendTexture::Execute(DDImageProcessorExecuteContext_GameThread& context)
{
    FScopeLock Lock(&DataMutex);
    if (CurrentOp == BTO_Lerp)
    {
        check(SlotToDependency.Num() == 3);
        check(SlotToDependency[BTDS_InputTexture_A] && SlotToDependency[BTDS_InputTexture_B] && SlotToDependency[BTDS_MaskTexture]);
        ExecuteDependencies(context);

        auto inputTexture_0 = SlotToDependency[0]->GetResultRenderTarget();
        auto inputTexture_1 = SlotToDependency[1]->GetResultRenderTarget();
        auto inputTexture_2 = SlotToDependency[2]->GetResultRenderTarget();
        check(inputTexture_0->SizeX == inputTexture_1->SizeX);
        check(inputTexture_0->SizeX == inputTexture_2->SizeX);
        check(inputTexture_0->SizeY == inputTexture_1->SizeY);
        check(inputTexture_0->SizeY == inputTexture_2->SizeY);
    }
    else // only needs two inputs and no mask
    {
        check(SlotToDependency.Num() == 2);
        check(SlotToDependency[BTDS_InputTexture_A] && SlotToDependency[BTDS_InputTexture_B]);
        ExecuteDependencies(context);
        auto inputTexture_0 = SlotToDependency[0]->GetResultRenderTarget();
        auto inputTexture_1 = SlotToDependency[1]->GetResultRenderTarget();
    }
}

void UDDBlendTexture::Execute(DDImageProcessorExecuteContext_RenderThread& context)
{
    ensure(IsInRenderingThread());
    ExecuteDependencies(context);

    SCOPE_CYCLE_COUNTER(STAT_Popul8BlendTextureCS_Execute);
    DECLARE_GPU_STAT(Popul8BlendTextureCS)
    RDG_EVENT_SCOPE(*context.GraphBuilder, "Popul8BlendTextureCS");
    RDG_GPU_STAT_SCOPE(*context.GraphBuilder, Popul8BlendTextureCS);

    FScopeLock Lock(&DataMutex);
    typename FDDBlendTextureCS::FPermutationDomain PermutationVector;
    TShaderMapRef<FDDBlendTextureCS> ComputeShader(GetGlobalShaderMap(GMaxRHIFeatureLevel), PermutationVector);
    bool bIsShaderValid = ComputeShader.IsValid();
    check(bIsShaderValid);
    auto PassParameters = context.GraphBuilder->AllocParameters<FDDBlendTextureCS::FParameters>();

    // Output
    FRDGTextureDesc TempTextureDesc = FRDGTextureDesc::Create2D(
        FIntPoint(ResultRenderTarget->SizeX, ResultRenderTarget->SizeY),
        PF_B8G8R8A8,
        FClearValueBinding::White,
        TexCreate_RenderTargetable | TexCreate_ShaderResource | TexCreate_UAV,
        /* NumMips */ 1u);
    FRDGTextureRef TmpTexture = context.GraphBuilder->CreateTexture(TempTextureDesc, TEXT("Popul8BlendTextureCS.TempTexture"));
    FTextureRenderTargetResource* ResultTextureResource = ResultRenderTarget->GetRenderTargetResource();
    FRDGTextureRef ResultTextureRef = RegisterExternalTexture(*context.GraphBuilder, ResultTextureResource->GetRenderTargetTexture(), TEXT("Popul8DilationCS.RT"));
    PassParameters->OutputTexture = context.GraphBuilder->CreateUAV(TmpTexture);

    // InputTexture
    PassParameters->InputTexture_A = InputDependencyAsSRV(context, BTDS_InputTexture_A);
    PassParameters->InputTexture_B = InputDependencyAsSRV(context, BTDS_InputTexture_B);

    if (CurrentOp == BTO_Lerp)
    {
        PassParameters->InputTexture_Mask = InputDependencyAsSRV(context, BTDS_MaskTexture);
    }
    else
    {
        PassParameters->InputTexture_Mask = InputDependencyAsSRV(context, 0); // JUST PASS ANY, FIX LATER
    }

    PassParameters->InputTextureSampler = TStaticSamplerState<SF_Bilinear, AM_Clamp, AM_Clamp, AM_Clamp>::GetRHI();
    PassParameters->Weight = CurrentWeight;
    PassParameters->Operation = CurrentOp;

    // Add Pass
    auto GroupCount = FComputeShaderUtils::GetGroupCount(
        FIntVector(ResultRenderTarget->SizeX, ResultRenderTarget->SizeY, 1),
        FIntVector(NUM_THREADS_X, NUM_THREADS_Y, NUM_THREADS_Z));
    context.GraphBuilder->AddPass(
        RDG_EVENT_NAME("ExecutePopul8BlendTextureCS"),
        PassParameters,
        ERDGPassFlags::AsyncCompute,
        [PassParameters, ComputeShader, GroupCount](FRHIComputeCommandList& RHICmdList)
        {
            FComputeShaderUtils::Dispatch(RHICmdList, ComputeShader, *PassParameters, GroupCount);
        });

    // Copy to output render target
    check(ResultTextureRef->Desc.Format == PF_B8G8R8A8);
    AddCopyTexturePass(*context.GraphBuilder, TmpTexture, ResultTextureRef, FRHICopyTextureInfo());
}

void UDDBlendTexture::SetWeight(float Weight)
{
    FScopeLock Lock(&DataMutex);
    CurrentWeight = Weight;
}

void UDDBlendTexture::SetOperation(EDDBlendTextureOperation Operation)
{
    CurrentOp = Operation;
}
