#include "Controllers/ImageProcessor/DDBrightnessContrast.h"
#include "RHICommandList.h"
#include "RHIStaticStates.h"
#include "Engine/Texture2D.h"
#include "Engine/TextureRenderTarget2D.h"
#include "RenderGraph.h"
#include "RenderGraphUtils.h"
#include "RenderUtils.h"
#include "GlobalShader.h"
#include "TextureResource.h"


DECLARE_STATS_GROUP(TEXT("Popul8BrightnessContrastCS"), STATGROUP_Popul8BrightnessContrastCS, STATCAT_Advanced);
DECLARE_CYCLE_STAT(TEXT("Popul8BrightnessContrastCS Execute"), STAT_Popul8BrightnessContrastCS_Execute, STATGROUP_Popul8BrightnessContrastCS);

#define NUM_THREADS_X 8
#define NUM_THREADS_Y 8
#define NUM_THREADS_Z 1

class FDDBrightnessContrastCS : public FGlobalShader
{
    DECLARE_GLOBAL_SHADER(FDDBrightnessContrastCS)
    SHADER_USE_PARAMETER_STRUCT(FDDBrightnessContrastCS, FGlobalShader)

public:
    BEGIN_SHADER_PARAMETER_STRUCT(FParameters, )
        SHADER_PARAMETER_RDG_TEXTURE_SRV(Texture2D, InputTexture)
        SHADER_PARAMETER_RDG_TEXTURE_UAV(RWTexture2D<float4>, OutputTexture)
        SHADER_PARAMETER(float, BrightnessValue)
        SHADER_PARAMETER(float, ContrastValue)
    END_SHADER_PARAMETER_STRUCT()

public:
    static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
    {
        const FPermutationDomain PermutationVector(Parameters.PermutationId);
        return true;
    }

    static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
    {
        FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
        const FPermutationDomain PermutationVector(Parameters.PermutationId);

        OutEnvironment.SetDefine(TEXT("THREADS_X"), NUM_THREADS_X);
        OutEnvironment.SetDefine(TEXT("THREADS_Y"), NUM_THREADS_Y);
        OutEnvironment.SetDefine(TEXT("THREADS_Z"), NUM_THREADS_Z);
    }
};

IMPLEMENT_GLOBAL_SHADER(FDDBrightnessContrastCS, "/Plugin/Popul8/DDBrightnessContrast.usf", "MainCS", SF_Compute);


void UDDBrightnessContrast::Execute(DDImageProcessorExecuteContext_GameThread& context)
{
    FScopeLock Lock(&DataMutex);
    check(SlotToDependency.Num() == 1);
    check(SlotToDependency[0]);
    ExecuteDependencies(context);
}

void UDDBrightnessContrast::Execute(DDImageProcessorExecuteContext_RenderThread& context)
{
    ensure(IsInRenderingThread());
    ExecuteDependencies(context);

    SCOPE_CYCLE_COUNTER(STAT_Popul8BrightnessContrastCS_Execute);
    DECLARE_GPU_STAT(Popul8BrightnessContrastCS)
    RDG_EVENT_SCOPE(*context.GraphBuilder, "Popul8BrightnessContrastCS");
    RDG_GPU_STAT_SCOPE(*context.GraphBuilder, Popul8BrightnessContrastCS);

    FScopeLock Lock(&DataMutex);
    typename FDDBrightnessContrastCS::FPermutationDomain PermutationVector;
    TShaderMapRef<FDDBrightnessContrastCS> ComputeShader(GetGlobalShaderMap(GMaxRHIFeatureLevel), PermutationVector);
    bool bIsShaderValid = ComputeShader.IsValid();
    check(bIsShaderValid);
    auto PassParameters = context.GraphBuilder->AllocParameters<FDDBrightnessContrastCS::FParameters>();

    // Output
    FRDGTextureDesc TempTextureDesc = FRDGTextureDesc::Create2D(
        FIntPoint(ResultRenderTarget->SizeX, ResultRenderTarget->SizeY),
        PF_B8G8R8A8,
        FClearValueBinding::White,
        TexCreate_RenderTargetable | TexCreate_ShaderResource | TexCreate_UAV,
        /* NumMips */ 1u);
    FRDGTextureRef TmpTexture = context.GraphBuilder->CreateTexture(TempTextureDesc, TEXT("Popul8BrightnessContrastCS.TempTexture"));
    FTextureRenderTargetResource* ResultTextureResource = ResultRenderTarget->GetRenderTargetResource();
    FRDGTextureRef ResultTextureRef = RegisterExternalTexture(*context.GraphBuilder, ResultTextureResource->GetRenderTargetTexture(), TEXT("Popul8DilationCS.RT"));
    PassParameters->OutputTexture = context.GraphBuilder->CreateUAV(TmpTexture);

    // Input
    PassParameters->InputTexture = InputDependencyAsSRV(context, 0);
    PassParameters->BrightnessValue = CurrentBrightness;
    PassParameters->ContrastValue = CurrentContrast;

    // Add Pass
    auto GroupCount = FComputeShaderUtils::GetGroupCount(
        FIntVector(ResultRenderTarget->SizeX, ResultRenderTarget->SizeY, 1),
        FIntVector(NUM_THREADS_X, NUM_THREADS_Y, NUM_THREADS_Z));
    context.GraphBuilder->AddPass(
        RDG_EVENT_NAME("ExecutePopul8BrightnessContrastCS"),
        PassParameters,
        ERDGPassFlags::AsyncCompute,
        [PassParameters, ComputeShader, GroupCount](FRHIComputeCommandList& RHICmdList)
        {
            FComputeShaderUtils::Dispatch(RHICmdList, ComputeShader, *PassParameters, GroupCount);
        });

    // Copy to output render target
    check(ResultTextureRef->Desc.Format == PF_B8G8R8A8);
    AddCopyTexturePass(*context.GraphBuilder, TmpTexture, ResultTextureRef, FRHICopyTextureInfo());
}

void UDDBrightnessContrast::SetBrightness(const float& Brightness)
{
    FScopeLock Lock(&DataMutex);
    CurrentBrightness = Brightness;
}

void UDDBrightnessContrast::SetContrast(const float& Contrast)
{
    FScopeLock Lock(&DataMutex);
    CurrentContrast = Contrast;
}
