#include "Controllers/ImageProcessor/DDColorize.h"
#include "RHICommandList.h"
#include "RHIStaticStates.h"
#include "RenderGraph.h"
#include "RenderGraphUtils.h"
#include "RenderUtils.h"
#include "GlobalShader.h"
#include "Engine/TextureRenderTarget2D.h"
#include "TextureResource.h"


DECLARE_STATS_GROUP(TEXT("Popul8ColorizeCS"), STATGROUP_Popul8ColorizeCS, STATCAT_Advanced);
DECLARE_CYCLE_STAT(TEXT("Popul8ColorizeCS Execute"), STAT_Popul8ColorizeCS_Execute, STATGROUP_Popul8ColorizeCS);

#define NUM_THREADS_X 8
#define NUM_THREADS_Y 8
#define NUM_THREADS_Z 1

class FDDColorizeCS : public FGlobalShader
{
    DECLARE_GLOBAL_SHADER(FDDColorizeCS)
    SHADER_USE_PARAMETER_STRUCT(FDDColorizeCS, FGlobalShader)

public:
    BEGIN_SHADER_PARAMETER_STRUCT(FParameters, )
        SHADER_PARAMETER_RDG_TEXTURE_SRV(Texture2D, InputTexture)
        SHADER_PARAMETER_RDG_TEXTURE_UAV(RWTexture2D<float4>, OutputTexture)
        SHADER_PARAMETER(FVector4f, Color)
    END_SHADER_PARAMETER_STRUCT()

public:
    static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
    {
        const FPermutationDomain PermutationVector(Parameters.PermutationId);
        return true;
    }

    static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
    {
        FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
        const FPermutationDomain PermutationVector(Parameters.PermutationId);

        OutEnvironment.SetDefine(TEXT("THREADS_X"), NUM_THREADS_X);
        OutEnvironment.SetDefine(TEXT("THREADS_Y"), NUM_THREADS_Y);
        OutEnvironment.SetDefine(TEXT("THREADS_Z"), NUM_THREADS_Z);
    }
};

IMPLEMENT_GLOBAL_SHADER(FDDColorizeCS, "/Plugin/Popul8/DDColorize.usf", "MainCS", SF_Compute);


void UDDColorize::Execute(DDImageProcessorExecuteContext_GameThread& context)
{
    FScopeLock Lock(&DataMutex);
    check(SlotToDependency.Num() == 1);
    check(SlotToDependency[0]);
    ExecuteDependencies(context);
}

void UDDColorize::Execute(DDImageProcessorExecuteContext_RenderThread& context)
{
    ensure(IsInRenderingThread());
    ExecuteDependencies(context);

    SCOPE_CYCLE_COUNTER(STAT_Popul8ColorizeCS_Execute);
    DECLARE_GPU_STAT(Popul8ColorizeCS)
    RDG_EVENT_SCOPE(*context.GraphBuilder, "Popul8ColorizeCS");
    RDG_GPU_STAT_SCOPE(*context.GraphBuilder, Popul8ColorizeCS);

    FScopeLock Lock(&DataMutex);
    typename FDDColorizeCS::FPermutationDomain PermutationVector;
    TShaderMapRef<FDDColorizeCS> ComputeShader(GetGlobalShaderMap(GMaxRHIFeatureLevel), PermutationVector);
    bool bIsShaderValid = ComputeShader.IsValid();
    check(bIsShaderValid);
    auto PassParameters = context.GraphBuilder->AllocParameters<FDDColorizeCS::FParameters>();

    // Output
    FRDGTextureDesc TempTextureDesc = FRDGTextureDesc::Create2D(
        FIntPoint(ResultRenderTarget->SizeX, ResultRenderTarget->SizeY),
        PF_B8G8R8A8,
        FClearValueBinding::White,
        TexCreate_RenderTargetable | TexCreate_ShaderResource | TexCreate_UAV,
        /* NumMips */ 1u);
    FRDGTextureRef TmpTexture = context.GraphBuilder->CreateTexture(TempTextureDesc, TEXT("Popul8ColorizeCS.TempTexture"));
    FTextureRenderTargetResource* ResultTextureResource = ResultRenderTarget->GetRenderTargetResource();
    FRDGTextureRef ResultTextureRef = RegisterExternalTexture(*context.GraphBuilder, ResultTextureResource->GetRenderTargetTexture(), TEXT("Popul8ColorizeCS.RT"));
    PassParameters->OutputTexture = context.GraphBuilder->CreateUAV(TmpTexture);

    // Input
    PassParameters->InputTexture = InputDependencyAsSRV(context, 0);
    PassParameters->Color = FVector4f(CurrentColor);

    // Add Pass
    auto GroupCount = FComputeShaderUtils::GetGroupCount(
        FIntVector(ResultRenderTarget->SizeX, ResultRenderTarget->SizeY, 1),
        FIntVector(NUM_THREADS_X, NUM_THREADS_Y, NUM_THREADS_Z));
    context.GraphBuilder->AddPass(
        RDG_EVENT_NAME("ExecutePopul8ColorizeCS"),
        PassParameters,
        ERDGPassFlags::AsyncCompute,
        [PassParameters, ComputeShader, GroupCount](FRHIComputeCommandList& RHICmdList)
        {
            FComputeShaderUtils::Dispatch(RHICmdList, ComputeShader, *PassParameters, GroupCount);
        });

    // Copy to output render target
    check(ResultTextureRef->Desc.Format == PF_B8G8R8A8);
    AddCopyTexturePass(*context.GraphBuilder, TmpTexture, ResultTextureRef, FRHICopyTextureInfo());
}

void UDDColorize::SetColor(const FColor& Color)
{
    FScopeLock Lock(&DataMutex);
    CurrentColor = Color;
}

const FColor& UDDColorize::GetColor() const
{
    return CurrentColor;
}
