#include "Controllers/ImageProcessor/DDDilation.h"
#include "RHICommandList.h"
#include "Engine/Texture2D.h"
#include "Engine/TextureRenderTarget2D.h"
#include "RenderGraph.h"
#include "RenderGraphUtils.h"
#include "RenderUtils.h"
#include "GlobalShader.h"
#include "TextureResource.h"


DECLARE_STATS_GROUP(TEXT("Popul8DilationCS"), STATGROUP_Popul8DilationCS, STATCAT_Advanced);
DECLARE_CYCLE_STAT(TEXT("Popul8DilationCS Execute"), STAT_Popul8DilationCS_Execute, STATGROUP_Popul8DilationCS);

#define NUM_THREADS_X 8
#define NUM_THREADS_Y 8
#define NUM_THREADS_Z 1

class FDDDilationCS : public FGlobalShader
{
    DECLARE_GLOBAL_SHADER(FDDDilationCS)
    SHADER_USE_PARAMETER_STRUCT(FDDDilationCS, FGlobalShader)

public:
    BEGIN_SHADER_PARAMETER_STRUCT(FParameters, )
        SHADER_PARAMETER_RDG_TEXTURE_SRV(Texture2D, InputTexture)
        SHADER_PARAMETER_RDG_TEXTURE_UAV(RWTexture2D<float4>, OutputTexture)
        SHADER_PARAMETER_RDG_BUFFER_SRV(StructuredBuffer<int2>, CellBuffer)
        SHADER_PARAMETER(int, Count)
    END_SHADER_PARAMETER_STRUCT()

public:
    static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
    {
        const FPermutationDomain PermutationVector(Parameters.PermutationId);
        return true;
    }

    static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
    {
        FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
        const FPermutationDomain PermutationVector(Parameters.PermutationId);

        OutEnvironment.SetDefine(TEXT("THREADS_X"), NUM_THREADS_X);
        OutEnvironment.SetDefine(TEXT("THREADS_Y"), NUM_THREADS_Y);
        OutEnvironment.SetDefine(TEXT("THREADS_Z"), NUM_THREADS_Z);
    }
};

IMPLEMENT_GLOBAL_SHADER(FDDDilationCS, "/Plugin/Popul8/DDDilation.usf", "MainCS", SF_Compute);


void UDDDilation::Execute(DDImageProcessorExecuteContext_GameThread& context)
{
    FScopeLock Lock(&DataMutex);
    check(SlotToDependency.Num() == 1);
    check(SlotToDependency[0]);
    ExecuteDependencies(context);
}

void UDDDilation::Execute(DDImageProcessorExecuteContext_RenderThread& context)
{
    ensure(IsInRenderingThread());
    ExecuteDependencies(context);

    SCOPE_CYCLE_COUNTER(STAT_Popul8DilationCS_Execute);
    DECLARE_GPU_STAT(Popul8DilationCS)
    RDG_EVENT_SCOPE(*context.GraphBuilder, "Popul8DilationCS");
    RDG_GPU_STAT_SCOPE(*context.GraphBuilder, Popul8DilationCS);

    FScopeLock Lock(&DataMutex);
    typename FDDDilationCS::FPermutationDomain PermutationVector;
    TShaderMapRef<FDDDilationCS> ComputeShader(GetGlobalShaderMap(GMaxRHIFeatureLevel), PermutationVector);
    bool bIsShaderValid = ComputeShader.IsValid();
    check(bIsShaderValid);
    auto PassParameters = context.GraphBuilder->AllocParameters<FDDDilationCS::FParameters>();

    // Output
    FRDGTextureDesc TempTextureDesc = FRDGTextureDesc::Create2D(
        FIntPoint(ResultRenderTarget->SizeX, ResultRenderTarget->SizeY),
        PF_B8G8R8A8,
        FClearValueBinding::White,
        TexCreate_RenderTargetable | TexCreate_ShaderResource | TexCreate_UAV,
        /* NumMips */ 1u);
    FRDGTextureRef TmpTexture = context.GraphBuilder->CreateTexture(TempTextureDesc, TEXT("Popul8DilationCS.TempTexture"));
    FTextureRenderTargetResource* ResultTextureResource = ResultRenderTarget->GetRenderTargetResource();
    FRDGTextureRef ResultTextureRef = RegisterExternalTexture(*context.GraphBuilder, ResultTextureResource->GetRenderTargetTexture(), TEXT("Popul8DilationCS.RT"));
    PassParameters->OutputTexture = context.GraphBuilder->CreateUAV(TmpTexture);

    // Input
    PassParameters->InputTexture = InputDependencyAsSRV(context, 0);

    // Create an RDG buffer from data
    TArray<FIntPoint> cellBufferData;
    int32 circleCount = CurrentRadius;
    for (int32 i = -circleCount; i <= circleCount; i++)
    {
        for (int32 j = -circleCount; j <= circleCount; j++)
        {
            float dist = FVector2f(FIntPoint(i, j)).Length();
            if (i != 0 && j != 0 && dist < circleCount)
                cellBufferData.Add(FIntPoint(i, j));
        }
    }
    cellBufferData.Sort([](const FIntPoint& A, const FIntPoint& B) {
        float distA = FVector2f(A).Length();
        float distB = FVector2f(B).Length();
        return distA < distB;
        });
    FRDGBufferRef CellBufferRDG = CreateStructuredBuffer(
        *context.GraphBuilder,
        TEXT("Popul8DilationCS.CellBuffer"),
        sizeof(FIntPoint),
        cellBufferData.Num(),
        cellBufferData.GetData(),
        cellBufferData.Num() * sizeof(FIntPoint)
    );

    // Assign the buffer's SRV to the pass parameters
    PassParameters->CellBuffer = context.GraphBuilder->CreateSRV(CellBufferRDG);
    PassParameters->Count = cellBufferData.Num();

    // Add Pass
    auto GroupCount = FComputeShaderUtils::GetGroupCount(
        FIntVector(ResultRenderTarget->SizeX, ResultRenderTarget->SizeY, 1),
        FIntVector(NUM_THREADS_X, NUM_THREADS_Y, NUM_THREADS_Z));
    context.GraphBuilder->AddPass(
        RDG_EVENT_NAME("ExecutePopul8DilationCS"),
        PassParameters,
        ERDGPassFlags::AsyncCompute,
        [PassParameters, ComputeShader, GroupCount](FRHIComputeCommandList& RHICmdList)
        {
            FComputeShaderUtils::Dispatch(RHICmdList, ComputeShader, *PassParameters, GroupCount);
        });

    // Copy to output render target
    check(ResultTextureRef->Desc.Format == PF_B8G8R8A8);
    AddCopyTexturePass(*context.GraphBuilder, TmpTexture, ResultTextureRef, FRHICopyTextureInfo());
}

void UDDDilation::SetRadius(int32 Brightness)
{
    FScopeLock Lock(&DataMutex);
    CurrentRadius = Brightness;
}
