#include "Controllers/ImageProcessor/DDImageTexture.h"
#include "RHICommandList.h"
#include "RHIStaticStates.h"
#include "Engine/Texture2D.h"
#include "Engine/TextureRenderTarget2D.h"
#include "RenderGraph.h"
#include "RenderGraphUtils.h"
#include "RenderUtils.h"
#include "GlobalShader.h"
#include "TextureResource.h"


DECLARE_STATS_GROUP(TEXT("Popul8ImageTextureCS"), STATGROUP_Popul8ImageTextureCS, STATCAT_Advanced);
DECLARE_CYCLE_STAT(TEXT("Popul8ImageTextureCS Execute"), STAT_Popul8ImageTextureCS_Execute, STATGROUP_Popul8ImageTextureCS);

#define NUM_THREADS_X 8
#define NUM_THREADS_Y 8
#define NUM_THREADS_Z 1

class FDDImageTextureCS : public FGlobalShader
{
    DECLARE_GLOBAL_SHADER(FDDImageTextureCS)
    SHADER_USE_PARAMETER_STRUCT(FDDImageTextureCS, FGlobalShader)

public:
    BEGIN_SHADER_PARAMETER_STRUCT(FParameters, )
        SHADER_PARAMETER_RDG_TEXTURE_SRV(Texture2D, InputTexture)
        SHADER_PARAMETER_RDG_TEXTURE_UAV(RWTexture2D<float4>, OutputTexture)
        SHADER_PARAMETER_SAMPLER(SamplerState, InputTextureSampler)
        SHADER_PARAMETER(FVector2f, Scale)
        SHADER_PARAMETER(FVector2f, Offset)
    END_SHADER_PARAMETER_STRUCT()

public:
    static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
    {
        return true;
    }

    static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
    {
        FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
        const FPermutationDomain PermutationVector(Parameters.PermutationId);
        OutEnvironment.SetDefine(TEXT("THREADS_X"), NUM_THREADS_X);
        OutEnvironment.SetDefine(TEXT("THREADS_Y"), NUM_THREADS_Y);
        OutEnvironment.SetDefine(TEXT("THREADS_Z"), NUM_THREADS_Z);
    }
};

IMPLEMENT_GLOBAL_SHADER(FDDImageTextureCS, "/Plugin/Popul8/DDImageTexture.usf", "MainCS", SF_Compute);

void UDDImageTexture::Prepare()
{
    Super::Prepare();

    // Ensure the texture is fully loaded
    check(InputTexture);
    if (bNeedToUpdateTextureResource)
    {
        InputTexture->UpdateResource();
        bNeedToUpdateTextureResource = false;
    }
}

void UDDImageTexture::Execute(DDImageProcessorExecuteContext_GameThread& context)
{
    FScopeLock Lock(&DataMutex);
}

void UDDImageTexture::Execute(DDImageProcessorExecuteContext_RenderThread& context)
{
    ensure(IsInRenderingThread());
    SCOPE_CYCLE_COUNTER(STAT_Popul8ImageTextureCS_Execute);
    DECLARE_GPU_STAT(Popul8ImageTextureCS)
    RDG_EVENT_SCOPE(*context.GraphBuilder, "Popul8ImageTextureCS");
    RDG_GPU_STAT_SCOPE(*context.GraphBuilder, Popul8ImageTextureCS);

    FScopeLock Lock(&DataMutex);
    typename FDDImageTextureCS::FPermutationDomain PermutationVector;
    TShaderMapRef<FDDImageTextureCS> ComputeShader(GetGlobalShaderMap(GMaxRHIFeatureLevel), PermutationVector);
    bool bIsShaderValid = ComputeShader.IsValid();
    check(bIsShaderValid);

    FDDImageTextureCS::FParameters* PassParameters = context.GraphBuilder->AllocParameters<FDDImageTextureCS::FParameters>();
    FTextureResource* InputTextureResource = InputTexture->GetResource();
    FTextureRenderTargetResource* RenderTargetResource = ResultRenderTarget->GetRenderTargetResource();

    // Output
    FRDGTextureDesc TempTextureDesc = FRDGTextureDesc::Create2D(
        RenderTargetResource->GetSizeXY(),
        PF_B8G8R8A8,
        FClearValueBinding::White,
        TexCreate_RenderTargetable | TexCreate_ShaderResource | TexCreate_UAV,
        /* NumMips */ 1u);
    FRDGTextureRef TmpTexture = context.GraphBuilder->CreateTexture(TempTextureDesc, TEXT("Popul8ImageTextureCS.TempTexture"));
    FRDGTextureRef TargetTexture = RegisterExternalTexture(*context.GraphBuilder, RenderTargetResource->GetRenderTargetTexture(), TEXT("Popul8ImageTextureCS.RT"));
    PassParameters->OutputTexture = context.GraphBuilder->CreateUAV(TmpTexture);

    // Input
    FRDGTextureRef Source_0 = RegisterExternalTexture(*context.GraphBuilder, InputTextureResource->GetTexture2DRHI(), TEXT("Popul8ImageTextureCS.InputTexture"));
    PassParameters->InputTexture = context.GraphBuilder->CreateSRV(FRDGTextureSRVDesc::CreateForMipLevel(Source_0, 0));
    PassParameters->InputTextureSampler = TStaticSamplerState<SF_Bilinear, AM_Clamp, AM_Clamp, AM_Clamp>::GetRHI();
    PassParameters->Scale = Scale;
    PassParameters->Offset = Offset;

    // Add Pass
    auto GroupCount = FComputeShaderUtils::GetGroupCount(
        FIntVector(ResultRenderTarget->SizeX, ResultRenderTarget->SizeY, 1),
        FIntVector(NUM_THREADS_X, NUM_THREADS_Y, NUM_THREADS_Z));
    context.GraphBuilder->AddPass(
        RDG_EVENT_NAME("ExecutePopul8ImageTextureCS"),
        PassParameters,
        ERDGPassFlags::AsyncCompute,
        [PassParameters, ComputeShader, GroupCount](FRHIComputeCommandList& RHICmdList)
        {
            FComputeShaderUtils::Dispatch(RHICmdList, ComputeShader, *PassParameters, GroupCount);
        });

    // Copy to output render target
    check(TargetTexture->Desc.Format == PF_B8G8R8A8);
    AddCopyTexturePass(*context.GraphBuilder, TmpTexture, TargetTexture, FRHICopyTextureInfo());
}

void UDDImageTexture::SetInputTexture(UTexture2D* Input) 
{
    FScopeLock Lock(&DataMutex);
    InputTexture = Input;
    bNeedToUpdateTextureResource = true;
}

void UDDImageTexture::SetScaleAndOffset(const FVector2f& ImageTextureScale, const FVector2f& ImageTextureOffset)
{
    FScopeLock Lock(&DataMutex);
    Scale = ImageTextureScale;
    Offset = ImageTextureOffset;
}
