#include "Controllers/ImageProcessor/DDMeshUVColor.h"
#include "RHICommandList.h"
#include "RHIStaticStates.h"
#include "RenderGraph.h"
#include "RenderGraphUtils.h"
#include "RenderUtils.h"
#include "GlobalShader.h"
#include "Engine/TextureRenderTarget2D.h"
#include "TextureResource.h"


DECLARE_STATS_GROUP(TEXT("Popul8MeshUVColorRS"), STATGROUP_Popul8MeshUVColorRS, STATCAT_Advanced);
DECLARE_CYCLE_STAT(TEXT("Popul8MeshUVColorRS Execute"), STAT_Popul8MeshUVColorRS_Execute, STATGROUP_Popul8MeshUVColorRS);

// Vertex shader
class FDDMeshUVColorVS : public FGlobalShader
{
    DECLARE_GLOBAL_SHADER(FDDMeshUVColorVS);
    SHADER_USE_PARAMETER_STRUCT(FDDMeshUVColorVS, FGlobalShader)

public:
    BEGIN_SHADER_PARAMETER_STRUCT(FParameters, )
        SHADER_PARAMETER(FVector2f, Scale)
        SHADER_PARAMETER(FVector2f, Offset)
    END_SHADER_PARAMETER_STRUCT()

public:
    static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters) { return true; }

    static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
    {
        FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
    }
};
IMPLEMENT_GLOBAL_SHADER(FDDMeshUVColorVS, "/Plugin/Popul8/DDMeshUVColor.usf", "MainVS", SF_Vertex);

// Pixel shader
class FDDMeshUVColorPS : public FGlobalShader
{
    DECLARE_GLOBAL_SHADER(FDDMeshUVColorPS)
    SHADER_USE_PARAMETER_STRUCT(FDDMeshUVColorPS, FGlobalShader)

public:
    BEGIN_SHADER_PARAMETER_STRUCT(FParameters, )
        SHADER_PARAMETER(FVector4f, ColorMultiplier)
        RENDER_TARGET_BINDING_SLOTS()
    END_SHADER_PARAMETER_STRUCT()

public:
    static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters) { return true; }

    static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
    {
        FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
    }
};
IMPLEMENT_GLOBAL_SHADER(FDDMeshUVColorPS, "/Plugin/Popul8/DDMeshUVColor.usf", "MainPS", SF_Pixel);


void UDDMeshUVColor::Execute(DDImageProcessorExecuteContext_GameThread& context)
{
    FScopeLock Lock(&DataMutex);
}

void UDDMeshUVColor::Execute(DDImageProcessorExecuteContext_RenderThread& context)
{
    ensure(IsInRenderingThread());

    SCOPE_CYCLE_COUNTER(STAT_Popul8MeshUVColorRS_Execute);
    DECLARE_GPU_STAT(Popul8MeshUVColorRS)
    RDG_EVENT_SCOPE(*context.GraphBuilder, "Popul8MeshUVColorRS");
    RDG_GPU_STAT_SCOPE(*context.GraphBuilder, Popul8MeshUVColorRS);

    FScopeLock Lock(&DataMutex);
    auto VS_PassParameters = context.GraphBuilder->AllocParameters<FDDMeshUVColorVS::FParameters>();
    auto PS_PassParameters = context.GraphBuilder->AllocParameters<FDDMeshUVColorPS::FParameters>();

    auto* ShaderMap = GetGlobalShaderMap(GMaxRHIFeatureLevel);
    typename FDDMeshUVColorVS::FPermutationDomain PermutationVectorVS;
    TShaderMapRef<FDDMeshUVColorVS> VertexShader(ShaderMap, PermutationVectorVS);
    bool bIsVSValid = VertexShader.IsValid();
    check(bIsVSValid);
    typename FDDMeshUVColorPS::FPermutationDomain PermutationVectorPS;
    TShaderMapRef<FDDMeshUVColorPS> PixelShader(ShaderMap, PermutationVectorVS);
    bool bIsPSValid = PixelShader.IsValid();
    check(bIsPSValid);

    // Output
    FLinearColor CustomClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    FRDGTextureDesc TempTextureDesc = FRDGTextureDesc::Create2D(
        FIntPoint(SizeX, SizeY),
        PF_B8G8R8A8,
        FClearValueBinding(CustomClearColor),
        TexCreate_RenderTargetable | TexCreate_ShaderResource | TexCreate_UAV,
        /* NumMips */ 1u);
    FRDGTextureRef TmpTexture = context.GraphBuilder->CreateTexture(TempTextureDesc, TEXT("Popul8MeshUVColorRS.TempTexture"));
    FTextureRenderTargetResource* ResultTextureResource = ResultRenderTarget->GetRenderTargetResource();
    FRDGTextureRef ResultTextureRef = RegisterExternalTexture(*context.GraphBuilder, ResultTextureResource->GetRenderTargetTexture(), TEXT("Popul8MeshUVColorRS.RT"));
    //PassParameters->OutputTexture = context.GraphBuilder->CreateUAV(TmpTexture);
    PS_PassParameters->RenderTargets[0] = FRenderTargetBinding(TmpTexture, ERenderTargetLoadAction::EClear, 0);


    // VS Input
    VS_PassParameters->Scale = FVector2f(1.0f, 1.0f);
    VS_PassParameters->Offset = FVector2f(0.0f, 0.0f);

    // PS Input
    PS_PassParameters->ColorMultiplier = ColorMultiplier;

    // Add Pass
    context.GraphBuilder->AddPass(
        RDG_EVENT_NAME("ExecutePopul8MeshUVColorRS"),
        PS_PassParameters,
        ERDGPassFlags::Raster,
        [VertexShader, PixelShader, VS_PassParameters, PS_PassParameters, this](FRHICommandListImmediate& RHICmdList)
        {
            FGraphicsPipelineStateInitializer GraphicsPSOInit;
            RHICmdList.ApplyCachedRenderTargets(GraphicsPSOInit);

            // Set up blending - opaque
            GraphicsPSOInit.BlendState = TStaticBlendState<>::GetRHI();

            // Disable backface culling
            GraphicsPSOInit.RasterizerState = TStaticRasterizerState<FM_Solid, CM_None, ERasterizerDepthClipMode::DepthClip, false>::GetRHI();

            // Disable Depth/Stencil test
            GraphicsPSOInit.DepthStencilState = TStaticDepthStencilState<false, CF_Always>::GetRHI();

            // Define vertex buffer
            FVertexDeclarationElementList Elements;
            Elements.Add(FVertexElement(0, 0, EVertexElementType::VET_Float2, 0, sizeof(FVector2f)));
            Elements.Add(FVertexElement(1, 0, EVertexElementType::VET_Float4, 1, sizeof(FVector4f)));
            GraphicsPSOInit.BoundShaderState.VertexDeclarationRHI = PipelineStateCache::GetOrCreateVertexDeclaration(Elements);
            GraphicsPSOInit.BoundShaderState.VertexShaderRHI = VertexShader.GetVertexShader();
            GraphicsPSOInit.PrimitiveType = PT_TriangleList;

            // Set VS/PS shaders
            GraphicsPSOInit.BoundShaderState.PixelShaderRHI = PixelShader.GetPixelShader();
            SetGraphicsPipelineState(RHICmdList, GraphicsPSOInit, 0);
            SetShaderParameters(RHICmdList, PixelShader, PixelShader.GetPixelShader(), *PS_PassParameters);
            SetShaderParameters(RHICmdList, VertexShader, VertexShader.GetVertexShader(), *VS_PassParameters);

            // Set vertex buffer
            RHICmdList.SetStreamSource(0, UV_VB_RHI, 0);
            RHICmdList.SetStreamSource(1, Color_VB_RHI, 0);
            RHICmdList.DrawPrimitive(0, UVs.Num() / 3, 1);
        });

    // Copy to output render target
    check(ResultTextureRef->Desc.Format == PF_B8G8R8A8);
    AddCopyTexturePass(*context.GraphBuilder, TmpTexture, ResultTextureRef, FRHICopyTextureInfo());
}

void UDDMeshUVColor::setUVsAndColors(const TArray<FVector2f>& uvs, const TArray<FColor>& colors)
{
    check(uvs.Num() == colors.Num());

    // UVs
    if (UV_VB_RHI.IsValid())
        UV_VB_RHI.SafeRelease();
    UVs = uvs;
    ENQUEUE_RENDER_COMMAND(LockAndModifyVertexBuffer)(
        [this](FRHICommandListImmediate& RHICmdList)
        {
            // Allocate vertex buffer memory and set the buffer to non-volatile
            FRHIResourceCreateInfo CreateInfo(TEXT("Popul8MeshUVColorRS.UV_VertexBuffer"));
            UV_VB_RHI = RHICmdList.CreateVertexBuffer(sizeof(FVector2f) * UVs.Num(), EBufferUsageFlags::Static, CreateInfo);

            // Lock the buffer to access it for writing and then unlock it
            void* VoidPtr = RHICmdList.LockBuffer(UV_VB_RHI, 0, sizeof(FVector2f) * UVs.Num(), RLM_WriteOnly);
            FVector2f* Vertices = reinterpret_cast<FVector2f*>(VoidPtr);
            for (int32 i = 0; i < UVs.Num(); i++)
            {
                Vertices[i] = UVs[i];
            }
            RHICmdList.UnlockBuffer(UV_VB_RHI);
        }
        );

    // Colors
    if (Color_VB_RHI.IsValid())
        Color_VB_RHI.SafeRelease();
    Colors.SetNum(colors.Num());
    for (int32 i = 0; i < colors.Num(); i++)
        Colors[i] = FVector4f(colors[i]);
    ENQUEUE_RENDER_COMMAND(LockAndModifyVertexBuffer)(
        [this](FRHICommandListImmediate& RHICmdList)
        {
            // Allocate vertex buffer memory and set the buffer to non-volatile
            FRHIResourceCreateInfo CreateInfo(TEXT("Popul8MeshUVColorRS.Colors_VertexBuffer"));
            Color_VB_RHI = RHICmdList.CreateVertexBuffer(sizeof(FVector4f) * Colors.Num(), EBufferUsageFlags::Static, CreateInfo);

            // Lock the buffer to access it for writing and then unlock it
            void* VoidPtr = RHICmdList.LockBuffer(Color_VB_RHI, 0, sizeof(FVector4f) * Colors.Num(), RLM_WriteOnly);
            FVector4f* Vertices = reinterpret_cast<FVector4f*>(VoidPtr);
            for (int32 i = 0; i < Colors.Num(); i++)
            {
                Vertices[i] = Colors[i];
            }
            RHICmdList.UnlockBuffer(Color_VB_RHI);
        }
        );
}

void UDDMeshUVColor::setColorMultiplier(const FColor& colorMultiplier)
{
    ColorMultiplier = FVector4f(colorMultiplier);
}

UDDMeshUVColor::~UDDMeshUVColor()
{
    if (UV_VB_RHI.IsValid())
        UV_VB_RHI.SafeRelease();
    if (Color_VB_RHI.IsValid())
        Color_VB_RHI.SafeRelease();
}
