#include "Controllers/ImageProcessor/DDMeshUVTexture.h"
#include "RHICommandList.h"
#include "RHIStaticStates.h"
#include "Engine/Texture2D.h"
#include "Engine/TextureRenderTarget2D.h"
#include "RenderGraph.h"
#include "RenderGraphUtils.h"
#include "RenderUtils.h"
#include "GlobalShader.h"
#include "TextureResource.h"


DECLARE_STATS_GROUP(TEXT("Popul8MeshUVTextureRS"), STATGROUP_Popul8MeshUVTextureRS, STATCAT_Advanced);
DECLARE_CYCLE_STAT(TEXT("Popul8MeshUVTextureRS Execute"), STAT_Popul8MeshUVTextureRS_Execute, STATGROUP_Popul8MeshUVTextureRS);

// Vertex shader
class FDDMeshUVTextureVS : public FGlobalShader
{
    DECLARE_GLOBAL_SHADER(FDDMeshUVTextureVS);
    SHADER_USE_PARAMETER_STRUCT(FDDMeshUVTextureVS, FGlobalShader)

public:
    BEGIN_SHADER_PARAMETER_STRUCT(FParameters, )
        SHADER_PARAMETER(FVector2f, Scale)
        SHADER_PARAMETER(FVector2f, Offset)
    END_SHADER_PARAMETER_STRUCT()

public:
    static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters) { return true; }

    static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
    {
        FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
    }
};
IMPLEMENT_GLOBAL_SHADER(FDDMeshUVTextureVS, "/Plugin/Popul8/DDMeshUVTexture.usf", "MainVS", SF_Vertex);

// Pixel shader
class FDDMeshUVTexturePS : public FGlobalShader
{
    DECLARE_GLOBAL_SHADER(FDDMeshUVTexturePS)
    SHADER_USE_PARAMETER_STRUCT(FDDMeshUVTexturePS, FGlobalShader)

public:
    BEGIN_SHADER_PARAMETER_STRUCT(FParameters, )
        SHADER_PARAMETER(FVector4f, ColorMultiplier)
        SHADER_PARAMETER_RDG_TEXTURE_SRV(Texture2D, InputTexture)
        SHADER_PARAMETER_SAMPLER(SamplerState, InputTextureSampler)
        RENDER_TARGET_BINDING_SLOTS()
    END_SHADER_PARAMETER_STRUCT()

public:
    static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters) { return true; }

    static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
    {
        FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
    }
};
IMPLEMENT_GLOBAL_SHADER(FDDMeshUVTexturePS, "/Plugin/Popul8/DDMeshUVTexture.usf", "MainPS", SF_Pixel);


void UDDMeshUVTexture::Execute(DDImageProcessorExecuteContext_GameThread& context)
{
    FScopeLock Lock(&DataMutex);   
    check(SlotToDependency.Num() == 1);
    check(SlotToDependency[0]);
    ExecuteDependencies(context);
}

void UDDMeshUVTexture::Execute(DDImageProcessorExecuteContext_RenderThread& context)
{
    ensure(IsInRenderingThread());
    ExecuteDependencies(context);

    SCOPE_CYCLE_COUNTER(STAT_Popul8MeshUVTextureRS_Execute);
    DECLARE_GPU_STAT(Popul8MeshUVTextureRS)
    RDG_EVENT_SCOPE(*context.GraphBuilder, "Popul8MeshUVTextureRS");
    RDG_GPU_STAT_SCOPE(*context.GraphBuilder, Popul8MeshUVTextureRS);

    FScopeLock Lock(&DataMutex);
    auto VS_PassParameters = context.GraphBuilder->AllocParameters<FDDMeshUVTextureVS::FParameters>();
    auto PS_PassParameters = context.GraphBuilder->AllocParameters<FDDMeshUVTexturePS::FParameters>();

    auto* ShaderMap = GetGlobalShaderMap(GMaxRHIFeatureLevel);
    typename FDDMeshUVTextureVS::FPermutationDomain PermutationVectorVS;
    TShaderMapRef<FDDMeshUVTextureVS> VertexShader(ShaderMap, PermutationVectorVS);
    bool bIsVSValid = VertexShader.IsValid();
    check(bIsVSValid);
    typename FDDMeshUVTexturePS::FPermutationDomain PermutationVectorPS;
    TShaderMapRef<FDDMeshUVTexturePS> PixelShader(ShaderMap, PermutationVectorVS);
    bool bIsPSValid = PixelShader.IsValid();
    check(bIsPSValid);

    // Output
    FLinearColor CustomClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    FRDGTextureDesc TempTextureDesc = FRDGTextureDesc::Create2D(
        FIntPoint(SizeX, SizeY),
        PF_B8G8R8A8,
        FClearValueBinding(CustomClearColor),
        TexCreate_RenderTargetable | TexCreate_ShaderResource | TexCreate_UAV,
        /* NumMips */ 1u);
    FRDGTextureRef TmpTexture = context.GraphBuilder->CreateTexture(TempTextureDesc, TEXT("Popul8MeshUVTextureRS.TempTexture"));
    FTextureRenderTargetResource* ResultTextureResource = ResultRenderTarget->GetRenderTargetResource();
    FRDGTextureRef ResultTextureRef = RegisterExternalTexture(*context.GraphBuilder, ResultTextureResource->GetRenderTargetTexture(), TEXT("Popul8MeshUVTextureRS.RT"));
    //PassParameters->OutputTexture = context.GraphBuilder->CreateUAV(TmpTexture);
    PS_PassParameters->RenderTargets[0] = FRenderTargetBinding(TmpTexture, ERenderTargetLoadAction::EClear, 0);


    // VS Input
    VS_PassParameters->Scale = FVector2f(1.0f, 1.0f);
    VS_PassParameters->Offset = FVector2f(0.0f, 0.0f);

    // PS Input
    PS_PassParameters->ColorMultiplier = ColorMultiplier;
    PS_PassParameters->InputTexture = InputDependencyAsSRV(context, 0);
    PS_PassParameters->InputTextureSampler = TStaticSamplerState<SF_Bilinear, AM_Clamp, AM_Clamp, AM_Clamp>::GetRHI();

    // Add Pass
    context.GraphBuilder->AddPass(
        RDG_EVENT_NAME("ExecutePopul8MeshUVTextureRS"),
        PS_PassParameters,
        ERDGPassFlags::Raster,
        [VertexShader, PixelShader, VS_PassParameters, PS_PassParameters, this](FRHICommandListImmediate& RHICmdList)
        {
            FGraphicsPipelineStateInitializer GraphicsPSOInit;
            RHICmdList.ApplyCachedRenderTargets(GraphicsPSOInit);

            // Set up blending - opaque
            GraphicsPSOInit.BlendState = TStaticBlendState<>::GetRHI();

            // Disable backface culling
            GraphicsPSOInit.RasterizerState = TStaticRasterizerState<FM_Solid, CM_None, ERasterizerDepthClipMode::DepthClip, false>::GetRHI();

            // Disable Depth/Stencil test
            GraphicsPSOInit.DepthStencilState = TStaticDepthStencilState<false, CF_Always>::GetRHI();

            // Define vertex buffer
            FVertexDeclarationElementList Elements;
            Elements.Add(FVertexElement(0, 0, EVertexElementType::VET_Float2, 0, sizeof(FVector2f)));
            GraphicsPSOInit.BoundShaderState.VertexDeclarationRHI = PipelineStateCache::GetOrCreateVertexDeclaration(Elements);
            GraphicsPSOInit.BoundShaderState.VertexShaderRHI = VertexShader.GetVertexShader();
            GraphicsPSOInit.PrimitiveType = PT_TriangleList;

            // Set VS/PS shaders
            GraphicsPSOInit.BoundShaderState.PixelShaderRHI = PixelShader.GetPixelShader();
            SetGraphicsPipelineState(RHICmdList, GraphicsPSOInit, 0);
            SetShaderParameters(RHICmdList, PixelShader, PixelShader.GetPixelShader(), *PS_PassParameters);
            SetShaderParameters(RHICmdList, VertexShader, VertexShader.GetVertexShader(), *VS_PassParameters);

            // Set vertex buffer
            RHICmdList.SetStreamSource(0, VertexBufferRHI, 0);
            RHICmdList.DrawPrimitive(0, UVs.Num() / 3, 1);
        });

    // Copy to output render target
    check(ResultTextureRef->Desc.Format == PF_B8G8R8A8);
    AddCopyTexturePass(*context.GraphBuilder, TmpTexture, ResultTextureRef, FRHICopyTextureInfo());
}

void UDDMeshUVTexture::setUVs(const TArray<FVector2f>& uvs)
{
    // Release first if already created
    if (VertexBufferRHI.IsValid())
        VertexBufferRHI.SafeRelease();

    UVs = uvs;
    ENQUEUE_RENDER_COMMAND(LockAndModifyVertexBuffer)(
        [this](FRHICommandListImmediate& RHICmdList)
        {
            // Allocate vertex buffer memory and set the buffer to non-volatile
            FRHIResourceCreateInfo CreateInfo(TEXT("Popul8MeshUVTextureRS.VertexBuffer"));
            VertexBufferRHI = RHICmdList.CreateVertexBuffer(sizeof(FVector2f) * UVs.Num(), EBufferUsageFlags::Static, CreateInfo);

            // Lock the buffer to access it for writing and then unlock it
            void* VoidPtr = RHICmdList.LockBuffer(VertexBufferRHI, 0, sizeof(FVector2f) * UVs.Num(), RLM_WriteOnly);
            FVector2f* Vertices = reinterpret_cast<FVector2f*>(VoidPtr);
            for (int32 i = 0; i < UVs.Num(); i++)
            {
                Vertices[i] = UVs[i];
            }
            RHICmdList.UnlockBuffer(VertexBufferRHI);
        }
        );
}

void UDDMeshUVTexture::setColorMultiplier(const FColor& colorMultiplier)
{
    ColorMultiplier = FVector4f(colorMultiplier);
}

UDDMeshUVTexture::~UDDMeshUVTexture()
{
    if (VertexBufferRHI.IsValid())
        VertexBufferRHI.SafeRelease();
}
