// Fill out your copyright notice in the Description page of Project Settings.

#include "DDModelSerializer.h"
#include "Misc/Paths.h"
#include "Misc/FileHelper.h"
#include "HAL/PlatformFileManager.h"
#include "Serialization/MemoryWriter.h"
#include "Serialization/MemoryReader.h"
#include "DDPopul8.h"
#include "Models/DDTextureEditingModel.h"
#include "ActorComponents/DDEditableSkeletalMeshComponent.h"
#include "Controllers/ImageProcessor/DDColorize.h"


struct SerializedDataHeader
{
    bool UseBoneShaping = false;
    bool UseMorphTargets = false;
    bool UseTextureEditing = false;
    int Version = 1;

    friend FArchive& operator<<(FArchive& Ar, SerializedDataHeader& Header)
    {
        Ar << Header.UseBoneShaping;
        Ar << Header.UseMorphTargets;
        Ar << Header.UseTextureEditing;
        Ar << Header.Version;
        return Ar;
    }
};

//inline FArchive& operator<<(FArchive& Ar, FDDBoneTransformationState& State)
//{
//    Ar << State.Position;
//    Ar << State.Scale;
//    Ar << State.Rotation;
//    return Ar;
//}

struct TextureEditingColorValue
{
    int32 ComponentIndex;
    int32 StackIndex;
    int32 InterfaceProcessorIndex;
    FColor Color;

    friend FArchive& operator<<(FArchive& Ar, TextureEditingColorValue& Value)
    {
        Ar << Value.ComponentIndex;
        Ar << Value.StackIndex;
        Ar << Value.InterfaceProcessorIndex;
        Ar << Value.Color;
        return Ar;
    }
};


UDDModelSerializer::UDDModelSerializer(const FObjectInitializer& ObjectInitializer)
    : Super(ObjectInitializer)
{

}

UDDModelSerializer::~UDDModelSerializer()
{

}

void UDDModelSerializer::AssignDataModels(UDDEditableSkeletalMeshComponent* SkeletalMeshComponent)
{
    if (!IsValid(SkeletalMeshComponent->CEConfigData))
    {
        UE_LOG(LogDidimo, Error, TEXT("%s - Unable to assign data models. No valid CEConfigData assigned"), ANSI_TO_TCHAR(__FUNCTION__));
        return;
    }
    TextureEditingModel = SkeletalMeshComponent->getTextureEditingModel();
}

bool UDDModelSerializer::SerializeToBytes(TArray<uint8>& OutBytes)
{
    FMemoryWriter MemoryWriter(OutBytes, true);
    SerializedDataHeader Header;
    Header.UseBoneShaping = false;
    Header.UseMorphTargets = false;
    Header.UseTextureEditing = TextureEditingModel != nullptr;
    MemoryWriter << Header;
    if (Header.UseBoneShaping)
    {
        //FScopeLock Lock(&BoneShapingModel->DataMutex);
        //MemoryWriter << BoneShapingModel->SegmentationStates;
    }
    if (Header.UseMorphTargets)
    {
        //FScopeLock Lock(&MorphShapingModel->DataMutex);
        //MemoryWriter << MorphShapingModel->SegmentationVariationWeights;
    }
    if (Header.UseTextureEditing)
    {
        TArray<TextureEditingColorValue> ColorValues;
        for (int32 i = 0; i < TextureEditingModel->ImageProcessorComponents.Num(); i++)
        {
            auto ImageProcessorComponent = &TextureEditingModel->ImageProcessorComponents[i];
            for (int32 j = 0; j < ImageProcessorComponent->ImageProcessorStacks.Num(); j++)
            {
                auto Stack = &ImageProcessorComponent->ImageProcessorStacks[j];
                for (int32 k = 0; k < Stack->InterfaceProcessors.Num(); k++)
                {
                    auto Processor = Stack->InterfaceProcessors[k];
                    auto ColorizeProcessor = Cast<UDDColorize>(Processor);
                    if (ColorizeProcessor)
                    {
                        TextureEditingColorValue ColorValue;
                        ColorValue.ComponentIndex = i;
                        ColorValue.StackIndex = j;
                        ColorValue.InterfaceProcessorIndex = k;
                        ColorValue.Color = ColorizeProcessor->GetColor();
                        ColorValues.Add(ColorValue);
                    }
                }
            }
        }
        MemoryWriter << ColorValues;
    }
    return true;
}

bool UDDModelSerializer::DeserializeFromBytes(const TArray<uint8>& InBytes)
{
    FMemoryReader MemoryReader(InBytes, true);
    SerializedDataHeader Header;
    MemoryReader << Header;
    //check(Header.UseBoneShaping == (BoneShapingModel != nullptr));
    //check(Header.UseMorphTargets == (MorphShapingModel != nullptr));
    check(Header.UseTextureEditing == (TextureEditingModel != nullptr));
    if (Header.UseBoneShaping)
    {
        //FScopeLock Lock(&BoneShapingModel->DataMutex);
        //BoneShapingModel->MarkedDirty = true;
        //MemoryReader << BoneShapingModel->SegmentationStates;
    }
    if (Header.UseMorphTargets)
    {
        //FScopeLock Lock(&MorphShapingModel->DataMutex);
        //MorphShapingModel->MarkedDirty = true;
        //MemoryReader << MorphShapingModel->SegmentationVariationWeights;
    }
    if (Header.UseTextureEditing)
    {
        //FScopeLock Lock(&TextureEditingModel->DataMutex);
        TextureEditingModel->MarkedDirty = true;
        TArray<TextureEditingColorValue> ColorValues;
        MemoryReader << ColorValues;
        for (auto& ColorValue : ColorValues)
        {
            auto ImageProcessorComponent = &TextureEditingModel->ImageProcessorComponents[ColorValue.ComponentIndex];
            auto Stack = &ImageProcessorComponent->ImageProcessorStacks[ColorValue.StackIndex];
            auto Processor = Stack->InterfaceProcessors[ColorValue.InterfaceProcessorIndex];
            auto ColorizeProcessor = Cast<UDDColorize>(Processor);
            check(ColorizeProcessor);
            ColorizeProcessor->SetColor(ColorValue.Color);
        }
    }
    return true;
}

bool UDDModelSerializer::SaveToFile(const FString& FilePath, const TArray<uint8>& Bytes)
{
    // Check if the array is empty
    if (Bytes.Num() == 0)
    {
        UE_LOG(LogTemp, Warning, TEXT("SaveToFile: Cannot save an empty byte array to file."));
        return false;
    }

    // Attempt to save the file
    if (FFileHelper::SaveArrayToFile(Bytes, *FilePath))
    {
        UE_LOG(LogTemp, Log, TEXT("SaveToFile: Successfully saved file to %s"), *FilePath);
        return true;
    }
    else
    {
        UE_LOG(LogTemp, Error, TEXT("SaveToFile: Failed to save file to %s"), *FilePath);
        return false;
    }
}

bool UDDModelSerializer::LoadFromFile(const FString& FilePath, TArray<uint8>& OutBytes)
{
    // Check if the file exists
    if (FPlatformFileManager::Get().GetPlatformFile().FileExists(*FilePath))
    {
        // Attempt to load the file into the OutBytes array
        if (FFileHelper::LoadFileToArray(OutBytes, *FilePath))
        {
            UE_LOG(LogTemp, Log, TEXT("Successfully loaded file from %s"), *FilePath);
            return true; // Success
        }
        else
        {
            UE_LOG(LogTemp, Warning, TEXT("Failed to load file from %s"), *FilePath);
        }
    }
    else
    {
        UE_LOG(LogTemp, Warning, TEXT("File not found: %s"), *FilePath);
    }

    return false; // Failure
}

bool UDDModelSerializer::SerializeEditableSkeletalMeshComponent(UDDEditableSkeletalMeshComponent* SkeletalMeshComponent, TArray<uint8>& OutBytes)
{
    UDDModelSerializer* Serializer = NewObject<UDDModelSerializer>();
    Serializer->AssignDataModels(SkeletalMeshComponent);
    return Serializer->SerializeToBytes(OutBytes);
}

bool UDDModelSerializer::DeserializeEditableSkeletalMeshComponent(UDDEditableSkeletalMeshComponent* SkeletalMeshComponent, const TArray<uint8>& Bytes)
{
    UDDModelSerializer* Serializer = NewObject<UDDModelSerializer>();
    Serializer->AssignDataModels(SkeletalMeshComponent);
    return Serializer->DeserializeFromBytes(Bytes);
}
