// Fill out your copyright notice in the Description page of Project Settings.

#include "DataAssets/DDDataAssets.h"
#include "JsonObjectConverter.h"
#include "IImageWrapperModule.h"
#include "IImageWrapper.h"
#include "ImageUtils.h"
#include "Engine/Texture2D.h"
#include "Misc/Paths.h"
#include "Misc/FileHelper.h"
#include "Modules/ModuleManager.h"
#include "TextureResource.h"


EPixelFormat ConvertRGBFormatToPixelFormat(ERGBFormat RGBFormat, int32 BitDepth)
{
    switch (RGBFormat)
    {
    case ERGBFormat::Gray:
        return (BitDepth == 16) ? PF_G16 : PF_G8; // 16-bit grayscale or 8-bit grayscale

    case ERGBFormat::RGBA:
        return (BitDepth == 16) ? PF_FloatRGBA : PF_R8G8B8A8; // 16-bit float RGBA or 8-bit RGBA

    case ERGBFormat::BGRA:
        return PF_B8G8R8A8; // 8-bit BGRA format

    default:
        return PF_Unknown; // Unknown format
    }
}

UTexture2D* LoadTexture2DFromFile(const FString& FilePath, bool bUseSRGB)
{
    UTexture2D* LoadedTexture = nullptr;
    TArray<uint8> RawFileData;

    // Load file into byte array
    if (FFileHelper::LoadFileToArray(RawFileData, *FilePath))
    {
        // Get the image wrapper module
        IImageWrapperModule& ImageWrapperModule = FModuleManager::LoadModuleChecked<IImageWrapperModule>(FName("ImageWrapper"));

        // Detect the image format
        EImageFormat ImageFormat = ImageWrapperModule.DetectImageFormat(RawFileData.GetData(), RawFileData.Num());
        if (ImageFormat != EImageFormat::PNG)
        {
            UE_LOG(LogTemp, Warning, TEXT("Failed to load image: %s"), *FilePath);
            return nullptr;
        }

        // Create an image wrapper for the detected format
        TSharedPtr<IImageWrapper> ImageWrapper = ImageWrapperModule.CreateImageWrapper(ImageFormat);
        if (ImageWrapper.IsValid() && ImageWrapper->SetCompressed(RawFileData.GetData(), RawFileData.Num()))
        {
            // Get the raw image data (RGBA 8-bit)
            int32 bitDepth = ImageWrapper->GetBitDepth();
            ERGBFormat format = ImageWrapper->GetFormat();
            EPixelFormat pixelFormat = ConvertRGBFormatToPixelFormat(format, bitDepth);
            TArray64<uint8> UncompressedRGBA;
            ImageWrapper->GetRaw(UncompressedRGBA);
            if (ImageWrapper->GetRaw(UncompressedRGBA))
            {
                // Create the texture from the raw image data
                LoadedTexture = UTexture2D::CreateTransient(ImageWrapper->GetWidth(), ImageWrapper->GetHeight(), pixelFormat);
                if (LoadedTexture == nullptr)
                {
                    UE_LOG(LogTemp, Warning, TEXT("Failed to create a texture while loading the image: %s"), *FilePath);
                    return nullptr;
                }

                // Set the SRGB flag
                LoadedTexture->SRGB = bUseSRGB;

                // Lock the texture so we can modify its content
                FTexture2DMipMap& Mip = LoadedTexture->GetPlatformData()->Mips[0];
                void* TextureData = Mip.BulkData.Lock(LOCK_READ_WRITE);
                FMemory::Memcpy(TextureData, UncompressedRGBA.GetData(), UncompressedRGBA.Num());
                Mip.BulkData.Unlock();
            }
        }
    }

    return LoadedTexture;
}

TStrongObjectPtr<UDDAssetCatalogData> UDDAssetCatalogData::Load(const FString& AssetCatalogFilePath)
{
    FString DirectoryPath = FPaths::GetPath(AssetCatalogFilePath);
    FString jsonString;
    if (!FFileHelper::LoadFileToString(jsonString, *AssetCatalogFilePath))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to load file: %s"), *AssetCatalogFilePath);
        return nullptr;
    }

    // Parse the JSON string
    TSharedPtr<FJsonObject> JsonObject;
    TSharedRef<TJsonReader<>> jsonReader = TJsonReaderFactory<>::Create(jsonString);
    bool result = FJsonSerializer::Deserialize(jsonReader, JsonObject);
    check(result);
    result = JsonObject.IsValid();
    check(result);

    // Create a new instance
    TStrongObjectPtr<UDDAssetCatalogData> NewAsset = TStrongObjectPtr<UDDAssetCatalogData>(NewObject<UDDAssetCatalogData>());
    if (!NewAsset)
    {
        UE_LOG(LogTemp, Error, TEXT("Failed to create UDDAssetCatalogData object."));
        return nullptr;
    }

    // Populate the asset from JSON
    if (!FJsonObjectConverter::JsonObjectToUStruct(JsonObject.ToSharedRef(), UDDAssetCatalogData::StaticClass(), NewAsset.Get(), 0, 0))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to parse JSON into UDDAssetCatalogData"));
        return nullptr;
    }

    //
    // Texture loading
    // 
    //// Array to hold the graph events (tasks) for synchronization
    //FGraphEventArray Tasks;
    //for (auto& texture : NewAsset->textures)
    //{
    //    FString TextureFilePath = DirectoryPath + "/" + texture.path;
    //    bool UseSRGB = !texture.isLinear;
    //    FGraphEventRef Task = FFunctionGraphTask::CreateAndDispatchWhenReady([TextureFilePath, UseSRGB, &texture]()
    //        {
    //            // Load texture in the worker thread
    //            UTexture2D* LoadedTexture = LoadTexture2DFromFile(TextureFilePath, UseSRGB);

    //            // Ensure the texture is updated and assigned back on the game thread
    //            AsyncTask(ENamedThreads::GameThread, [&texture, LoadedTexture]()
    //                {
    //                    LoadedTexture->UpdateResource();
    //                    texture.texture = LoadedTexture;
    //                });
    //        }, TStatId(), nullptr, ENamedThreads::AnyThread);
    //    Tasks.Add(Task);
    //}

    //// Wait for all tasks to complete before returning control to the game thread
    //if (Tasks.Num() > 0)
    //    FTaskGraphInterface::Get().WaitUntilTasksComplete(Tasks, ENamedThreads::GameThread);

    return NewAsset;
}

TStrongObjectPtr<UDDCharacterDescriptorData> UDDCharacterDescriptorData::Load(const FString& CharacterDescriptorFilePath)
{
    FString DirectoryPath = FPaths::GetPath(CharacterDescriptorFilePath);
    FString jsonString;
    if (!FFileHelper::LoadFileToString(jsonString, *CharacterDescriptorFilePath))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to load file: %s"), *CharacterDescriptorFilePath);
        return nullptr;
    }

    // Parse the JSON string
    TSharedPtr<FJsonObject> JsonObject;
    TSharedRef<TJsonReader<>> jsonReader = TJsonReaderFactory<>::Create(jsonString);
    bool result = FJsonSerializer::Deserialize(jsonReader, JsonObject);
    check(result);
    result = JsonObject.IsValid();
    check(result);

    // Create a new instance
    TStrongObjectPtr<UDDCharacterDescriptorData> NewAsset = TStrongObjectPtr<UDDCharacterDescriptorData>(NewObject<UDDCharacterDescriptorData>());
    if (!NewAsset)
    {
        UE_LOG(LogTemp, Error, TEXT("Failed to create UDDCharacterDescriptorData object."));
        return nullptr;
    }

    // Populate the asset from JSON
    if (!FJsonObjectConverter::JsonObjectToUStruct(JsonObject.ToSharedRef(), UDDCharacterDescriptorData::StaticClass(), NewAsset.Get(), 0, 0))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to parse JSON into UDDCharacterDescriptorData"));
        return nullptr;
    }

    return NewAsset;
}

TStrongObjectPtr<UDDCharacterEditorConfigData> UDDCharacterEditorConfigData::Load(const FString& AssetCatalogFilePath)
{
    FString DirectoryPath = FPaths::GetPath(AssetCatalogFilePath);
    FString jsonString;
    if (!FFileHelper::LoadFileToString(jsonString, *AssetCatalogFilePath))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to load file: %s"), *AssetCatalogFilePath);
        return nullptr;
    }

    // Parse the JSON string
    TSharedPtr<FJsonObject> JsonObject;
    TSharedRef<TJsonReader<>> jsonReader = TJsonReaderFactory<>::Create(jsonString);
    bool result = FJsonSerializer::Deserialize(jsonReader, JsonObject);
    check(result);
    result = JsonObject.IsValid();
    check(result);

    // Create a new instance
    TStrongObjectPtr<UDDCharacterEditorConfigData> NewAsset = TStrongObjectPtr<UDDCharacterEditorConfigData>(NewObject<UDDCharacterEditorConfigData>());
    if (!NewAsset)
    {
        UE_LOG(LogTemp, Error, TEXT("Failed to create UDDCharacterEditorConfigData object."));
        return nullptr;
    }

    // Populate the asset from JSON
    if (!FJsonObjectConverter::JsonObjectToUStruct(JsonObject.ToSharedRef(), UDDCharacterEditorConfigData::StaticClass(), NewAsset.Get(), 0, 0))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to parse JSON into UDDCharacterEditorConfigData"));
        return nullptr;
    }

    //
    // Texture loading
    // 
    //// Array to hold the graph events (tasks) for synchronization
    //FGraphEventArray Tasks;
    //for (auto& texture : NewAsset->textures)
    //{
    //    FString TextureFilePath = DirectoryPath + "/" + texture.path;
    //    bool UseSRGB = !texture.isLinear;
    //    FGraphEventRef Task = FFunctionGraphTask::CreateAndDispatchWhenReady([TextureFilePath, UseSRGB, &texture]()
    //        {
    //            // Load texture in the worker thread
    //            UTexture2D* LoadedTexture = LoadTexture2DFromFile(TextureFilePath, UseSRGB);

    //            // Ensure the texture is updated and assigned back on the game thread
    //            AsyncTask(ENamedThreads::GameThread, [&texture, LoadedTexture]()
    //                {
    //                    LoadedTexture->UpdateResource();
    //                    texture.texture = LoadedTexture;
    //                });
    //        }, TStatId(), nullptr, ENamedThreads::AnyThread);
    //    Tasks.Add(Task);
    //}

    //// Wait for all tasks to complete before returning control to the game thread
    //if (Tasks.Num() > 0)
    //    FTaskGraphInterface::Get().WaitUntilTasksComplete(Tasks, ENamedThreads::GameThread);

    return NewAsset;
}
