#include "DataAssets/DDVertexCacheData.h"
#include "Rendering/SkeletalMeshRenderData.h"
#include "HAL/PlatformFileManager.h"
#include "Misc/FileHelper.h"
#include "DDKdTree.h"


void BuildKdTreeFromMesh(
	FPositionVertexBuffer* PositionVertexBuffer,
	FDDKdTree* outKdTree)
{
	const int32 NumSoftVertices = PositionVertexBuffer->GetNumVertices();
	TArray<FVector> KdTreeData;
	KdTreeData.SetNum(NumSoftVertices);
	for (int32 VertexIndex = 0; VertexIndex < NumSoftVertices; VertexIndex++)
	{
		FVector3f VPosition = PositionVertexBuffer->VertexPosition(VertexIndex);
		KdTreeData[VertexIndex] = FVector(VPosition);
	}
	BuildKdtree(outKdTree, KdTreeData);
	ValidateKdtree(*outKdTree);
}

/// <summary>
/// Attachment Vertex Cache Data
/// </summary>
void UDDAttachmentVertexCacheData::BuildData()
{
	TArray<float> CullVertexData;
	if (!LoadBinaryFloats(BinFileDirectory, CullVertexData))
	{
		UE_LOG(LogTemp, Error, TEXT("There are no culls on this file"));
		return;
	}

	auto RenderResource_SM = TargetSkeletalMesh->GetResourceForRendering();
	auto LODRenderData_SM = &RenderResource_SM->LODRenderData[0];

	// SM
	auto StaticVertexBuffers_SM = &LODRenderData_SM->StaticVertexBuffers;
	auto StaticVertexBuffer_SM = &StaticVertexBuffers_SM->StaticMeshVertexBuffer;
	auto PositionVB_SM = &StaticVertexBuffers_SM->PositionVertexBuffer;

	// Create the SM_To_MTSM_IndexMap and SM_To_BSSM_IndexMap map
	FDDKdTree KdTree_SM;
	BuildKdTreeFromMesh(PositionVB_SM, &KdTree_SM);

	VertexOpacity.SetNum(StaticVertexBuffer_SM->GetNumVertices());
	for (auto& vo : VertexOpacity)
		vo = 1.0f;

	for (int32 CullIndex = 0; CullIndex < CullVertexData.Num(); CullIndex += 3)
	{
		FVector3f CullVertex = FVector3f(CullVertexData[CullIndex], -CullVertexData[CullIndex + 2], CullVertexData[CullIndex + 1]) * 100;

		TArray<int> Result;
		// TODO: std::numeric_limits::
		static float Radius = 0.001f;
		CollectFromKdtree(KdTree_SM, FVector(CullVertex), Radius, &Result);

		for (int index : Result)
		{
			check(index != -1);
			VertexOpacity[index] = 0.0f;
		}
		//int best_VBIndex_SM = FindClosestInKdtree(KdTree_SM, FVector(CullVertex));
		//check(best_VBIndex_SM != -1);
		//VertexOpacity[best_VBIndex_SM] = 0.0f;
	}

	UE_LOG(LogTemp, Log, TEXT("Culling data computed successfully"));
}

bool UDDAttachmentVertexCacheData::LoadBinaryFloats(const FString& FilePath, TArray<float>& OutFloats)
{
	if (!FPlatformFileManager::Get().GetPlatformFile().FileExists(*FilePath))
	{
		UE_LOG(LogTemp, Error, TEXT("File does not exist: %s"), *FilePath);
		return false;
	}

	TArray<uint8> BinaryData;
	if (!FFileHelper::LoadFileToArray(BinaryData, *FilePath))
	{
		UE_LOG(LogTemp, Error, TEXT("Failed to load file: %s"), *FilePath);
		return false;
	}

	// Convert byte data to floats
	int32 FloatCount = BinaryData.Num() / sizeof(float);
	OutFloats.SetNumUninitialized(FloatCount);
	FMemory::Memcpy(OutFloats.GetData(), BinaryData.GetData(), BinaryData.Num());

	return true;
}
