#include "Models/DDTextureEditingModel.h"
#include "DataAssets/DDDataAssets.h"
#include "DataAssets/DDCharacterEditorSkinCompositionData.h"
#include "RHICommandList.h"
#include "RHIStaticStates.h"
#include "Engine/Texture2D.h"
#include "Engine/TextureRenderTarget2D.h"
#include "Misc/PackageName.h"

#include "Controllers/ImageProcessor/DDImageTexture.h"
#include "Controllers/ImageProcessor/DDColorize.h"
#include "Controllers/ImageProcessor/DDBrightnessContrast.h"
#include "Controllers/ImageProcessor/DDDilation.h"
#include "Controllers/ImageProcessor/DDSwizzle.h"
#include "Controllers/ImageProcessor/DDBlendTexture.h"
#include "Controllers/ImageProcessor/DDMeshUVColor.h"
#include "Controllers/ImageProcessor/DDMeshUVTexture.h"


FDDTextureEditingBindingKey::FDDTextureEditingBindingKey()
{
    MeshName = "";
    MaterialKey = EDDTextureEditingMaterialKey::MK_None;
}

FDDTextureEditingBindingKey::FDDTextureEditingBindingKey(const FString& MeshNameStr, const FString& MaterialKeyStr)
{
    MeshName = MeshNameStr;
    if (MaterialKeyStr.Compare("_BaseColorMap") == 0)
        MaterialKey = EDDTextureEditingMaterialKey::MK_Albedo;
    else if (MaterialKeyStr.Compare("_NormalMap") == 0)
        MaterialKey = EDDTextureEditingMaterialKey::MK_Normal;
    else
        MaterialKey = EDDTextureEditingMaterialKey::MK_None;
}

FDDTextureEditingBindingKey::FDDTextureEditingBindingKey(const FString& MeshNameStr, EDDTextureEditingMaterialKey KeyMaterialKeyStr)
{
    MeshName = MeshNameStr;
    MaterialKey = KeyMaterialKeyStr;
}

bool FDDTextureEditingBindingKey::operator==(const FDDTextureEditingBindingKey& Other) const
{
    return MeshName == Other.MeshName && MaterialKey == Other.MaterialKey;
}

uint32 GetTypeHash(const FDDTextureEditingBindingKey& Key)
{
    return HashCombine(GetTypeHash(Key.MeshName), GetTypeHash(Key.MaterialKey));
}

void UDDTextureEditingModel::SetImageTextureInputTexture(UDDImageTexture* ImageTextureProcessor, const FString& Path)
{
    FString NormalizedPath = FPaths::ConvertRelativePathToFull(FPaths::Combine(ConfigDir, Path));
    FString AssetPath = FPaths::GetPath(NormalizedPath) / FPaths::GetBaseFilename(NormalizedPath);
    UTexture2D* Tex = Cast<UTexture2D>(StaticLoadObject(UTexture2D::StaticClass(), nullptr, *AssetPath));
    check(Tex);
    ImageTextureProcessor->SetInputTexture(Tex);
}

UDDImageProcessor* UDDTextureEditingModel::AddImageProcessor(
    const FString& BlendMode,
    const TArray<FString>& stringParams,
    const TArray<int32>& intParams,
    UDDImageProcessor* PreviousProcessor)
{
    UDDImageProcessor* aProc = nullptr;
    UDDImageProcessor* bProc = nullptr;
    UDDImageProcessor* maskPreSwizzleProc = nullptr;

    if (PreviousProcessor != nullptr)
        aProc = PreviousProcessor;

    if (stringParams.Num() == 1 && intParams.Num() == 0)
    {
        auto proc = NewObject<UDDImageTexture>();
        CurrentStack->ImageProcessorNodes.Add(proc);
        SetImageTextureInputTexture(proc, stringParams[0]);
        bProc = proc;
    }
    else if (stringParams.Num() == 1 && intParams.Num() == 2)
    {
        check(intParams[0] == -1);
        check(intParams[1] != -1);
        auto proc = NewObject<UDDImageTexture>();
        CurrentStack->ImageProcessorNodes.Add(proc);
        SetImageTextureInputTexture(proc, stringParams[0]);
        bProc = proc;

        auto RefStack_1 = &CurrentComponent->ImageProcessorStacks[intParams[1]];
        maskPreSwizzleProc = RefStack_1->ImageProcessorNodes.Last();
    }
    else if (stringParams.Num() == 0 && intParams.Num() == 2)
    {
        check(intParams[0] != -1);
        check(intParams[1] != -1);

        auto RefStack_0 = &CurrentComponent->ImageProcessorStacks[intParams[0]];
        auto refProc_0 = RefStack_0->ImageProcessorNodes.Last();
        bProc = refProc_0;

        auto RefStack_1 = &CurrentComponent->ImageProcessorStacks[intParams[1]];
        maskPreSwizzleProc = RefStack_1->ImageProcessorNodes.Last();
    }
    else
    {
        throw;
    }

    CurrentStack->InterfaceProcessors.Add(nullptr); // UDDImageProcessor can't be interfaced
    return AddBlendingProcessor(aProc, bProc, maskPreSwizzleProc);
}

UDDImageProcessor* UDDTextureEditingModel::AddColorizeProcessor(
    const FString& BlendMode,
    const TArray<FString>& stringParams,
    const TArray<int32>& intParams,
    UDDImageProcessor* PreviousProcessor)
{
    UDDImageProcessor* aProc = nullptr;
    UDDImageProcessor* bProc = nullptr;
    UDDImageProcessor* maskPreSwizzleProc = nullptr;

    if (PreviousProcessor != nullptr)
        aProc = PreviousProcessor;
    else
        throw; // must have the previous proc

    auto proc = NewObject<UDDColorize>();
    CurrentStack->ImageProcessorNodes.Add(proc);
    CurrentStack->InterfaceProcessors.Add(proc);
    proc->SetDependency(aProc);
    bProc = proc;

    if (stringParams.Num() == 0 && intParams.Num() == 0)
    {
        // this is fine, nothing to do
    }
    else if (stringParams.Num() == 1 && intParams.Num() == 0)
    {
        auto imageProc = NewObject<UDDImageTexture>();
        CurrentStack->ImageProcessorNodes.Add(imageProc);
        SetImageTextureInputTexture(imageProc, stringParams[0]);
        maskPreSwizzleProc = imageProc;
    }
    else if (stringParams.Num() == 0 && intParams.Num() == 1)
    {
        check(intParams[0] != -1);
        auto RefStack_0 = &CurrentComponent->ImageProcessorStacks[intParams[0]];
        auto refProc_0 = RefStack_0->ImageProcessorNodes.Last();
        maskPreSwizzleProc = refProc_0;
    }
    else
    {
        throw;
    }

    return AddBlendingProcessor(aProc, bProc, maskPreSwizzleProc);
}

UDDImageProcessor* UDDTextureEditingModel::AddBlendingProcessor(
    UDDImageProcessor* AProcessor,
    UDDImageProcessor* BProcessor,
    UDDImageProcessor* MaskProcessor)
{
    UDDImageProcessor* MaskProcAfterSwizzle = nullptr;
    if (MaskProcessor)
    {
        auto siwzzleProc = NewObject<UDDSwizzle>();
        CurrentStack->ImageProcessorNodes.Add(siwzzleProc);
        siwzzleProc->SetDependency(MaskProcessor);
        siwzzleProc->SetArrangement(
            EDDSwizzleChannel::SC_Red,
            EDDSwizzleChannel::SC_Red,
            EDDSwizzleChannel::SC_Red,
            EDDSwizzleChannel::SC_One);
        MaskProcAfterSwizzle = siwzzleProc;
    }

    check(BProcessor);
    if (MaskProcAfterSwizzle == nullptr)
    {
        return BProcessor; // nothing to blend
    }
    else if (AProcessor)
    {
        auto blendProc = NewObject<UDDBlendTexture>();
        CurrentStack->ImageProcessorNodes.Add(blendProc);
        blendProc->SetDependency(AProcessor, 0);
        blendProc->SetDependency(BProcessor, 1);
        blendProc->SetDependency(MaskProcAfterSwizzle, 2);
        blendProc->SetOperation(EDDBlendTextureOperation::BTO_Lerp);
        return blendProc;
    }
    else
    {
        auto  blendProc = NewObject<UDDBlendTexture>();
        CurrentStack->ImageProcessorNodes.Add(blendProc);
        blendProc->SetDependency(BProcessor, 0);
        blendProc->SetDependency(MaskProcAfterSwizzle, 1);
        blendProc->SetOperation(EDDBlendTextureOperation::BTO_Mul);
        return blendProc;
    }
}

UDDImageProcessor* UDDTextureEditingModel::AddOptionalBlending(
    const FString& BlendMode,
    UDDImageProcessor* CurrentProcessor,
    UDDImageProcessor* PreviousProcessor)
{
    if (!BlendMode.IsEmpty())
    {
        if (BlendMode.Compare("Normal") == 0)
        {
            return CurrentProcessor;
        }
        else
        {
            auto blendProc = NewObject<UDDBlendTexture>();
            CurrentStack->ImageProcessorNodes.Add(blendProc);
            blendProc->SetDependency(PreviousProcessor, 0);
            blendProc->SetDependency(CurrentProcessor, 1);
            if (BlendMode.Compare("LinearDodge") == 0)
            {
                blendProc->SetOperation(EDDBlendTextureOperation::BTO_Add);
            }
            else
            {
                throw;
            }
            return blendProc;
        }
    }
    else
    {
        return CurrentProcessor;
    }
}

void UDDTextureEditingModel::Initialize(UDDCharacterEditorSkinCompositionData* Config, USkeletalMesh* SkinnedMesh)
{
    MarkedDirty = true;
    ResultNodes.Empty();
    BindingToRenderTarget.Empty();
    FString ConfigPath = FPackageName::ObjectPathToPackageName(Config->GetPathName());
    ConfigDir = FPaths::GetPath(ConfigPath);

    ImageProcessorComponents.Empty();
    for (auto& component : Config->components)
    {
        ImageProcessorComponents.Add(FDDTextureEditingComponent());
        CurrentComponent = &ImageProcessorComponents.Last();
        UDDImageProcessor* prevProc;
        for (auto& stack : component.processorStacks)
        {
            CurrentComponent->ImageProcessorStacks.Add(FDDTextureEditingProcessorStack());
            CurrentStack = &CurrentComponent->ImageProcessorStacks.Last();
            prevProc = nullptr;
            for (auto& processor : stack.processors)
            {
                if (processor.type.Compare("ImageTexture") == 0)
                {
                    prevProc = AddImageProcessor(processor.blendMode, processor.stringParams, processor.intParams, prevProc);
                }
                else if (processor.type.Compare("Swizzle") == 0)
                {
                    if (processor.intParams.Num() == 4)
                    {
                        auto proc = NewObject<UDDSwizzle>();
                        CurrentStack->ImageProcessorNodes.Add(proc);
                        CurrentStack->InterfaceProcessors.Add(proc);
                        check(prevProc);
                        proc->SetDependency(prevProc);
                        proc->SetArrangement(
                            (EDDSwizzleChannel)processor.intParams[0],
                            (EDDSwizzleChannel)processor.intParams[1],
                            (EDDSwizzleChannel)processor.intParams[2],
                            (EDDSwizzleChannel)processor.intParams[3]);

                        prevProc = AddOptionalBlending(processor.blendMode, proc, prevProc);
                    }
                    else
                    {
                        throw;
                    }
                }
                else if (processor.type.Compare("Colorize") == 0)
                {
                    prevProc = AddColorizeProcessor(processor.blendMode, processor.stringParams, processor.intParams, prevProc);
                }
                else
                {
                    throw;
                }
            }

            for (auto imageProcessor : CurrentStack->ImageProcessorNodes)
            {
                imageProcessor->Initialize(4096, 4096);
            }

            if (!stack.materialKey.IsEmpty())
            {
                auto ImageProc = CurrentStack->ImageProcessorNodes.Last();
                ResultNodes.Add(ImageProc);
                for (auto& meshName : component.meshes)
                {
                    auto BindingKey = FDDTextureEditingBindingKey(meshName, stack.materialKey);
                    BindingToRenderTarget.Add(BindingKey, ImageProc->GetResultRenderTarget());
                }
            }
        }
    }
}

bool UDDTextureEditingModel::SupportsMesh(const FDDTextureEditingBindingKey& Key) const
{
    return BindingToRenderTarget.Contains(Key);
}

UTextureRenderTarget2D* UDDTextureEditingModel::GetMeshRenderTarget(const FDDTextureEditingBindingKey& Key)
{
    return BindingToRenderTarget[Key];
}

