#include "DDBridgeCharacterEditorImport.h"
#include "AssetImportTask.h"
#include "AssetToolsModule.h"
#include "ActorComponents/DDEditableSkeletalMeshComponent.h"
#include "Kismet2/KismetEditorUtilities.h"
#include "Kismet2/BlueprintEditorUtils.h"
#include "GameFramework/Character.h"

FString UDDBridgeCharacterEditorImport::BlueprintTemplatePath = TEXT("/Popul8/BP_DDEditableCharacterTemplate");

void UDDBridgeCharacterEditorImport::Import(const FString& SourceDir)
{
	FString RelativeContentPath = TEXT("Popul8Characters"); // Path relative to the Content folder
	FString DestinationDir = FPaths::Combine(TEXT("/Game"), RelativeContentPath);

	// Get all files and directories in the source directory
	TArray<FString> FoundFiles;
	TArray<FString> FoundDirectories;
	IFileManager::Get().FindFilesRecursive(FoundFiles, *SourceDir, TEXT("*.*"), true, false);
	FString LastDirectoryName = FPaths::GetCleanFilename(SourceDir);

	// Asset catalog first
	TStrongObjectPtr<UDDAssetCatalogData> AssetCatalogData = nullptr;
	FString AssetCatalogDir;
	if (!LoadAssetCatalog(FoundFiles, &AssetCatalogData, AssetCatalogDir))
	{
		UE_LOG(LogTemp, Warning, TEXT("Failed to import dir: %s"), *SourceDir);
		return;
	}

	// Looking for the template asset to load
	FString templatePath = FPaths::GetPath(AssetCatalogData->characterEditors[0].characterEditorBasePath);
	FDDTemplateAsset* TemplateAsset = nullptr;
	for (auto& ta : AssetCatalogData->templateAssets)
		if (ta.path.Equals(templatePath, ESearchCase::IgnoreCase))
			TemplateAsset = &ta;
	if (TemplateAsset == nullptr)
	{
		UE_LOG(LogTemp, Warning, TEXT("AssetCatalog inconsistency TemplateAsset %s does not exist in the AssetCatalog"),
		       *templatePath);
		return;
	}

	// Load CE_config
	check(AssetCatalogData->characterEditors.Num() == 1);
	TStrongObjectPtr<UDDCharacterEditorConfigData> CharacterEditorConfigData = nullptr;
	FString ceConfigDir;
	if (!LoadCEConfig(FPaths::Combine(AssetCatalogDir,
	                                  AssetCatalogData->characterEditors[0].characterEditorConfigPath),
	                  &CharacterEditorConfigData, ceConfigDir))
	{
		UE_LOG(LogTemp, Warning, TEXT("Failed to import dir: %s"), *SourceDir);
		return;
	}

	// Loading ce_skin_composition
	TStrongObjectPtr<UDDCharacterEditorSkinCompositionData> SkinCompositionConfigData = nullptr;
	FString SkinCompositionConfigDir;
	if (!LoadCESkinComposition(FPaths::Combine(ceConfigDir, "ce_skin_composition.json"), // TODO: Hardcoded path get this from AssetCatalog
		&SkinCompositionConfigData, SkinCompositionConfigDir))
	{
		UE_LOG(LogTemp, Warning, TEXT("Failed to import dir: %s"), *SourceDir);
		return;
	}

	// Import textures
	TArray<UTexture2D*> Textures;
	Textures.SetNum(AssetCatalogData->textures.Num());
	for (int32 i = 0; i < AssetCatalogData->textures.Num(); i++)
	{
		if (!AssetCatalogData->textures[i].path.IsEmpty())
		{
			FString FilePath = FPaths::Combine(AssetCatalogDir, AssetCatalogData->textures[i].path);
			FString RelativePath = FilePath.RightChop(SourceDir.Len());
			FString DestinationFilePath = FPaths::Combine(DestinationDir, LastDirectoryName, RelativePath);
			DestinationFilePath = FPaths::GetPath(DestinationFilePath);
			Textures[i] = ImportTexture(FilePath, DestinationFilePath, AssetCatalogData->textures[i].isLinear, false);
		}
		else
		{
			UE_LOG(LogTemp, Warning, TEXT("Texture not found in the Popul8 Bridge import"));
		}
	}

	// Importing CharacterEditor textures
	for (auto& meshAlbedo : CharacterEditorConfigData->meshAlbedos)
	{
		if (!meshAlbedo.texturePath.IsEmpty())
		{
			FString FilePath = FPaths::Combine(ceConfigDir, meshAlbedo.texturePath);
			FString RelativePath = FilePath.RightChop(SourceDir.Len());
			FString DestinationFilePath = FPaths::Combine(DestinationDir, LastDirectoryName, RelativePath);
			DestinationFilePath = FPaths::GetPath(DestinationFilePath);
			ImportTexture(FilePath, DestinationFilePath, false, false);
		}
		else
		{
			UE_LOG(LogTemp, Warning, TEXT("Texture not found in the Popul8 Bridge import"));
		}
	}

	// Import Segmentations
	//CharacterEditorConfigData 
	
	// Import MeshMasks
	for (auto& segmentation : CharacterEditorConfigData->segmentations)
	{
		//for (auto& textureEdit : segmentation.textureEdit)
		//{
		//	// Gradient
		//	if (!textureEdit.gradient.gradientTexturePath.IsEmpty())
		//	{
		//		FString FilePath = FPaths::Combine(ceConfigDir, textureEdit.gradient.gradientTexturePath);
		//		FString RelativePath = FilePath.RightChop(SourceDir.Len());
		//		FString DestinationFilePath = FPaths::Combine(DestinationDir, LastDirectoryName, RelativePath);
		//		DestinationFilePath = FPaths::GetPath(DestinationFilePath);
		//		ImportTexture(FilePath, DestinationFilePath, true, false);
		//	}
		//	else
		//	{
		//		UE_LOG(LogTemp, Warning, TEXT("Texture not found in the Popul8 Bridge import"));
		//	}

		//	// Mask path
		//	if (!textureEdit.maskPath.IsEmpty())
		//	{
		//		FString FilePath = FPaths::Combine(ceConfigDir, textureEdit.maskPath);
		//		FString RelativePath = FilePath.RightChop(SourceDir.Len());
		//		FString DestinationFilePath = FPaths::Combine(DestinationDir, LastDirectoryName, RelativePath);
		//		DestinationFilePath = FPaths::GetPath(DestinationFilePath);
		//		ImportTexture(FilePath, DestinationFilePath, true, false);
		//	}
		//	else
		//	{
		//		UE_LOG(LogTemp, Warning, TEXT("Texture not found in the Popul8 Bridge import"));
		//	}
		//}

		for (auto& meshMask : segmentation.meshMasks)
		{
			FString FilePath = FPaths::Combine(ceConfigDir, meshMask.texturePath);
			FString RelativePath = FilePath.RightChop(SourceDir.Len());
			FString DestinationFilePath = FPaths::Combine(DestinationDir, LastDirectoryName, RelativePath);
			DestinationFilePath = FPaths::GetPath(DestinationFilePath);
			ImportTexture(FilePath, DestinationFilePath, true, true);
		}
	}

	// Create materials
	ImportMaterialsFrom(AssetCatalogData.Get(), TemplateAsset->skinnedRenderObjects, Textures,
	                    DestinationDir / LastDirectoryName / TemplateAsset->path);

	//
	// Import FBXs
	//
	USkeletalMesh* TemplateSKM = nullptr;
	check(AssetCatalogData->templateAssets.Num() == 1);
	// Importing Template from asset catalog
	{
		FString FilePath = FPaths::Combine(AssetCatalogDir, AssetCatalogData->templateAssets[0].LODData[0].fbxPath);
		FString RelativePath = FilePath.RightChop(SourceDir.Len());
		FString DestinationFilePath = FPaths::Combine(DestinationDir, LastDirectoryName, RelativePath);
		DestinationFilePath = FPaths::GetPath(DestinationFilePath);
		TemplateSKM = ImportFBXMesh(nullptr, FilePath, DestinationFilePath, true);
	}

	USkeletalMesh* RigShapeSKM = nullptr;
	// Importing RigShape from CharacterEditor
	{
		FString FilePath = FPaths::Combine(ceConfigDir, CharacterEditorConfigData->rigShapePath);
		FString RelativePath = FilePath.RightChop(SourceDir.Len());
		FString DestinationFilePath = FPaths::Combine(DestinationDir, LastDirectoryName, RelativePath);
		DestinationFilePath = FPaths::GetPath(DestinationFilePath);
		RigShapeSKM = ImportFBXMesh(nullptr, FilePath, DestinationFilePath, false);
	}

	// Importing Animation from CharacterEditor (optional)
	if (CharacterEditorConfigData->animationPath.IsEmpty())
	{
		FString FilePath = FPaths::Combine(ceConfigDir, CharacterEditorConfigData->animationPath);
		FString RelativePath = FilePath.RightChop(SourceDir.Len());
		FString DestinationFilePath = FPaths::Combine(DestinationDir, LastDirectoryName, RelativePath);
		DestinationFilePath = FPaths::GetPath(DestinationFilePath);
		ImportFBXMesh(nullptr, FilePath, DestinationFilePath, false);
	}

	FString RootDestinationDir = FPaths::Combine(DestinationDir, LastDirectoryName);
	// Import AssetCatalog as its data asset type
	{
		auto JsonFactory = GetFactoryForFileType(TEXT("json"));
		FAssetToolsModule& AssetToolsModule = FModuleManager::LoadModuleChecked<FAssetToolsModule>("AssetTools");
		IAssetTools& AssetTools = AssetToolsModule.Get();

		UAssetImportTask* ImportTask = NewObject<UAssetImportTask>();
		ImportTask->Filename = AssetCatalogDir / "AssetCatalog.json";
		ImportTask->Factory = JsonFactory;
		ImportTask->DestinationPath = RootDestinationDir;
		ImportTask->bAutomated = true; // Set to true for automated import without dialogs

		TArray<UAssetImportTask*> ImportTasks;
		ImportTasks.Add(ImportTask);
		AssetTools.ImportAssetTasks(ImportTasks);
	}
	// Import CE config as its data asset type
	UDDCharacterEditorConfigData* CE_ConfigData = nullptr;
	{
		auto JsonFactory = GetFactoryForFileType(TEXT("json"));
		FAssetToolsModule& AssetToolsModule = FModuleManager::LoadModuleChecked<FAssetToolsModule>("AssetTools");
		IAssetTools& AssetTools = AssetToolsModule.Get();

		UAssetImportTask* ImportTask = NewObject<UAssetImportTask>();
		ImportTask->Filename = FPaths::Combine(AssetCatalogDir,
		                                       AssetCatalogData->characterEditors[0].characterEditorConfigPath);
		ImportTask->Factory = JsonFactory;
		ImportTask->DestinationPath = FPaths::GetPath(
			RootDestinationDir / AssetCatalogData->characterEditors[0].characterEditorConfigPath);
		ImportTask->bAutomated = true; // Set to true for automated import without dialogs

		TArray<UAssetImportTask*> ImportTasks;
		ImportTasks.Add(ImportTask);
		AssetTools.ImportAssetTasks(ImportTasks);

		CE_ConfigData = Cast<UDDCharacterEditorConfigData>(ImportTask->GetObjects()[0]);
	}

	// Create character
	auto CharacterName = AssetCatalogData->templateAssets[0].assetName;
	UBlueprint* BPCharacter = Cast<UBlueprint>(CreateBlueprint("BP_" + CharacterName, RootDestinationDir, BlueprintTemplatePath));

	ACharacter* Character = Cast<ACharacter>(BPCharacter->GeneratedClass->GetDefaultObject());
	UDDEditableSkeletalMeshComponent* UDDSKMC = Cast<UDDEditableSkeletalMeshComponent>(Character->GetMesh());
	UDDSKMC->SetSkeletalMesh(RigShapeSKM);
	throw; // THIS NEEDS TO BE MADE TO WORK WITH THE NEW DATA (INCLUDING UDDCEVertexCacheData )
	//UDDSKMC->CEConfigData = CE_ConfigData;
	//UDDSKMC->MorphTargetsSkeletalMesh = TemplateSKM;
	//UDDSKMC->BoneShapingSkeletalMesh = RigShapeSKM;
	//UDDSKMC->MTCEConfigData = MTCEConfigData;


	//FKismetEditorUtilities::CompileBlueprint(BPCharacter);
	//FBlueprintEditorUtils::MarkBlueprintAsStructurallyModified(BPCharacter);
}
