#include "DDBridgeImportBase.h"
#include "Factories/FbxFactory.h"
#include "Factories/FbxImportUI.h"
#include "Factories/TextureFactory.h"
#include "Factories/MaterialFactoryNew.h"
#include "Factories/AnimBlueprintFactory.h"
#include "Factories/FbxSkeletalMeshImportData.h"
#include "Materials/Material.h"
#include "Materials/MaterialExpression.h"
#include "Materials/MaterialExpressionTextureSampleParameter2D.h"
#include "Materials/MaterialExpressionVertexColor.h"
#include "MaterialEditingLibrary.h"
#include "Materials/MaterialExpressionScalarParameter.h"
#include "Materials/MaterialExpressionVectorParameter.h"
#include "Engine/SkinnedAssetCommon.h"
#include "GameFramework/Character.h"
#include "Misc/Paths.h"
#include "AssetToolsModule.h"
#include "IAssetTools.h"
#include "AssetImportTask.h"
#include "ActorComponents/DDEditableSkeletalMeshComponent.h"
#include "DDDataAssetFactory.h"
#include "Materials/MaterialExpressionScalarParameter.h"
#include <ObjectTools.h>
#include "DDBridgeUtils.h"


UFactory* UDDBridgeImportBase::GetFactoryForFileType(const FString& FileExtension)
{
	if (FileExtension.Equals(TEXT("fbx"), ESearchCase::IgnoreCase))
	{
		return NewObject<UFbxFactory>();
	}
	else if (FileExtension.Equals(TEXT("png"), ESearchCase::IgnoreCase))
	{
		return NewObject<UTextureFactory>();
	}
	else if (FileExtension.Equals(TEXT("tga"), ESearchCase::IgnoreCase))
	{
		return NewObject<UTextureFactory>();
	}
	else if (FileExtension.Equals(TEXT("json"), ESearchCase::IgnoreCase))
	{
		return NewObject<UDDDataAssetFactory>();
	}
	return nullptr;
}

bool UDDBridgeImportBase::LoadAssetCatalog(const TArray<FString> FoundFiles, TStrongObjectPtr<UDDAssetCatalogData>* AssetCatalogData,
                                           FString& AssetCatalogDir)
{
	for (const FString& FilePath : FoundFiles)
	{
		if (FPaths::GetCleanFilename(FilePath).Equals(TEXT("AssetCatalog.json"), ESearchCase::IgnoreCase))
		{
			*AssetCatalogData = UDDAssetCatalogData::Load(FilePath);
			AssetCatalogDir = FPaths::GetPath(FilePath);
			return true;
		}
	}
	return false;
}

bool UDDBridgeImportBase::LoadCharacterDescriptor(const TArray<FString> FoundFiles,
											      TStrongObjectPtr<UDDCharacterDescriptorData>* CharacterDescriptorData,
                                                  FString& CharacterDescriptorDir)
{
	for (const FString& FilePath : FoundFiles)
	{
		if (FPaths::GetCleanFilename(FilePath).Equals(TEXT("CharacterDescriptor.json"), ESearchCase::IgnoreCase))
		{
			*CharacterDescriptorData = UDDCharacterDescriptorData::Load(FilePath);
			CharacterDescriptorDir = FPaths::GetPath(FilePath);
			return true;
		}
	}
	return false;
}

bool UDDBridgeImportBase::LoadCESkinComposition(const FString& Path, 
	TStrongObjectPtr<UDDCharacterEditorSkinCompositionData>* CharacterEditorSkinCompositionData,
	FString& CESkinCompositionDir)
{
	*CharacterEditorSkinCompositionData = UDDCharacterEditorSkinCompositionData::Load(Path);
	if (CharacterEditorSkinCompositionData->Get() == nullptr) return false;

	CESkinCompositionDir = FPaths::GetPath(Path);
	return true;
}


bool UDDBridgeImportBase::LoadCEConfig(const FString& Path, TStrongObjectPtr<UDDCharacterEditorConfigData>* CharacterEditorConfigData,
                                       FString& CEConfigDir)
{
	*CharacterEditorConfigData = UDDCharacterEditorConfigData::Load(Path);
	if (CharacterEditorConfigData->Get() == nullptr) return false;

	CEConfigDir = FPaths::GetPath(Path);
	return true;
}

USkeletalMesh* UDDBridgeImportBase::ImportFBXMesh(
	TArray<FDDSkinnedRenderObject>* SkinnedRenderObjects,
	const FString& FilePath,
	const FString& DestinationPath,
	bool ImportMorphTargets,
	USkeleton* OverrideSkeleton)
{
	FAssetToolsModule& AssetToolsModule = FModuleManager::LoadModuleChecked<FAssetToolsModule>("AssetTools");
	IAssetTools& AssetTools = AssetToolsModule.Get();
	FString FileExtension = FPaths::GetExtension(FilePath);
	UFactory* Factory = GetFactoryForFileType(FileExtension);
	if (UFbxFactory* FbxFactory = Cast<UFbxFactory>(Factory))
	{
		FbxFactory->ImportUI->bImportMaterials = false; // Do not create materials
		FbxFactory->ImportUI->bImportTextures = false; // Do not import textures
		FbxFactory->ImportUI->bImportAnimations = true; // Do import the animations
		
		if (OverrideSkeleton)
			FbxFactory->ImportUI->Skeleton = OverrideSkeleton;

		if (FbxFactory->ImportUI->SkeletalMeshImportData)
		{
			FbxFactory->ImportUI->bImportAsSkeletal = true;
			FbxFactory->ImportUI->SkeletalMeshImportData->bImportMorphTargets = ImportMorphTargets;
			FbxFactory->ImportUI->SkeletalMeshImportData->bUseT0AsRefPose = false;
			FbxFactory->ImportUI->SkeletalMeshImportData->bImportVertexAttributes = true;
			FbxFactory->ImportUI->SkeletalMeshImportData->VertexColorImportOption = EVertexColorImportOption::Override;
			FbxFactory->ImportUI->SkeletalMeshImportData->VertexOverrideColor = FColor::Red;
			
		}
	}
	else
	{
		UE_LOG(LogTemp, Warning, TEXT("No correct factory found for file: %s"), *FilePath);
		return nullptr;
	}

	UAssetImportTask* ImportTask = NewObject<UAssetImportTask>();
	ImportTask->Filename = FilePath;
	ImportTask->Factory = Factory;
	ImportTask->DestinationPath = DestinationPath;
	ImportTask->bAutomated = true; // Set to true for automated import without dialogs

	TArray<UAssetImportTask*> ImportTasks;
	ImportTasks.Add(ImportTask);
	AssetTools.ImportAssetTasks(ImportTasks);
	UE_LOG(LogTemp, Log, TEXT("Import attempt for file: %s"), *FilePath);

	// Generate the asset path
	FString FileNameWithoutExtension = FPaths::GetBaseFilename(FilePath);
	FString AssetPath = FPaths::Combine(DestinationPath, FileNameWithoutExtension);

	// Load the imported asset
	USkeletalMesh* Imported_SM = Cast<USkeletalMesh>(
		StaticLoadObject(USkeletalMesh::StaticClass(), nullptr, *AssetPath));
	if (!Imported_SM)
	{
		UE_LOG(LogTemp, Warning, TEXT("Failed to load imported asset: %s"), *AssetPath);
		return nullptr;
	}

	TMap<FString, TArray<FString>> meshNameToMaterialNames;
	for (size_t i = 0; i < SkinnedRenderObjects->Num(); i++)
	{
		TArray<FString> materialNames;
		for (size_t j = 0; j < (*SkinnedRenderObjects)[i].materialNames.Num(); j++)
		{
			FString materialName = (*SkinnedRenderObjects)[i].materialNames[j].ToLower();
			materialName = UDDBridgeUtils::NameSanitizer(materialName);
			materialNames.Add(materialName);
		}
		meshNameToMaterialNames.Add((*SkinnedRenderObjects)[i].meshName, materialNames);
	}

	// Assign the materials
	// TODO: THIS HAS TO BE DONE IN A BETTER WAY
	TArray<FSkeletalMaterial> Materials_SM = Imported_SM->GetMaterials();
	for (size_t i = 0; i < Materials_SM.Num(); i++)
	{
		bool Found = false;
		FString MaterialSlotName_SM = Materials_SM[i].MaterialSlotName.GetPlainNameString().ToLower();
		FString ImportedMaterialSlotName_SM = Materials_SM[i].ImportedMaterialSlotName.GetPlainNameString().ToLower();
		TArray<FString> ValidMaterialNames;
		if (meshNameToMaterialNames.Find(MaterialSlotName_SM))
			ValidMaterialNames = meshNameToMaterialNames[MaterialSlotName_SM];
		else if (meshNameToMaterialNames.Find(ImportedMaterialSlotName_SM))
			ValidMaterialNames = meshNameToMaterialNames[ImportedMaterialSlotName_SM];
		else for (auto kv : meshNameToMaterialNames) 
		{
			// This is a Fallback mechanism if anything fails,
			// we need to think this better
			if (MaterialSlotName_SM.StartsWith(kv.Key)) 
			{
				Found = true;
				ValidMaterialNames = kv.Value;
				break;
			}
		}

		for (auto& PathMaterial : PathToMaterial)
		{
			if (PathMaterial.Key.StartsWith(DestinationPath) == false)
				continue; // not in the right folder to be the material we need

			FString MaterialFileName = FPaths::GetBaseFilename(PathMaterial.Key);
			bool MaterialMatch = false;
			for (auto& ValideMaterialName : ValidMaterialNames)
			{
				if (MaterialFileName.Equals(ValideMaterialName))
				{
					MaterialMatch = true;
					break;
				}
			}

			if (MaterialMatch)
			{
				Materials_SM[i] = PathMaterial.Value;
				Materials_SM[i].UVChannelData.bInitialized = true; // to prevent the warning
				Found = true;
				break;
			}
		}

		//// If the material is not found is the case that the name of the material inside the FBX is not the same that the one
		//// that is in the asset catalog, selecting the first one as a measure
		//if (!Found)
		//{
		//	FString FallbackMaterialName_SM = FPaths::GetPath(MaterialName_SM);
		//	for (auto& NameMaterial : NameToMaterial)
		//	{
		//		FString NameMaterialPath = FPaths::GetPath(NameMaterial.Key);
		//		if (FallbackMaterialName_SM.StartsWith(NameMaterialPath))
		//		{
		//			Materials_SM[i] = NameMaterial.Value;
		//			Materials_SM[i].UVChannelData.bInitialized = true; // to prevent the warning
		//			break;
		//		}
		//		UE_LOG(LogTemp, Warning, TEXT("Material: %s, for asset: %s, not found, falling back to "), 
		//			*AssetPath, *MaterialName_SM, *NameMaterial.Key);
		//	}
		//}
	}
	Imported_SM->SetMaterials(Materials_SM);

	// Cleaning the autoimported skeleton
	/*if (OverrideSkeleton)
	{
		USkeleton* AutoImportedSkeleton = Imported_SM->Skeleton;
		Imported_SM->Skeleton = OverrideSkeleton;

		if (AutoImportedSkeleton) ObjectTools::DeleteSingleObject(AutoImportedSkeleton);
	}*/

	return Imported_SM;
}

UTexture2D* UDDBridgeImportBase::ImportTexture(const FString& FilePath, const FString& DestinationPath, bool isLinear,
                                               bool useCPUAccess = false)
{
	FAssetToolsModule& AssetToolsModule = FModuleManager::LoadModuleChecked<FAssetToolsModule>("AssetTools");
	IAssetTools& AssetTools = AssetToolsModule.Get();
	FString FileExtension = FPaths::GetExtension(FilePath);
	UFactory* Factory = GetFactoryForFileType(FileExtension);

	if (UTextureFactory* TextureFactory = Cast<UTextureFactory>(Factory))
	{
		TextureFactory->SuppressImportOverwriteDialog(true); // Avoid overwriting dialog
		TextureFactory->ColorSpaceMode = isLinear ? ETextureSourceColorSpace::Linear : ETextureSourceColorSpace::SRGB;
		TextureFactory->MipGenSettings = TMGS_NoMipmaps; // No mipmaps
	}
	else
	{
		UE_LOG(LogTemp, Warning, TEXT("No correct factory found for file: %s"), *FilePath);
		return nullptr;
	}

	UAssetImportTask* ImportTask = NewObject<UAssetImportTask>();
	ImportTask->Filename = FilePath;
	ImportTask->Factory = Factory;
	ImportTask->DestinationPath = DestinationPath;
	ImportTask->bAutomated = true; // Set to true for automated import without dialogs

	TArray<UAssetImportTask*> ImportTasks;
	ImportTasks.Add(ImportTask);
	AssetTools.ImportAssetTasks(ImportTasks);

	if (useCPUAccess)
	{
		auto ImportedObjects = ImportTask->GetObjects();
		UObject* ImportedAsset = ImportedObjects.Num() > 0 ? ImportTask->GetObjects()[0] : nullptr;
		if (UTexture* ImportedTexture = Cast<UTexture>(ImportedAsset))
		{
			ImportedTexture->Availability = ETextureAvailability::CPU;
			ImportedTexture->UpdateResource();
			UE_LOG(LogTemp, Log, TEXT("Texture imported with CPU access set: %s"), *FilePath);
		}
		else
		{
			UE_LOG(LogTemp, Warning, TEXT("Failed to cast imported asset to UTexture: %s"), *FilePath);
		}
	}

	UE_LOG(LogTemp, Log, TEXT("Import attempt for file: %s"), *FilePath);

	// Generate the asset path
	FString FileNameWithoutExtension = FPaths::GetBaseFilename(FilePath);
	FString AssetPath = FPaths::Combine(DestinationPath, FileNameWithoutExtension);

	// Load the imported texture
	UTexture2D* ImportedTexture = Cast<UTexture2D>(StaticLoadObject(UTexture2D::StaticClass(), nullptr, *AssetPath));
	if (!ImportedTexture)
	{
		UE_LOG(LogTemp, Warning, TEXT("Failed to load imported texture: %s"), *AssetPath);
		return nullptr;
	}
	return ImportedTexture;
}

UMaterial* UDDBridgeImportBase::CreateMaterial(
	const FString& MaterialName,
	const FString& DestinationPath,
	UTexture2D* BC_Texture,
	UTexture2D* N_Texture,
	UTexture2D* ARM_Texture,
	bool IsAlphaBlend)
{
	FAssetToolsModule& AssetToolsModule = FModuleManager::LoadModuleChecked<FAssetToolsModule>("AssetTools");
	IAssetTools& AssetTools = AssetToolsModule.Get();
	FString Filepath = FPaths::Combine(DestinationPath, MaterialName);

	UMaterialFactoryNew* MaterialFactory = NewObject<UMaterialFactoryNew>();
	UObject* NewMaterialAsset = AssetTools.CreateAsset(MaterialName, DestinationPath, UMaterial::StaticClass(),
	                                                   MaterialFactory);

	if (UMaterial* NewMaterial = Cast<UMaterial>(NewMaterialAsset))
	{
		NewMaterial->BlendMode = BLEND_Masked;
		NewMaterial->TwoSided = IsAlphaBlend;

		UE_LOG(LogTemp, Log, TEXT("Created material: %s"), *Filepath);
		// Save the material
		NewMaterial->Modify();

		// Create parameters
		auto BC_TextureParameter = CreateMaterialTextureParameter(NewMaterial, "BC_Texture", BC_Texture,
		                                                          SAMPLERTYPE_Color, FIntPoint(-400, 0));
		auto N_TextureParameter = CreateMaterialTextureParameter(NewMaterial, "N_Texture", N_Texture,
		                                                         SAMPLERTYPE_Normal, FIntPoint(-400, 300));
		auto ARM_TextureParameter = CreateMaterialTextureParameter(NewMaterial, "ARM_Texture", ARM_Texture,
		                                                           SAMPLERTYPE_LinearColor, FIntPoint(-400, 600));

		// Connect nodes
		if (IsAlphaBlend)
		{
			NewMaterial->GetEditorOnlyData()->BaseColor.Connect(0, BC_TextureParameter);
			NewMaterial->GetEditorOnlyData()->OpacityMask.Connect(4, BC_TextureParameter);
		}
		else if (BC_Texture)
		{
			NewMaterial->GetEditorOnlyData()->BaseColor.Connect(0, BC_TextureParameter);
		}
		else // just connect white color
		{
			auto BC_ColorNode = CreateMaterialVectorParameter(NewMaterial, "Albedo_Default", FIntPoint(-400, -300));
			BC_ColorNode->DefaultValue = FLinearColor::White;
			NewMaterial->GetEditorOnlyData()->BaseColor.Connect(0, BC_ColorNode);
		}

		if (N_Texture)
		{
			NewMaterial->GetEditorOnlyData()->Normal.Connect(0, N_TextureParameter);
		}
		
		if (ARM_Texture)
		{
			NewMaterial->GetEditorOnlyData()->Metallic.Connect(3, ARM_TextureParameter);
			NewMaterial->GetEditorOnlyData()->Roughness.Connect(2, ARM_TextureParameter);
			NewMaterial->GetEditorOnlyData()->AmbientOcclusion.Connect(1, ARM_TextureParameter);
		}

		NewMaterial->PostEditChange();
	}

	// Load the imported material
	UMaterial* ImportedMaterial = Cast<UMaterial>(StaticLoadObject(UMaterial::StaticClass(), nullptr, *Filepath));
	if (!ImportedMaterial)
	{
		UE_LOG(LogTemp, Warning, TEXT("Failed to load created material: %s"), *Filepath);
	}
	return ImportedMaterial;
}

UAnimBlueprint* UDDBridgeImportBase::CreateAnimationBlueprint(const FString& Name, const FString& DestinationPath, USkeleton* TargetSkeleton)
{
	FAssetToolsModule& AssetToolsModule = FModuleManager::LoadModuleChecked<FAssetToolsModule>("AssetTools");
	IAssetTools& AssetTools = AssetToolsModule.Get();
	FString Filepath = FPaths::Combine(DestinationPath, Name);

	UAnimBlueprintFactory* AnimBPFactory = NewObject<UAnimBlueprintFactory>();
	UObject* NewABPAsset = AssetTools.CreateAsset(Name, DestinationPath, UAnimBlueprint::StaticClass(), AnimBPFactory);

	if (UAnimBlueprint* AnimBP = Cast<UAnimBlueprint>(NewABPAsset))
	{
		AnimBP->TargetSkeleton = TargetSkeleton;

		AnimBP->PostEditChange();
	}

	// Load the imported material
	UAnimBlueprint* ImportedABP = Cast<UAnimBlueprint>(StaticLoadObject(UAnimBlueprint::StaticClass(), nullptr, *Filepath));
	if (!ImportedABP)
	{
		UE_LOG(LogTemp, Warning, TEXT("Failed to load created ABP: %s"), *Filepath);
	}
	return ImportedABP;
}

UMaterial* UDDBridgeImportBase::CreateInvisibleMaterial(
	const FString& MaterialName,
	const FString& DestinationPath)
{
	FAssetToolsModule& AssetToolsModule = FModuleManager::LoadModuleChecked<FAssetToolsModule>("AssetTools");
	IAssetTools& AssetTools = AssetToolsModule.Get();
	FString MaterialPath = FPaths::Combine(DestinationPath, MaterialName);

	UMaterialFactoryNew* MaterialFactory = NewObject<UMaterialFactoryNew>();
	UObject* NewMaterialAsset = AssetTools.CreateAsset(MaterialName, DestinationPath, UMaterial::StaticClass(),
	                                                   MaterialFactory);

	if (UMaterial* NewMaterial = Cast<UMaterial>(NewMaterialAsset))
	{
		NewMaterial->BlendMode = BLEND_Translucent;

		UE_LOG(LogTemp, Log, TEXT("Created material: %s"), *MaterialPath);
		// Save the material
		NewMaterial->Modify();

		// Create parameters
		auto Opacity_TextureParameter = CreateMaterialScalarParameter(NewMaterial, "Opacity", FIntPoint(-400, 900));

		// Connect nodes
		NewMaterial->GetEditorOnlyData()->Opacity.Connect(0, Opacity_TextureParameter);

		NewMaterial->PostEditChange();
	}

	// Load the imported material
	UMaterial* ImportedMaterial = Cast<UMaterial>(StaticLoadObject(UMaterial::StaticClass(), nullptr, *MaterialPath));
	if (!ImportedMaterial)
	{
		UE_LOG(LogTemp, Warning, TEXT("Failed to load created material: %s"), *MaterialPath);
	}
	return ImportedMaterial;
}

UMaterialExpressionTextureSampleParameter2D* UDDBridgeImportBase::CreateMaterialTextureParameter(
	UMaterial* Material,
	const FString& ParameterName,
	UTexture* Texture,
	EMaterialSamplerType SamplerType,
	const FIntPoint& NodePos)
{
	UMaterialExpression* Expression = UMaterialEditingLibrary::CreateMaterialExpression(
		Material, UMaterialExpressionTextureSampleParameter2D::StaticClass());
	UMaterialExpressionTextureSampleParameter2D* Parameter = Cast<
		UMaterialExpressionTextureSampleParameter2D>(Expression);
	Material->AddExpressionParameter(Parameter, Material->EditorParameters);

	Parameter->ParameterName = *ParameterName;
	Parameter->MaterialExpressionEditorX = NodePos.X;
	Parameter->MaterialExpressionEditorY = NodePos.Y;
	Parameter->Texture = Texture;
	Parameter->SamplerType = SamplerType;
	return Parameter;
}

UMaterialExpressionScalarParameter* UDDBridgeImportBase::CreateMaterialScalarParameter(
	UMaterial* Material, const FString& ParameterName, const FIntPoint& NodePos)
{
	UMaterialExpression* Expression = UMaterialEditingLibrary::CreateMaterialExpression(
		Material, UMaterialExpressionScalarParameter::StaticClass());
	UMaterialExpressionScalarParameter* Parameter = Cast<UMaterialExpressionScalarParameter>(Expression);

	Parameter->ParameterName = *ParameterName;
	Parameter->DefaultValue = 0.0f;
	Parameter->MaterialExpressionEditorX = NodePos.X;
	Parameter->MaterialExpressionEditorY = NodePos.Y;
	return Parameter;
}

UMaterialExpressionVectorParameter* UDDBridgeImportBase::CreateMaterialVectorParameter(
	UMaterial* Material, const FString& ParameterName, const FIntPoint& NodePos)
{
	UMaterialExpression* Expression = UMaterialEditingLibrary::CreateMaterialExpression(
		Material, UMaterialExpressionVectorParameter::StaticClass());
	UMaterialExpressionVectorParameter* Parameter = Cast<UMaterialExpressionVectorParameter>(Expression);

	if (Parameter)
	{
		Parameter->ParameterName = *ParameterName;
		Parameter->DefaultValue = FLinearColor::Black; // Default to black (0,0,0)
		Parameter->MaterialExpressionEditorX = NodePos.X;
		Parameter->MaterialExpressionEditorY = NodePos.Y;
	}

	return Parameter;
}

UMaterialExpressionVertexColor* UDDBridgeImportBase::CreateMaterialVertexColor(UMaterial* Material, const FIntPoint& NodePos)
{
	UMaterialExpression* Expression = UMaterialEditingLibrary::CreateMaterialExpression(
		Material, UMaterialExpressionVertexColor::StaticClass());

	UMaterialExpressionVertexColor* Node = Cast<UMaterialExpressionVertexColor>(Expression);
	Node->MaterialExpressionEditorX = NodePos.X;
	Node->MaterialExpressionEditorY = NodePos.Y;
	return Node;
}

UMaterial* UDDBridgeImportBase::DuplicateDDMaterial(const FString& OriginalAssetPath,
	const FString& DuplicatedFileDirectory,
	const FString& AssetName)
{
	// Load the original Material asset
	UMaterial* OriginalMaterial = Cast<UMaterial>(
		StaticLoadObject(UMaterial::StaticClass(), nullptr, *OriginalAssetPath));

	if (OriginalMaterial)
	{
		FAssetToolsModule& AssetToolsModule = FModuleManager::LoadModuleChecked<FAssetToolsModule>("AssetTools");
		IAssetTools& AssetTools = AssetToolsModule.Get();
		UObject* Obj = AssetTools.DuplicateAsset(AssetName, DuplicatedFileDirectory, OriginalMaterial);
		UMaterial* MATObj = Cast<UMaterial>(Obj);

		if (MATObj)
		{
			UE_LOG(LogTemp, Log, TEXT("Material duplicated successfully! New Path: %s"),
				*(DuplicatedFileDirectory / AssetName));
			return MATObj;
		}
		else
		{
			UE_LOG(LogTemp, Error, TEXT("Failed to duplicate Material asset."));
			return nullptr;
		}
	}
	else
	{
		UE_LOG(LogTemp, Error, TEXT("Original Material asset not found at path: %s"), *OriginalAssetPath);
		return nullptr;
	}
}

UBlueprint* UDDBridgeImportBase::DuplicateCharacterBlueprint(const FString& OriginalAssetPath,
                                                             const FString& DuplicatedFileDirectory,
                                                             const FString& AssetName)
{
	// Load the original Blueprint asset (ACharacter subclass)
	UBlueprint* OriginalBlueprint = Cast<UBlueprint>(
		StaticLoadObject(UBlueprint::StaticClass(), nullptr, *OriginalAssetPath));

	if (OriginalBlueprint)
	{
		// Ensure the blueprint is a subclass of ACharacter
		if (OriginalBlueprint->GeneratedClass->IsChildOf(ACharacter::StaticClass()))
		{
			// Duplicate the asset using DuplicateAsset method
			FAssetToolsModule& AssetToolsModule = FModuleManager::LoadModuleChecked<FAssetToolsModule>("AssetTools");
			IAssetTools& AssetTools = AssetToolsModule.Get();
			UObject* Obj = AssetTools.DuplicateAsset(AssetName, DuplicatedFileDirectory, OriginalBlueprint);
			UBlueprint* BPObj = Cast<UBlueprint>(Obj);

			if (BPObj)
			{
				UE_LOG(LogTemp, Log, TEXT("Blueprint duplicated successfully! New Path: %s"),
				       *(DuplicatedFileDirectory / AssetName));
				return BPObj;
			}
			else
			{
				UE_LOG(LogTemp, Error, TEXT("Failed to duplicate Blueprint asset."));
				return nullptr;
			}
		}
		else
		{
			UE_LOG(LogTemp, Warning, TEXT("The original asset is not a subclass of ACharacter."));
			return nullptr;
		}
	}
	else
	{
		UE_LOG(LogTemp, Error, TEXT("Original Blueprint asset not found at path: %s"), *OriginalAssetPath);
		return nullptr;
	}
}

UObject* UDDBridgeImportBase::CreateBlueprint(
	const FString& CharacterName,
	const FString& DestinationPath,
	const FString& BlueprintTemplatePath)
{
	// Validate DestinationPath and format it for a valid package path
	FString ValidDestinationPath = DestinationPath;
	if (!ValidDestinationPath.StartsWith(TEXT("/Game")))
	{
		ValidDestinationPath = FPaths::Combine(TEXT("/Game"), ValidDestinationPath);
	}
	ValidDestinationPath = ValidDestinationPath.Replace(TEXT("\\"), TEXT("/"));

	// Create the BP
	UBlueprint* NewBlueprint = DuplicateCharacterBlueprint(BlueprintTemplatePath, ValidDestinationPath,
	                                                       CharacterName);
	if (!NewBlueprint)
	{
		UE_LOG(LogTemp, Error, TEXT("Failed to create Blueprint for ACharacter."));
		return nullptr;
	}

	return NewBlueprint;
}

TMap<FString, FString> textureSlotToShaderReference = {
	{ "ALBEDO", "_BaseMap" },
	{ "_BaseColorMap", "_BaseMap" },
	{ "NORMAL", "_NormalMap" },
	{ "_NormalMap", "_NormalMap" },
	{ "MASK", "_MaskMap" },
	{ "_MaskMap", "_MaskMap" },
};

void UDDBridgeImportBase::ImportMaterialsFrom(UDDAssetCatalogData* AssetCatalogData,
                                              TArray<FDDSkinnedRenderObject> SkinnedRenderObjects,
                                              TArray<UTexture2D*> Textures, const FString& DestinationDir)
{
	// TODO 
	//NameToMaterial.Empty();
	for (auto SkinnedRenderObject : SkinnedRenderObjects)
	{
		for (const auto MaterialIndex : SkinnedRenderObject.materialExportIndex)
		{
			FDDMaterial Material = AssetCatalogData->materials[MaterialIndex];
			FString MaterialName = UDDBridgeUtils::NameSanitizer(Material.name.ToLower());
			if (PathToMaterial.Contains(DestinationDir / MaterialName))
				continue;

			UTexture2D *BC_Texture = nullptr, *N_Texture = nullptr, *ARM_Texture = nullptr;
			for (int32 j = 0; j < Material.textureExports.Num(); j++)
			{
				if (!textureSlotToShaderReference.Find(Material.textureExports[j].slot))
					continue;
				FString shaderReference = textureSlotToShaderReference[Material.textureExports[j].slot];
				if (shaderReference.Compare("_BaseMap") == 0)
				{
					check(Material.textureExports[j].textureVariationIndices.Num() > 0);
					int32 TextureIndex = Material.textureExports[j].textureVariationIndices[0];
					BC_Texture = Textures[TextureIndex];
				}
				else if (shaderReference.Compare("_NormalMap") == 0)
				{
					check(Material.textureExports[j].textureVariationIndices.Num() > 0);
					int32 TextureIndex = Material.textureExports[j].textureVariationIndices[0];
					N_Texture = Textures[TextureIndex];
				}
				else if (shaderReference.Compare("_MaskMap") == 0)
				{
					check(Material.textureExports[j].textureVariationIndices.Num() > 0);
					int32 TextureIndex = Material.textureExports[j].textureVariationIndices[0];
					ARM_Texture = Textures[TextureIndex];
				}
			}

			bool IsInvisible = Material.shader.Equals(TEXT("invisible"), ESearchCase::IgnoreCase);
			bool IsAlphaBlend = Material.shader.Contains(TEXT("WithAlphaWithBlends"), ESearchCase::IgnoreCase);
			UMaterial* NewMaterial;
			if (IsInvisible)
				NewMaterial = CreateInvisibleMaterial(MaterialName, DestinationDir);
			else
				NewMaterial = CreateMaterial(MaterialName, DestinationDir, BC_Texture, N_Texture, ARM_Texture, IsAlphaBlend);

			PathToMaterial.Add(DestinationDir / MaterialName.ToLower(), NewMaterial);
		}
	}
}
