// Fill out your copyright notice in the Description page of Project Settings.

#include "DDDataAssetFactory.h"
#include "DataAssets/DDDataAssets.h"
#include "DataAssets/DDCharacterEditorSkinCompositionData.h"
#include "Misc/FileHelper.h"
#include "Json.h"
#include "JsonUtilities.h"


UDDDataAssetFactory::UDDDataAssetFactory()
{
    SupportedClass = nullptr;
    Formats.Add(TEXT("json;Didimo Asset Catalog"));
    bCreateNew = false;
    bEditAfterNew = true;
    bEditorImport = true;
}

UObject* UDDDataAssetFactory::FactoryCreateFile(
    UClass* InClass,
    UObject* InParent,
    FName InName,
    EObjectFlags Flags,
    const FString& Filename,
    const TCHAR* Parms,
    FFeedbackContext* Warn,
    bool& bOutOperationCanceled)
{
    check(JsonObject->HasField(TEXT("fileType")));
    FString FileType = JsonObject->GetStringField(TEXT("fileType"));

    bOutOperationCanceled = false;
    if (FileType.Compare("ce_skin_composition") == 0)
    {
        return CreateCharacterEditorSkinComposition(InClass, InParent, InName, Flags);
    }

    // TODO: implement FileType-based check, like above
    else if (JsonObject->HasField(TEXT("characterEditors")))
    {
        return CreateAssetCatalog(InClass, InParent, InName, Flags);
    }
    else if (JsonObject->HasField(TEXT("rigShapePath")))
    {
        return CreateCharacterEditorConfig(InClass, InParent, InName, Flags);
    }
    else if (JsonObject->HasField(TEXT("groups")))
    {
        return CreateCharacterDescriptorConfig(InClass, InParent, InName, Flags);
    }
    else
    {
        UE_LOG(LogTemp, Error, TEXT("Failed to parse JSON file: %s"), *Filename);
        bOutOperationCanceled = true; // Mark the operation as canceled in case of failure
        return nullptr;
    }
}

UObject* UDDDataAssetFactory::FactoryCreateNew(
    UClass* InClass,
    UObject* InParent,
    FName InName,
    EObjectFlags Flags,
    UObject* Context,
    FFeedbackContext* Warn)
{
    return NewObject<UDDAssetCatalogData>(InParent, InClass, InName, Flags);
}

bool UDDDataAssetFactory::FactoryCanImport(const FString& Filename)
{
    // Load the file into a string
    FString jsonString;
    if (!FFileHelper::LoadFileToString(jsonString, *Filename))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to load file: %s"), *Filename);
        return false;
    }

    // Parse the JSON string
    TSharedRef<TJsonReader<>> jsonReader = TJsonReaderFactory<>::Create(jsonString);
    if (!FJsonSerializer::Deserialize(jsonReader, JsonObject))
        return false;
    if (!JsonObject.IsValid())
        return false;
    if (ResolveSupportedClass() == nullptr)
        return false;

    // Is ok
    return true;
}

UClass* UDDDataAssetFactory::ResolveSupportedClass()
{
    check(JsonObject->HasField(TEXT("fileType")));
    FString FileType = JsonObject->GetStringField(TEXT("fileType"));
    if (FileType.Compare("ce_skin_composition") == 0)
    {
        return UDDCharacterEditorSkinCompositionData::StaticClass();
    }
    else if (JsonObject->HasField(TEXT("characterEditors")))
    {
        return UDDAssetCatalogData::StaticClass();
    }
    else if (JsonObject->HasField(TEXT("rigShapePath")))
    {
        return UDDCharacterEditorConfigData::StaticClass();
    }
    else if (JsonObject->HasField(TEXT("groups")))
    {
        return UDDCharacterDescriptorData::StaticClass();
    }
    else
    {
        return nullptr;
    }
}

bool UDDDataAssetFactory::DoesSupportClass(UClass* Class)
{
    return 
        Class == UDDAssetCatalogData::StaticClass() ||
        Class == UDDCharacterEditorConfigData::StaticClass() ||
        Class == UDDCharacterDescriptorData::StaticClass();
}

bool UDDDataAssetFactory::ConfigureProperties()
{
    // Implement any specific configuration logic here
    return true;
}

UObject* UDDDataAssetFactory::CreateAssetCatalog(
    UClass* InClass,
    UObject* InParent,
    FName InName,
    EObjectFlags Flags)
{
    // Create a new instance
    auto NewAsset = NewObject<UDDAssetCatalogData>(InParent, InClass, InName, Flags);
    if (!NewAsset)
    {
        UE_LOG(LogTemp, Error, TEXT("Failed to create UDDAssetCatalogData object."));
        return nullptr;
    }

    // Populate the asset from JSON
    if (!FJsonObjectConverter::JsonObjectToUStruct(JsonObject.ToSharedRef(), UDDAssetCatalogData::StaticClass(), NewAsset, 0, 0))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to parse JSON into UDDAssetCatalogData"));
        return nullptr;
    }

    // At this point, NewAsset contains the parsed data from the json file
    UE_LOG(LogTemp, Log, TEXT("Successfully created UDDAssetCatalogData from JSON."));

    // Try to map the mask textures to their places (textures need to be imported at this point)
    for (auto& texture : NewAsset->textures)
    {
        if (texture.path.IsEmpty())
        {
            texture.texture = nullptr;
            UE_LOG(LogTemp, Warning, TEXT("AssetCatalog texture not found"));
        }
        else
        {
            FString LongPackagePath = FPackageName::GetLongPackagePath(NewAsset->GetOutermost()->GetName());
            FString Filepath = LongPackagePath / texture.path;
            FString FilepathWithoutExtension = FPaths::GetBaseFilename(Filepath, false);
            texture.texture = Cast<UTexture2D>(StaticLoadObject(UTexture2D::StaticClass(), nullptr, *FilepathWithoutExtension));
        }
    }

    return NewAsset;
}

UObject* UDDDataAssetFactory::CreateCharacterDescriptorConfig(
    UClass* InClass,
    UObject* InParent,
    FName InName,
    EObjectFlags Flags)
{
    // Create a new instance
    auto NewAsset = NewObject<UDDCharacterDescriptorData>(InParent, InClass, InName, Flags);
    if (!NewAsset)
    {
        UE_LOG(LogTemp, Error, TEXT("Failed to create UDDAssetCatalogData object."));
        return nullptr;
    }

    // Populate the asset from JSON
    if (!FJsonObjectConverter::JsonObjectToUStruct(JsonObject.ToSharedRef(), UDDCharacterDescriptorData::StaticClass(), NewAsset, 0, 0))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to parse JSON into UDDCharacterDescriptorData"));
        return nullptr;
    }

    // At this point, NewAsset contains the parsed data from the json file
    UE_LOG(LogTemp, Log, TEXT("Successfully created UDDCharacterDescriptorData from JSON."));

    return NewAsset;
}

UObject* UDDDataAssetFactory::CreateCharacterEditorConfig(
    UClass* InClass,
    UObject* InParent,
    FName InName,
    EObjectFlags Flags)
{
    // Create a new instance
    auto NewAsset = NewObject<UDDCharacterEditorConfigData>(InParent, InClass, InName, Flags);
    if (!NewAsset)
    {
        UE_LOG(LogTemp, Error, TEXT("Failed to create UDDCharacterEditorConfigData object."));
        return nullptr;
    }

    // Populate the asset from JSON
    if (!FJsonObjectConverter::JsonObjectToUStruct(JsonObject.ToSharedRef(), UDDCharacterEditorConfigData::StaticClass(), NewAsset, 0, 0))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to parse JSON into UDDCharacterEditorConfigData"));
        return nullptr;
    }

    // At this point, NewAsset contains the parsed data from the json file
    UE_LOG(LogTemp, Log, TEXT("Successfully created UDDCharacterEditorConfigData from JSON."));

    // Try to map the albedo textures to their places
    for (int32 MeshAlbedoIndex = 0; MeshAlbedoIndex < NewAsset->meshAlbedos.Num(); MeshAlbedoIndex++)
    {
        auto MeshAlbedo = &NewAsset->meshAlbedos[MeshAlbedoIndex];
        FString LongPackagePath = FPackageName::GetLongPackagePath(NewAsset->GetOutermost()->GetName());
        FString Filepath = LongPackagePath + TEXT("/") + MeshAlbedo->texturePath;
        FString FilepathWithoutExtension = FPaths::GetBaseFilename(Filepath, false);
        UTexture2D* TextureAsset = Cast<UTexture2D>(StaticLoadObject(UTexture2D::StaticClass(), nullptr, *FilepathWithoutExtension));
        MeshAlbedo->texture = TextureAsset;
    }

    //// Try to map the mask textures to their places (textures need to be imported at this point)
    //for (int32 SegmentationIndex = 0; SegmentationIndex < NewAsset->segmentations.Num(); SegmentationIndex++)
    //{
    //    auto Segmentation = &NewAsset->segmentations[SegmentationIndex];
    //    for (int32 TextureEditIndex = 0; TextureEditIndex < Segmentation->textureEdit.Num(); TextureEditIndex++)
    //    {
    //        auto TextureEdit = &Segmentation->textureEdit[TextureEditIndex];
    //        FString LongPackagePath = FPackageName::GetLongPackagePath(NewAsset->GetOutermost()->GetName());

    //        // Set gradient texture
    //        if (TextureEdit->gradient.gradientTexturePath.IsEmpty())
    //        {
    //            TextureEdit->gradient.gradientTexture = nullptr;
    //            UE_LOG(LogTemp, Warning, TEXT("CE Config gradient texture not found"));
    //        }
    //        else
    //        {
    //            FString Filepath = LongPackagePath + TEXT("/") + TextureEdit->gradient.gradientTexturePath;
    //            FString FilepathWithoutExtension = FPaths::GetBaseFilename(Filepath, false);
    //            UTexture2D* GradientTextureAsset = Cast<UTexture2D>(StaticLoadObject(UTexture2D::StaticClass(), nullptr, *FilepathWithoutExtension));
    //            TextureEdit->gradient.gradientTexture = GradientTextureAsset;
    //        }

    //        // Set mask texture
    //        if (TextureEdit->maskPath.IsEmpty())
    //        {
    //            TextureEdit->maskTexture = nullptr;
    //            UE_LOG(LogTemp, Warning, TEXT("CE Config mask texture not found"));
    //        }
    //        else
    //        {
    //            FString Filepath = LongPackagePath + TEXT("/") + TextureEdit->maskPath;
    //            FString FilepathWithoutExtension = FPaths::GetBaseFilename(Filepath, false);
    //            UTexture2D* MaskTextureAsset = Cast<UTexture2D>(StaticLoadObject(UTexture2D::StaticClass(), nullptr, *FilepathWithoutExtension));
    //            TextureEdit->maskTexture = MaskTextureAsset;
    //        }
    //    }
    //}

    return NewAsset;
}

UObject* UDDDataAssetFactory::CreateCharacterEditorSkinComposition(UClass* InClass, UObject* InParent, FName InName, EObjectFlags Flags)
{
    // Create a new instance
    auto NewAsset = NewObject<UDDCharacterEditorSkinCompositionData>(InParent, InClass, InName, Flags);
    if (!NewAsset)
    {
        UE_LOG(LogTemp, Error, TEXT("Failed to create UDDCharacterEditorSkinCompositionData object."));
        return nullptr;
    }

    // Populate the asset from JSON
    if (!FJsonObjectConverter::JsonObjectToUStruct(JsonObject.ToSharedRef(), UDDCharacterEditorSkinCompositionData::StaticClass(), NewAsset, 0, 0))
    {
        UE_LOG(LogTemp, Warning, TEXT("Failed to parse JSON into UDDCharacterEditorSkinCompositionData"));
        return nullptr;
    }

    // At this point, NewAsset contains the parsed data from the json file
    UE_LOG(LogTemp, Log, TEXT("Successfully created UDDCharacterEditorSkinCompositionData from JSON."));

    // Try to map the mask textures to their places (textures need to be imported at this point)
    FString LongPackagePath = FPackageName::GetLongPackagePath(NewAsset->GetOutermost()->GetName());
    for (int32 ComponentIndex = 0; ComponentIndex < NewAsset->components.Num(); ComponentIndex++)
    {
        auto Component = &NewAsset->components[ComponentIndex];

        // Handler textures (e.g. color pickers)
        for (int32 HandlerIndex = 0; HandlerIndex < Component->handlers.Num(); HandlerIndex++)
        {
            auto Handler = &Component->handlers[HandlerIndex];
            if (Handler->type.Compare("PixelPickerElement") == 0)
            {
                check(Handler->stringParams.Num() == 1);
                FString Filepath = LongPackagePath / Handler->stringParams[0];
                FString FilepathWithoutExtension = FPaths::GetBaseFilename(Filepath, false);
                UTexture2D* TextureAsset = Cast<UTexture2D>(StaticLoadObject(UTexture2D::StaticClass(), nullptr, *FilepathWithoutExtension));

                Handler->textureParams.SetNum(1);
                Handler->textureParams[0] = TextureAsset;
            }
        }

        // Make sure the textures that are referenced by strings are also referenced as assets
        // otherwise they might get lost during the COOK process
        for (int32 j = 0; j < Component->processorStacks.Num(); j++)
        {
            auto Stack = &Component->processorStacks[j];
            for (size_t k = 0; k < Stack->processors.Num(); k++)
            {
                auto Processor = &Stack->processors[k];
                for (int32 PathIndex = 0; PathIndex < Processor->stringParams.Num(); PathIndex++)
                {
                    if (Processor->stringParams[PathIndex].Len() > 0)
                    {
                        FString NormalizedPath = FPaths::ConvertRelativePathToFull(FPaths::Combine(LongPackagePath, Processor->stringParams[PathIndex]));
                        FString AssetPath = FPaths::GetPath(NormalizedPath) / FPaths::GetBaseFilename(NormalizedPath);
                        UTexture2D* TextureAsset = Cast<UTexture2D>(StaticLoadObject(UTexture2D::StaticClass(), nullptr, *AssetPath));
                        if (TextureAsset)
                            Processor->textureReferences.Add(TextureAsset);
                    }
                }
            }
        }
    }

    return NewAsset;
}
