// Copyright Epic Games, Inc. All Rights Reserved.

#include "DDPopul8Editor.h"
#include "ContentBrowserModule.h"
#include "IContentBrowserSingleton.h"
#include "ToolMenus.h"
#include "ContentBrowserModule.h"
#include "IContentBrowserSingleton.h"
#include "ToolMenus.h"
#include "IDesktopPlatform.h"
#include "DesktopPlatformModule.h"
#include "PropertyCustomizationHelpers.h"
#include "DataAssets/DDVertexCacheData.h"
#include "AssetRegistry/AssetRegistryModule.h"
#include <DDBridgeProjectImport.h>


#define LOCTEXT_NAMESPACE "FPopul8EditorModule"
//#define ADDITIONAL_OPTIONS


void FDDPopul8EditorModule::StartupModule()
{
	// Register menu extension
	FToolMenuOwnerScoped MenuOwner(this);
	UToolMenus::RegisterStartupCallback(FSimpleMulticastDelegate::FDelegate::CreateRaw(this, &FDDPopul8EditorModule::RegisterMenus));

	// Register tab
	FGlobalTabmanager::Get()->RegisterNomadTabSpawner("Popul8Tab", FOnSpawnTab::CreateRaw(this, &FDDPopul8EditorModule::CreatePluginTab))
		.SetDisplayName(LOCTEXT("Popul8TabTitle", "Popul8"))
		.SetMenuType(ETabSpawnerMenuType::Hidden);
}

void FDDPopul8EditorModule::ShutdownModule()
{
	UToolMenus::UnregisterOwner(this);
	FGlobalTabmanager::Get()->UnregisterNomadTabSpawner("Popul8Tab");
}

UDDAttachmentVertexCacheData* FDDPopul8EditorModule::CreateGarmentCullingCacheData(
	const FString& BinFileDirectory,
	USkeletalMesh* TargetSkeletalMesh, 
	const FString& DestinationDirectory)
{
	//// Modify the asset name for the new data asset
	FString AttachmentVertexCacheDataName = TEXT("VertexCullData");
	FString AttachmentVertexCachePackagePath = DestinationDirectory / AttachmentVertexCacheDataName;

	//// Create a new package for the data asset
	UPackage* Package = CreatePackage(*AttachmentVertexCachePackagePath);
	if (!Package)
	{
		UE_LOG(LogTemp, Error, TEXT("Failed to create package for Vertex Cache Data."));
		return nullptr;
	}

	//// Create the Vertex Cache Data asset
	UDDAttachmentVertexCacheData* AttachmentVertexCacheData = NewObject<UDDAttachmentVertexCacheData>(Package, FName(*AttachmentVertexCacheDataName), RF_Public | RF_Standalone);
	if (!AttachmentVertexCacheData)
	{
		UE_LOG(LogTemp, Error, TEXT("Failed to create Vertex Cache Data asset."));
		return nullptr;
	}

	//// Fill out the data
	AttachmentVertexCacheData->TargetSkeletalMesh = TargetSkeletalMesh;
	AttachmentVertexCacheData->BinFileDirectory = BinFileDirectory;
	AttachmentVertexCacheData->BuildData();

	//// Mark the package dirty and save it
	AttachmentVertexCacheData->MarkPackageDirty();
	FAssetRegistryModule::AssetCreated(AttachmentVertexCacheData);
	//FString PackageFileName = FPackageName::LongPackageNameToFilename(VertexCachePackagePath, FPackageName::GetAssetPackageExtension());
	//if (!UPackage::SavePackage(Package, VertexCacheData, EObjectFlags::RF_Public | RF_Standalone, *PackageFileName))
	//{
	//    UE_LOG(LogTemp, Error, TEXT("Failed to save Vertex Cache Data package."));
	//}
	UE_LOG(LogTemp, Log, TEXT("Successfully created Vertex Cache Data: %s"), *AttachmentVertexCacheDataName);

	// Open the content browser at the asset's location
	TArray<UObject*> Selected = { AttachmentVertexCacheData };
	FModuleManager::LoadModuleChecked<FContentBrowserModule>("ContentBrowser").Get().SyncBrowserToAssets(Selected);
	return AttachmentVertexCacheData;
}

void FDDPopul8EditorModule::RegisterMenus()
{
	// Extend the Level Editor menu bar
	UToolMenu* Menu = UToolMenus::Get()->ExtendMenu("LevelEditor.MainMenu.Window");
	FToolMenuSection& Section = Menu->FindOrAddSection("WindowLayout");
	Section.AddMenuEntry(
		"OpenPopul8Window",
		LOCTEXT("OpenPopul8Window", "Popul8 Plugin Window"),
		LOCTEXT("OpenPopul8WindowTooltip", "Opens Popul8 Plugin Window"),
		FSlateIcon(),
		FUIAction(FExecuteAction::CreateRaw(this, &FDDPopul8EditorModule::OpenWindow))
	);
}

void FDDPopul8EditorModule::OpenWindow()
{
	FGlobalTabmanager::Get()->TryInvokeTab(FName("Popul8Tab"));
}

TSharedRef<SDockTab> FDDPopul8EditorModule::CreatePluginTab(const FSpawnTabArgs& TabArgs)
{
	FSlateFontInfo BoldFont = FCoreStyle::GetDefaultFontStyle("BoldFont", 18);

	// Create the dock tab with the asset picker inside
	return SNew(SDockTab)
		.TabRole(ETabRole::NomadTab)
		[
			SNew(SVerticalBox)

				//**************************************************************************************************************************
				// Bridge project import ///////////////////////////////////////////////////////////////////////////////////////////////////
				//**************************************************************************************************************************
				+SVerticalBox::Slot()
				.AutoHeight()
				.HAlign(HAlign_Center)
				.Padding(0, 15, 0, 10)
				[
					SNew(STextBlock)
						.Text(LOCTEXT("Popul8BridgeHeader", "Popul8 Bridge"))
						.Font(BoldFont)
				]

				// Button
				+ SVerticalBox::Slot()
				.AutoHeight()
				.HAlign(HAlign_Center)
				[
					SNew(SButton)
						.Text(LOCTEXT("MyButtonLabel", "Select Project Export Folder"))
						.OnClicked_Lambda([this]() -> FReply {
						this->SelectProjectExportFolder();
						return FReply::Handled();
							})
				]

#ifdef ADDITIONAL_OPTIONS

				//************************************************************************************************************************
				// Garment culling information ///////////////////////////////////////////////////////////////////////////////////////////
				//************************************************************************************************************************				

					// Spacing
					+SVerticalBox::Slot()
					.AutoHeight()
					.Padding(0.f, 15.f) // Add 15px of vertical space
					[
						SNew(SSpacer)
					]

					// Second Section Header
					+SVerticalBox::Slot()
					.AutoHeight()
					.HAlign(HAlign_Center)
					.Padding(0, 15, 0, 10)
					[
						SNew(STextBlock)
							.Text(LOCTEXT("GarmentCullingHeader", "Garment Culling Configuration"))
							.Font(BoldFont)
					]


				// Button
				+SVerticalBox::Slot()
				.AutoHeight()
				.HAlign(HAlign_Center)
				[
					SNew(SButton)
						.Text(LOCTEXT("MyButtonLabel", "Select cull bin information "))
						.OnClicked_Lambda([this]() -> FReply {
						this->SelectCullBin();
						return FReply::Handled();
							})
				]

				// Piker
				+ SVerticalBox::Slot()
				.AutoHeight()
				.HAlign(HAlign_Center)
				[
					SNew(SHorizontalBox)

						// Label
						+ SHorizontalBox::Slot()
						.AutoWidth()
						.VAlign(VAlign_Center)
						.Padding(5.f, 0.f)
						[
							SNew(STextBlock)
								.Text(LOCTEXT("ObjectPickerLabel", "Attachment Mesh"))
						]

						// Object Picker
						+ SHorizontalBox::Slot()
						.FillWidth(1.f)
						[
							SNew(SObjectPropertyEntryBox)
								.ObjectPath_Lambda([this]() -> FString {
								return TargetSkeletalMesh ? TargetSkeletalMesh->GetPathName() : FString();
									})
								.OnObjectChanged_Lambda([this](const FAssetData& AssetData) {
								TargetSkeletalMesh = Cast<USkeletalMesh>(AssetData.GetAsset());
									})
								.AllowedClass(USkeletalMesh::StaticClass())
								.DisplayThumbnail(true)
								.EnableContentPicker(true)
								.DisplayBrowse(true)
						]
				]

				// Spacing
				+ SVerticalBox::Slot()
				.AutoHeight()
				.Padding(0.f, 10.f) // Add 10px of vertical space
				[
					SNew(SSpacer)
				]

				// Button
				+ SVerticalBox::Slot()
				.AutoHeight()
				.HAlign(HAlign_Center)
				[
					SNew(SButton)
						.Text(LOCTEXT("MyButtonLabel", "Build Vertex and cull Data"))
						.OnClicked_Lambda([this]() -> FReply {
						this->CreateGarmentCullingCacheData();
						return FReply::Handled();
							})
				]
#endif
		];
}

void FDDPopul8EditorModule::SelectCullBin()
{
	IDesktopPlatform* DesktopPlatform = FDesktopPlatformModule::Get();
	if (DesktopPlatform)
	{
		const void* ParentWindowHandle = nullptr;

		// Define the file dialog filter for .bin files
		TArray<FString> OutFiles;
		const FString FileTypes = TEXT("bin files (*.bin)|*.bin");

		// Open the directory picker dialog
		bool bFileSelected = DesktopPlatform->OpenFileDialog(
			ParentWindowHandle,
			TEXT("Select Bin File"),  // Dialog title
			TEXT(""),                 // Default path (empty for default system path)
			TEXT(""),                 // Default file (empty for no default file)
			FileTypes,                // File types filter as a single string
			0,                        // No flags
			OutFiles                  // Output selected file(s)
		);

		// If the user selected a folder, use it
		if (bFileSelected)
		{
			SelectedBinFileDirectory = OutFiles[0];
		}
		else
		{
			UE_LOG(LogTemp, Warning, TEXT("No folder selected."));
			SelectedBinFileDirectory = "";
		}
	}
	else
	{
		UE_LOG(LogTemp, Error, TEXT("Desktop Platform is unavailable."));
		SelectedBinFileDirectory = "";
	}
}

void FDDPopul8EditorModule::CreateGarmentCullingCacheData()
{
	UE_LOG(LogTemp, Log, TEXT("Building Garment Cull Cache Data!"));
	if (SelectedBinFileDirectory.IsEmpty())
	{
		UE_LOG(LogTemp, Warning, TEXT("SelectedBinFileDirectory is empty."));
		return;
	}

	if (!TargetSkeletalMesh)
	{
		UE_LOG(LogTemp, Warning, TEXT("GarmentSkeletalMesh is null."));
		return;
	}

	FString MeshPath = TargetSkeletalMesh->GetPathName();
	FString PackageRoot, PackagePath, AssetName;

	if (!FPackageName::SplitLongPackageName(MeshPath, PackageRoot, PackagePath, AssetName))
	{
		UE_LOG(LogTemp, Warning, TEXT("Failed to split package path for skeletal mesh."));
		return;
	}

	CreateGarmentCullingCacheData(SelectedBinFileDirectory, TargetSkeletalMesh, PackageRoot / PackagePath);
}

void FDDPopul8EditorModule::SelectProjectExportFolder()
{
	IDesktopPlatform* DesktopPlatform = FDesktopPlatformModule::Get();
	if (DesktopPlatform)
	{
		const void* ParentWindowHandle = nullptr;

		// Open the directory picker dialog
		bool bFolderSelected = DesktopPlatform->OpenDirectoryDialog(
			ParentWindowHandle,
			TEXT("Select Directory"),         // Dialog title
			TEXT(""),                         // Default path (empty for default system path)
			SelectedProjectExportFolder                    // Output selected folder
		);

		// If the user selected a folder, use it
		if (bFolderSelected)
		{
			if (!SelectedProjectExportFolder.IsEmpty())
			{
				// Perform the import using the selected folder
				TStrongObjectPtr<UDDBridgeProjectImport> BridgeImport =
					TStrongObjectPtr<UDDBridgeProjectImport>(NewObject<UDDBridgeProjectImport>());
				BridgeImport->Import(SelectedProjectExportFolder);
			}
		}
		else
		{
			UE_LOG(LogTemp, Warning, TEXT("No folder selected."));
			SelectedProjectExportFolder =  "";
		}
	}
	else
	{
		UE_LOG(LogTemp, Error, TEXT("Desktop Platform is unavailable."));
		SelectedProjectExportFolder =  "";
	}

}


#undef LOCTEXT_NAMESPACE

IMPLEMENT_MODULE(FDDPopul8EditorModule, Popul8Editor)